package MaintainData;

import java.io.Serializable;
import java.util.Date;
import java.util.concurrent.Semaphore;

/*
 * WIRING DIAGRAM :
 * @LOOKSY: CelebrityTable.java
 * @LOOKSY:ArtLeavesInterfaces.java
 *
 * Class Name : public class CelebrityTable implements Serializable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH because it is too very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : None
 *
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap. The data file
 * will contain the different celebrity names recognised by our system. A readymade data entry file will 
 * be provided to you dont loose that take a physical backup and keep safe. You may also use Data Entry
 * programs to update this list.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists the celebrities recognized by this system.
 *
 * Technical References : Buy and Read Asterix and the Cauldron.
 * Table Dedicated to : King Crimson (In the Court of the Crimson King  An Observation by King Crimson)
 *                      Track : MoonChild including the DREAM and the ILLUSION
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 * 
 * Last Modification Date : 8th August 2012
 * Last Modifications : Added new clone () Method
 * 
 * Last Modification Date : 10th September 2012
 * Last Modifications : changed private static to private transient and removed static init block
 *
 * Last Modification Date : 11th September 2012
 * Last Modifications : removed rowlock.acquireUninterruptibly(); and rowlock.release()
 * statements from the clone method because each of the setters will do the above.
 * 
 * 
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class CelebrityTable implements Serializable, Tables, Cloneable {
    private String celebname; /* 1 FIELD */

    private transient CelebrityTable celeb;
    private transient int ctn ;
    private transient Semaphore rowlock;  // 10th September 2012

    {
        ctn = LiveDB.clt;
        rowlock = new Semaphore(1,true);
    }

    // This is the Superb constructor

    public CelebrityTable ()
    {
        celebname = null;
    }

    public CelebrityTable (String p_Celebname)
    {
        if (p_Celebname == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Celebname:" + p_Celebname));
            System.exit(1);
        }
        
        rowlock.acquireUninterruptibly();
        // this we will validate here the others from the user interface
        for (int i = 0; i < p_Celebname.toCharArray().length; i++)
        {
            if (Character.isDigit(p_Celebname.toCharArray()[i])
               || Character.isHighSurrogate(p_Celebname.toCharArray()[i])
               || Character.isISOControl(p_Celebname.toCharArray()[i])
               || Character.isLowSurrogate(p_Celebname.toCharArray()[i])
               || Character.isMirrored(p_Celebname.toCharArray()[i]))
            {
               LiveDB.ALRTLVL = 15;
               System.err.println(LiveDB.grassHopper("Bad Name for a Celebrity " + p_Celebname));
               System.exit(1);
            }
        }        
        this.celebname = new String(p_Celebname.toCharArray());
        rowlock.release();
    }

    // Added new clone () method on 8th August 2012
    @Override
    protected Object clone (){        
           
        try{
            celeb = (CelebrityTable)super.clone();
            rowlock.acquireUninterruptibly();
            celeb.celebname = celeb.getCelebname();
            rowlock.release();
        }catch (CloneNotSupportedException ex_NOKIA54){
            throw new InternalError (ex_NOKIA54.toString());
        }        
                
        return celeb;
    }
    
    protected String getCelebname(){
        return new String(celebname.toCharArray());
    }

    @Override
    public String toString(){
        return getCelebname();
    }

    @Override
    public final boolean COMMIT(){
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(ctn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
