/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package MaintainData;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static MaintainData.LiveDB.ft;
import static MaintainData.LiveDB.it;
import static MaintainData.LiveDB.crr;
import static MaintainData.LiveDB.sc;
import static MaintainData.LiveDB.rt;
import static MaintainData.LiveDB.pc;
import static MaintainData.LiveDB.ct;
import java.util.Arrays;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:ChangeTracker.java
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:RefreshDatabase.read_database()
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.CopyFile, LiveDB.commit_database(), LiveDB.write_IndexFile()
 * @LOOKSY:LiveDB.CTSR, LiveDB.LiveSharedDB, LiveDB.ALRTLVL, LiveDB.SerializeJustOnceTable(),
 *         LiveDB.prepareCrashRecovery()
 * @LOOKSY:InstallDB.PowerOutageSensor()
 *
 * Class Name : CrashRecovery implements Externalizable, Cloneable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here other than COMMIT, COMMITINDEXES, REFRESH.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxcrashRecovery
 * searchkey = AnyValidTableNumber(Long)
 * searchkey returns String containing the completed rowdata for the table
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : This is a critical table for crash recovery purposes. This table will contain a
 * single row for each and every table except this CrashRecovery table and LiveDB.SerializeJustOnceTable().
 * This single row is the last row for tableno on which some DML happened. These SerializeJustOnce tables
 * contain static data which is loaded at Install time and then serialized and then they are only used for
 * lookup gets .... Along with any kind of DML on tables that row is offered to crash recovery table to be
 * saved using LiveDB.prepareCrashRecovery(). This method is called by ChangeTracker constructor.
 * LiveDB.prepareCrashRecovery() prepares the CrashRecovery row and then calls this CrashRecovery constructor.
 * There are two kinds of crashes one is a clean crash where some DML happened on some table and then the
 * system crashes but that DML was or wasnot committed to the physical database. So if that DML was not
 * committed to the physical table on disk then we have a backup row contained in this table for crash
 * recovery purposes. So LiveDB.PowerOutageSensor() meant to deal with such clean crashes picks up the
 * rows in the CrashRecovery table and then checks to see if it has been Updated or Inserted in that
 * table. If not it does that for you. NOTE:- This table contains only rows for Update and Insert DML's.
 * So this table helps you to recover revenues if there was a customer billing in ShoppingCart or even
 * a new customer who registered.
 * The second type of crash is a Dirty Harry crash where fsck() does the job. Dirty Harry crash happens
 * due to a system crash during serialization by whatever mechanism and so the database is inconsistent.
 * we are not dealing with dirty harry crashes here.
 *
 * Technical References : Buy and Read Asterix and the Falling Sky.
 *
 * Start Date: 18th October 2010
 * End Date: 15th May 2011
 *
 * Last Modification Date : 26th May 2012
 * Last Modifications : Silly Me slight small bug added this patch of code in the constructor
 * Actually not really a bug but a small patch of code to catch a programming screwup
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid CrashRecovery table number field:" + p_Tableno));
            System.exit(1);
        }
 *
 * Last Modification Date : 3rd June 2012
 * Last Modifications : Made changes to setter setCrrowid to look for negative rowid passed as parameter
 * Also changed some hardcodings in the initialization blocks crn = LiveDB.crr and itn = LiveDB.it
 *
 * Last Modification Date : 25th June 2012
 * Last Modifications : Marked as 25th June 2012
 * 
 * Last Modification Date : 8th August 2012
 * Last Modifications : As below
        // for every writeUTF two extra bytes of length information are written to file
        // so we must add that length information to the return value of SIZEOF
        int utfextras = values.length * 2; // 8th August 2012
 *
 * Last Modification Date : 14th Augustus 2012
 * Last Modification : I have added this call to LiveDB.growMore(ct) whereever I am growing
 * the LiveDB.LiveSharedDB array with a call to new.
 * 
 * Last Modification Date : 8th September 2012
 * Last Modifications : Marked as 8th September 2012
 * 
 * Last Modification Date : 10th September 2012
 * Last Modifications : Marked as 10th September 2012
 * Declared a new variable private transient int thisrow; and its usage
 * 
 * Last Modification Date : 11th September 2012
 * Last Modifications : removed rowlock.acquireUninterruptibly(); and rowlock.release()
 * statements from the clone method because each of the setters will do the above.
 * 
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

final class CrashRecovery implements Externalizable, Cloneable, Tables {
    private long tableno; // tableno for which CrashRecovery row is being stored
    private long rowid;  /* this rowid is the rowid for other tableno and not this CrashRecovery table */
    private String values [];  /* row values  pertaining to tableno*/
    private long crrowid; // the rowid for this table CrashRecovery
    /* 4 FIELDS */
    private static final long serialVersionUID = 99997L;
    private transient int thisrow; // 10th September 2012
    private transient long fieldchanges;
    private transient File srf;
    private transient String [] var;
    private transient int crn;
    private transient int itn;    
    private transient CrashRecovery cr;
    private transient Semaphore rowlock;

    {
       thisrow = 0; 
       fieldchanges = 0L;
       srf = null;
       var = null; // just to trim and return values array
       crn = LiveDB.crr;
       itn = LiveDB.it;
       cr = null;       
       rowlock = new Semaphore(1, true);
    }

    public CrashRecovery () {
        tableno = 0;
        rowid = 0;
        values = (String []) new String [1];
        crrowid = 0;
    }
    
    public CrashRecovery (long p_Tableno, long p_Rowid, long p_crrid, String [] p_Values){
        rowlock.acquireUninterruptibly();
        if ( p_Tableno == ft || p_Tableno == ct || p_Tableno == LiveDB.cr || p_Tableno == sc || p_Tableno == crr
             || p_Tableno == it || p_Tableno == rt || p_Tableno == pc )
             tableno = p_Tableno;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid CrashRecovery table number field:" + p_Tableno));
            System.exit(1);
        }        

        if (p_Rowid < 0 || p_Rowid >= LiveDB.LiveSharedDB[(int)p_Tableno].length)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid " + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        rowid = p_Rowid;

        // the length count here will include this object that the constructor is constructing
        // the object has already been created and initialized. The constructor is filling in the blanks
        
        // 8th September 2012  --- this check below over and above the nested checks
        if (LiveDB.LiveSharedDB[crn] != null)
        {
            if (LiveDB.LiveSharedDB[crn].length > 1) // Added on 25th June 2012    
            {
                thisrow = LiveDB.LiveSharedDB[crn].length -2;
                if (p_crrid != ((CrashRecovery)LiveDB.LiveSharedDB[crn][thisrow]).getCrrowid() + 1)
                {
                    LiveDB.ALRTLVL = 7;
                    System.err.println(LiveDB.grassHopper("Invalid rowid " + String.valueOf(p_Rowid)));
                    System.exit(1);
                }
            }
        }
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid " + String.valueOf(p_Rowid)));
            System.exit(1);            
        }
        crrowid = p_crrid;
        rowlock.release();
        setValues(p_Values);
        
    }
    @Override
    protected Object clone (){        
        
        try{
            cr = (CrashRecovery)super.clone();
            rowlock.acquireUninterruptibly();
            for (int vi = 0; vi < this.getValues().length; vi++)
                cr.values[vi] = new String(this.values[vi].toCharArray());
            rowlock.release();
        }catch (CloneNotSupportedException ex_NOKIA88){
            throw new InternalError (ex_NOKIA88.toString());
        }
        
        return cr;
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
        int thisrow = 0;
        String file = null;
 
        // first find the rowid for this table itn = 14 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(crn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("CrashRecovery misses entry in indxinstalledTables"));
            // perhaps the index is corrupted so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == crn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("CrashRecovery misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(crn));
        
        thisrow = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno();
        file = ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][thisrow]).getFilename();
        file = LiveDB.serializedfilespath + file + ".ser";

        srf = new File (file);

        // outsize should be negative before every write
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(-1L);
        rowlock.acquireUninterruptibly();
        
        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);        
        rowlock.release();
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        rowlock.acquireUninterruptibly();
        * 
        */


        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(tableno);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CrashRecovery serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CrashRecovery serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        for (int vl = 0; vl < values.length; vl++)
        {
            try{
              stream.writeUTF(values[vl]);
            }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field values"));
              throw new IOException ();
            }
        }

        // This last flush is valid after each row Uncomment the previous flushes for UltraFastCommits
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CrashRecovery serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        // YOU WILL STILL NEED TO USE THIS MECHANISM EVEN IF YOU DISABLE THE SYSTEM (PC) ON/OFF INTERRUPT
        // USING AN EXTERNAL C PROGRAM
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(0L);

        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        * 
        */
        
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, ClassNotFoundException {
        rowlock.acquireUninterruptibly();
        try{
              tableno = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field tableno"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }

        for (int vl = 0; stream.available() > 0 ; vl++)
        {
            try{
              values[vl] = stream.readUTF();
            }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field values"));
              throw new EOFException ();
            }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field values"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
            }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field values"));
              throw new IOException ();
            }
        }
        rowlock.release();
    }

    protected int SIZEOF(){
        // for every writeUTF two extra bytes of length information are written to file
        // so we must add that length information to the return value of SIZEOF
        int utfextras = values.length * 2; // 8th August 2012
        int vallength = 0;
        for (int val = 0; val < this.values.length; val++)
            vallength += this.values[val].length() * 2;
        return (8 + 8 + vallength + utfextras); // Changed on 8th August 2012
    }
    // although each instance of this Class InstalledTables represents a single row
    // we are committing the entire table here but before that a few checks and balances

    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;

        // we are committing the FileTableDataFileMap so mark those rows with change_flag 'F'
        for (int rw = (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {             
            
            if (Thread.currentThread().getStackTrace()[3].toString().indexOf("ChangeTracker") != -1)
                break;                
           
            System.out.println("CTSR=" + LiveDB.CTSR);
            System.out.println("LiveDB.LiveSharedDB[ct].len=" + LiveDB.LiveSharedDB[ct].length);
            System.out.println(Thread.currentThread().getStackTrace()[1]);
            System.out.println(Thread.currentThread().getStackTrace()[2]);
            System.out.println(Thread.currentThread().getStackTrace()[3]);
            System.out.println(Thread.currentThread().getStackTrace()[4]);
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                fnd = true;
                // insert an entry in ChangeTracker table for this table
                // Basically what this does is to backup this entirerow where you
                // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                thisrow = LiveDB.growMore(ct); // 10th September 2012
                LiveDB.LiveSharedDB[ct][thisrow] =
                new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                LiveDB.LiveSharedDB[ct][rw],thisrow);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which may be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                    + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][crn]).getFilename() + ".ser",
                LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR 
                + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][crn]).getFilename() + ".ser" + ".res");
            
            if ( LiveDB.commit_database(crn) == false )
                return false;
        }
        return true;
    }

    @Override
    public final boolean COMMITINDEXES(){
        if (LiveDB.LiveSharedDB[crn].length == 0)
            return false;
        return LiveDB.write_IndexFile(crn);
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(ft) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
    
    @Override
    public String toString(){
        String filher = "%";
        String vals = null;
        for (int i = 0; i < getValues().length; i++)
        {
            vals += String.valueOf(getValues()[i]);
            if (i < getValues().length -1)
                vals.concat("#");
        }
        return String.valueOf(getTableno()).concat(filher).concat(String.valueOf(getRowid()))
               .concat(vals).concat(filher).concat(String.valueOf(getCrrowid()));
    }

    protected long getTableno(){
        return tableno;
    }
    protected long getRowid(){
        return rowid;  // this is the rowid of the other table not this
    }
    protected String [] getValues(){
        var = values;
        for (int i = 0; i < var.length; i++)
        {
            if (var[i] == null || var[i].equals(""))
               var[i] = new String("");
            else
               var[i] = new String(var[i].toCharArray()).trim();
        }
        return var;
    }
    protected long getCrrowid(){
        return crrowid;
    }
    
    protected void setValues (String [] p_Values){

        if (p_Values == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Values=null"));
            System.exit(1);
        }
        
        rowlock.acquireUninterruptibly();
        // dont validate this it will take very long
 //       for (int i = 0; i < p_Values.length; i++)
            
            //            p_Values[i] = p_Values[i].concat(new String(LiveDB.BlankPads.toCharArray(),0,
//                    LiveDB.MAXSTRINGFLDLEN - (p_Values[i] == null ? 0:p_Values[i].length())));
        
        for (int i = 0; i < p_Values.length; i++)
        {
            values = values == null ? new String[1]: Arrays.copyOf(values, values.length + 1);
            values[i] = new String(p_Values[i].toCharArray());
        }
        rowlock.release();
    }

    protected void setCrrowid(long p_crrow){
        rowlock.acquireUninterruptibly();
        if (p_crrow >= 0)
            crrowid = p_crrow;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_crrow));
            System.exit(1);
        }
        rowlock.release();
    }
}
