/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package MaintainData;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static MaintainData.LiveDB.ft;
import static MaintainData.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.tablenameidentifier
 *
 * Class Name : InstalledTables implements Externalizable, Tables, Cloneable
 *
 * Methods : Look in Class definition
 * There are two constructors here one is without parameters to help simulate a refresh in fsck
 * The other Constructor method public InstalledTables (long p_Tableno, Date p_Accesstime,
 * Date p_Modifiedtime, Date p_installtime, String p_country, String p_iso3lc, long p_Initialsize,
 * long p_Rowid) does a thorough validation before accepting data. If data is non-compliant it does a
 * System.exit(1) rather than a LiveDB.SmoothExitCondition.signalAll() because the object has
 * already been created the constructor is only filling in values into the instance.
 *
 * The Tables Interface has been faked here except for REFRESH, COMMIT, COMMITINDEXES because it is too
 * very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxinstalledTables
 * searchkey = Integer.toString(AnyTableNumber); Any table number is inclusive of fake tables
 * searchkey returns Long rowid in the table
 *
 * Data File : No data entry files to be loaded at Install time.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This is one of the three critical tables for the Application to run. It logs all the
 * installtime of all tables at Application install time and is constantly updated later with accesstime and
 * modifiedtime.
 *
 * Technical References : Buy and Read Asterix and the Black Gold.
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 *
 * Last Modification Date : 3rd June 2012
 * Last Modifications : Made changes to setter setRowid to check for negative value for param p_Rowid
 * Also changed access specifier for default constructor without params from public to protected
 *
 * Last Modification Date : 8th August 2012
 * Last Modifications : As Below
        // every writeUTF writes two extra bytes of length info to the file/stream
        // so we need to add this to the return value of SIZEOF
        int utfextras = 2 * 2;
 * 
 * Last Modification Date : 14th Augustus 2012
 * Last Modification : Through out this source file I have added this call to LiveDB.growMore(ct)
 * whereever I am growing the LiveDB.LiveSharedDB array with a call to new.
 * growMore is a popular enriched bomb grade fertilizer used in the Indian Agricultural sector
 * to spice up the fruit and it contains both Urea and Potash to make Matash. And when you cook
 * vegetable add Harissa (mild Afrikan spice from North Afrika) one ladel.
 * 
 * Last Modification Date: 8th September 2012
 * Last Modifications : MArked as 8th September 2012
 * 
 * Last Modification Date : 10th September 2012
 * Last Modifications : Marked as 10th September 2012
 * Declared a new variable thisrow and its usage
   private transient int thisrow;
 *
 * Last Modification Date : 11th September 2012
 * Last Modifications : removed rowlock.acquireUninterruptibly(); and rowlock.release()
 * statements from the clone method because each of the setters will do the above.
 * 
 * Last Modification Date : 13th September 2012
 * Last Modifications : Marked as 13th September 2012
 * 
 * 
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

final class InstalledTables implements Externalizable, Tables, Cloneable {
    private long tableno;
    private Date accesstime;
    private Date modifiedtime;
    private Date installtime;
    private String whichlocale; // which locale accessed files for tableno Locale.getDefault().getDisplayName()
    private String isocharset; // ISO character set for locale getISO3Language()
    private long initialsize; /* initial size. Size can change for some tables */
    private long outsize; // this is the last serialized size of file on disk will show signed value if
                          // there was a power outage or crash during the serialization process of any file
    private long rowid; /* 9 FIELDS */
    private static final long serialVersionUID = 99995L;
    
    private transient int thisrow; // 10th September 2012
    private transient long fieldchanges;
    private transient File srf; // the file being serialized
    private transient int itn;   // InstalledTables number
    private transient InstalledTables it; // use for cloning
    private transient Semaphore rowlock;

    {
       thisrow = 0;
       fieldchanges = 0L;
       srf = null;
       itn = LiveDB.it;
       it = null;
       rowlock = new Semaphore(1, true);
    }

    // this default constructor is to simulate a refresh
    // look in fsck () dont use otherwise
    public InstalledTables (){

        tableno = 0;
        accesstime = null;
        modifiedtime = null;
        installtime = null;
        whichlocale = " ";
        isocharset = " ";
        initialsize = 0;
        outsize = 0;
        rowid = 0;
    }

    public InstalledTables (long p_Tableno, Date p_Accesstime, Date p_Modifiedtime, Date p_installtime,
           String p_country, String p_iso3lc, long p_Initialsize, long p_Rowid ) {
        rowlock.acquireUninterruptibly();
        if (p_Tableno < 0 || p_Tableno > LiveDB.maxtables)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("InstalledTables constructor Invalid Table No :"
                    + String.valueOf(p_Tableno)));
            System.exit(1);
        }
        tableno = p_Tableno;
        rowlock.release();
        setAccesstime(p_Accesstime);
        setModifiedtime(p_Modifiedtime);
        setInstalltime(p_installtime);        
        // dont validate this we may need to stuff in some other locales forcibly in next release
        setWhichlocale(p_country);        
        setIsocharset(p_iso3lc);
        // ISO 639 is not supposedly stable but it auto converts to new lang code list
        setInitialsize(p_Initialsize);
        setOutsize(0L); // right now it is zero it will be constantly updated during serialization
        setRowid(p_Rowid);
    }

    @Override
    protected Object clone (){
        
        try{
        it = (InstalledTables) super.clone();
        it.setWhichlocale(it.getWhichlocale());        
        it.setIsocharset(it.getIsocharset());
        it.setAccesstime(it.getAccesstime());        
        it.setModifiedtime(it.getModifiedtime());        
        it.setInstalltime(it.getInstalltime());         
        }catch (CloneNotSupportedException ex_NOKIA76){
            throw new InternalError (ex_NOKIA76.toString());
        }
        
        return it;
    }

    @Override
    public void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
        int thisrow = 0;
        String file = null;
        
        // first find the rowid for this table itn = 14 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(itn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("InstalledTables misses entry for self in indxinstalledTables"));
            // perhaps the index is corrupted so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == itn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("InstalledTables misses entry for self in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(itn));
        
        thisrow = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno();
        file = ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][thisrow]).getFilename();
        file = LiveDB.serializedfilespath + file + ".ser";
        
        srf = new File (file);

        // outsize should be negative before every write
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        else
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(-1L);
        rowlock.acquireUninterruptibly();

        thisrow = LiveDB.growMore(ct);

        rowlock.release();
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);        
        rowlock.acquireUninterruptibly();

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(tableno);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getAccesstime() == null ? 0L: getAccesstime().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field accesstime"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getModifiedtime() == null ? 0L: getModifiedtime().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field modifiedtime"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getInstalltime() == null ? 0L: getInstalltime().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field installtime"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(whichlocale);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field whichlocale"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(isocharset);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field isocharset"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(initialsize);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field initialsize"));
              throw new IOException ();
        }

        
        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(outsize);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field outsize"));
              throw new IOException ();
        }
        
        

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        } */

        // outsize should be negative before every write
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }

        
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during InstalledTables serialization"));
            System.exit(1);
        }
        
        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        else
            ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(0L);
                              
        thisrow = LiveDB.growMore(ct);

        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);   
        

    }

    @Override
    public void readExternal (ObjectInput stream) throws IOException, ClassNotFoundException {
        rowlock.acquireUninterruptibly();

        try{
              tableno = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field tableno"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        try{
              accesstime = new Date(stream.readLong());
              if (accesstime.getTime() == new Date(0L).getTime())
                  accesstime = null;
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field accesstime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field accesstime"));
              throw new IOException ();
        }

        try{
              modifiedtime = new Date(stream.readLong());
              if (modifiedtime.getTime() == new Date(0L).getTime())
                  modifiedtime = null;
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field modifiedtime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field modifiedtime"));
              throw new IOException ();
        }

        try{
              installtime = new Date(stream.readLong());
              if (installtime.getTime() == new Date(0L).getTime())
                  installtime = null;
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field installtime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field installtime"));
              throw new IOException ();
        }

        try{
              whichlocale = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field whichlocale"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field whichlocale"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field whichlocale"));
              throw new IOException ();
        }

        try{
              isocharset = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field isocharset"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field isocharset"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field isocharset"));
              throw new IOException ();
        }

        try{
              initialsize = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field initialsize"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field initialsize"));
              throw new IOException ();
        }

        try{
              outsize = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field outsize"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field outsize"));
              throw new IOException ();
        }

        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }

    protected int SIZEOF(){
        // every writeUTF writes two extra bytes of length info to the file/stream
        // so we need to add this to the return value of SIZEOF
        int utfextras = 2 * 2;
        return (8 + 8 + 8 + 8 + (whichlocale.length() * 2) 
                + (isocharset.length() * 2) + 8 + 8 + utfextras ); // Changed 8th August 2012
    }

    protected long getTableno(){
        return tableno;
    }
    protected Date getAccesstime(){
        // 13th September 2012 Checking for null
        return (accesstime == null ? (Date) null : new Date(accesstime.getTime()));
    }
    protected Date getModifiedtime (){
        // 13th September 2012 Checking for null
        return (modifiedtime == null ? (Date)null : new Date(modifiedtime.getTime()));
    }
    protected Date getInstalltime (){
        // 13th September 2012 Checking for null
        return (installtime == null ? (Date) null : new Date(installtime.getTime()));
    }
    protected String getWhichlocale (){
        return new String(whichlocale.toCharArray()).trim();
    }
    protected String getIsocharset (){
        return new String(isocharset.toCharArray()).trim();
    }
    protected long getInitialsize (){
        return initialsize;
    }
    protected long getOutsize (){
        return outsize;
    }
    protected long getRowid(){
        return rowid;
    }

    // this toString is a common override for all tables it does a toString of the current row
    @Override
    public String toString() {
        String filher = "%";
        return String.valueOf(getTableno()).concat(filher)
               .concat(String.valueOf(getAccesstime() == null ? "":getAccesstime().getTime()))
               .concat(filher).concat(String.valueOf(getModifiedtime() == null ? "":getModifiedtime().getTime()))
               .concat(filher).concat(String.valueOf(getInstalltime() == null ? "":getInstalltime().getTime()))
               .concat(filher).concat(getWhichlocale())
               .concat(filher).concat(getIsocharset()).concat(filher)
               .concat(String.valueOf(getInitialsize()))
               .concat(filher).concat(String.valueOf(getOutsize()))
               .concat(filher).concat(String.valueOf(getRowid()));
    }

    protected void setAccesstime(Date p_Accesstime){
        rowlock.acquireUninterruptibly(); 
        // 13th September 2012 Checking for null value in parameter
        accesstime = p_Accesstime == null ? (Date) null : new Date(p_Accesstime.getTime());
        rowlock.release();
    }
    protected void setModifiedtime (Date p_Modifiedtime){
        rowlock.acquireUninterruptibly();
        // 13th September 2012 Checking for null value in parameter
        modifiedtime = p_Modifiedtime == null ? (Date)null : new Date(p_Modifiedtime.getTime());
        rowlock.release();
    }
    protected void setInstalltime (Date p_installtime){
        rowlock.acquireUninterruptibly();
        // 13th September 2012 Checking for null value in parameter
        this.installtime = p_installtime == null ? (Date)null : new Date(p_installtime.getTime());
        rowlock.release();
    }
    protected void setWhichlocale (String p_locale){
        if (p_locale == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("setWhichlocale Invalid p_locale=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        whichlocale = new String(p_locale.toCharArray());
        //        .concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_locale.length()));
        rowlock.release();
    }
    protected void setIsocharset (String p_charset){
        
        rowlock.acquireUninterruptibly();
        if (p_charset == null || p_charset.trim().length() == 0)
            p_charset = "eng";
        isocharset = new String(p_charset.toCharArray());
//.concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_charset.length()));
        rowlock.release();
    }
    protected void setInitialsize(long p_Initialsize){
        rowlock.acquireUninterruptibly(); 
        initialsize = p_Initialsize;
        rowlock.release();
    }
    // This outsize must match the insize after database is loaded otherwise there is an error
    protected void setOutsize (long p_outsize){
        rowlock.acquireUninterruptibly();
        outsize = p_outsize;
        rowlock.release();
    }

    void setRowid(long p_Rowid){
        rowlock.acquireUninterruptibly();
        // 8th September 2012
        // the previous if condition was as below
        // if (LiveDB.LiveSharedDB[itn].length == 0 && p_Rowid != 0)
        // this generates a null pointer exception
        if (LiveDB.LiveSharedDB[itn] == null && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        
        // 10th September 2012
        
        if (LiveDB.LiveSharedDB[itn].length > 1) // Added on 25th June 2012
        {
           thisrow = LiveDB.LiveSharedDB[itn].length -2; 
           if (p_Rowid != ((InstalledTables)LiveDB.LiveSharedDB[itn][thisrow]).getRowid() + 1)
           {
              LiveDB.ALRTLVL = 7;
              System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
              System.exit(1);
           }
        }
        
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }
    
    // although each instance of this Class InstalledTables represents a single row
    // we are committing the entire table here but before that a few checks and balances

    @Override
    public boolean COMMIT(){
        
        boolean fnd = false;
        // we are committing the ChangeTracker table itself so mark those rows with change_flag 'F'
        for (int rw =  (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == itn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                fnd = true;
                // insert an entry in ChangeTracker table for this table
                // Basically what this does is to backup this entirerow where you
                // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                thisrow = LiveDB.growMore(ct);
                LiveDB.LiveSharedDB[ct][thisrow] =
                new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                LiveDB.LiveSharedDB[ct][rw],thisrow);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which may be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                   + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][itn]).getFilename() + ".ser",
                   LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR +
                   ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][itn]).getFilename() + ".ser" + ".res");
            if ( LiveDB.commit_database(itn) == false )
                return false;
        }
        
        return true;
    }

    @Override
    public boolean COMMITINDEXES (){
        if (LiveDB.LiveSharedDB[itn].length == 0)
            return false;
        return LiveDB.write_IndexFile(itn);
    }

    @Override
    public boolean REFRESH(){
        
        if (RefreshDatabase.read_database(itn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
