/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package MaintainData;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static MaintainData.LiveDB.ft;
import static MaintainData.LiveDB.ct;

// caution dont static import pd

/*
 * WIRING DIAGRAM :
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:LiveDB.indxinstalledTables
 * 
 * Class Name : class PictureDatabase implements Externalizable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH, COMMIT because i dont want to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxpictureDatabase
 * searchkey = AnyValidPictureCode
 * searchkey returns Long rowid in the table
 *
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap. The data file
 * will contain all pictures up for sale. A readymade data entry file will be provided to you which you need
 * to customize a bit for your installation. Dont loose that take a physical backup and keep safe.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists all the pictures available for commerce. The constructor goes thru
 * a usual bunch of validations to ensure data compliance. A COMMIT has been provided for this table because
 * you may want to add more pictures to the database at a later stage based on your own logic which you write.
 * NOTE:- Use the slow commit on this table FastCommit Mechanism will not sense this table at all.
 * 
 * Technical References : Buy and Read Asterix and Son.
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 *
 * Last Modification Date : 3rd June 2012
 * Last Modifications : Changed access specifier to protected in default constructor without params
 * Also validated parameter p_Rowid in setter setRowid for negative value before assinging to rowid
 * Also changed some hardcodings for itn = 14 in the initialization blocks to itn = it and pdn = pd
 *
 * Last Modifications : 8th August 2012
 * Last Modifications : As Below
        // Every writeUTF writes two extra bytes of length info to file/stream
        // So we need to add that to the return value of SIZEOF
        int utfextras = 4 * 2;
 * 
 * Last Modification Date : 14th Augustus 2012
 * Last Modification : Through out this source file I have added this call to LiveDB.growMore(ct)
 * whereever I am growing the LiveDB.LiveSharedDB array with a call to new.
 * growMore is a popular enriched bomb grade fertilizer used in the Indian Agricultural sector
 * to spice up the fruit and it contains both Urea and Potash to make Matash. And when you cook
 * vegetable add Harissa (mild Afrikan spice from North Afrika) one ladel along with Bagar.
 * 
 * Last Modifications Date : 8th September 2012
 * Last Modification : Marked as 8th September 2012
 * 
 * Last Modification Date : 10th September 2012
 * Last Modifications : declared a new variable thisrow and its usage
 * 
 * Last Modification Date : 11th September 2012
 * Last Modifications : removed rowlock.acquireUninterruptibly(); and rowlock.release()
 * statements from the clone method because each of the setters will do the above.
 * 
 * Last Modification Date : 13th September 2012
 * Last Modifications : Marked as 13th September 2012
 * 
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class PictureDatabase implements Externalizable, Tables, Cloneable {
    private String pic; /* picture code */
    private Date add_date ; /* Date Added to Database */
    private Date last_access ; /* Last Access Date */
    private String pic_category ; /* Picture Category such as
                                  * 'Expressions', 'Coorg Pavillion'
                                  * 'Sale Items'
                                  */
    private String pic_title ; /* Picture Title */
    private float pic_price ; /* The picture price in USD default */
    private long pic_sale ; /* 0 means not sale item, 1 means sale item */
    private float pic_discount ; /* Picture Discount in usd by default */
    private String pic_img_location ; /* Picture image file location
                                      * Contains Path Name and Actual Image
                                      * file name on disk
                                      */
    private long rowid; /* 10 FIELDS */

    private static final long serialVersionUID = 99991L;
    
    private transient int thisrow;
    private transient Semaphore rowlock;
    private transient int itn;
    private transient int pdn;
    private transient File srf;
    private transient PictureDatabase pd;
    
    {
       thisrow = 0;
       rowlock = new Semaphore(1, true);
       itn = LiveDB.it;
       pdn = LiveDB.pd;
       pd = null;
       srf = null;
    }

    // thou shall not flirt with this constructor -- ten commandments
    public PictureDatabase (){
        pic =  pic_category = pic_title = pic_img_location = null;
        add_date = last_access = null;
        pic_price = pic_discount = pic_sale = 0;
    }

    public PictureDatabase (String p_Pic, Date p_Add_Date, Date p_Last_Access, String p_Pic_Category,
                            String p_Pic_Title, float p_Pic_Price, long p_Pic_Sale, float p_Pic_Discount,
                            String p_Pic_Img_Location, long p_Rowid) {

        rowlock.acquireUninterruptibly();
        if (p_Pic == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Pic=null"));
            System.exit(1);
        }
        pic = new String(p_Pic.toCharArray()) ;
        rowlock.release();
        
        setAdd_date(p_Add_Date);
        setLast_access(p_Last_Access);
        setPic_category(p_Pic_Category);
        setPic_title(p_Pic_Title);
        setPic_price(p_Pic_Price);
        setPic_sale(p_Pic_Sale);
        setPic_discount(p_Pic_Discount);
        setPic_img_location(p_Pic_Img_Location);
        setRowid(p_Rowid);
    }

    @Override
    protected Object clone(){
        
        try{
        pd = (PictureDatabase) super.clone();
        rowlock.acquireUninterruptibly();
        pd.pic = new String(pd.pic.toCharArray());
        rowlock.release();
        setAdd_date(pd.getAdd_date());
        setLast_access(pd.getLast_access());
        setPic_category(pd.getPic_category());
        setPic_title(pd.getPic_title());
        setPic_img_location(pd.getPic_img_location());
        }catch (CloneNotSupportedException ex_NOKIA91){
          throw new InternalError (ex_NOKIA91.toString());
        }
        
        return pd;
    }
    /* Getters and Setters provide a public interface to private data */
    /* They follow the rulebook of encapsulation */
    /* Dont provide Setters for all fields specially key fields !! */
    /* If you have to provide a setter for some sensitive data then
     * make that setter package private or protected.
     */
    protected int SIZEOF(){
        // Every writeUTF writes two extra bytes of length info to file/stream
        // So we need to add that to the return value of SIZEOF
        int utfextras = 4 * 2;
        return ((pic.length() * 2) + 8 + 8 + (pic_category.length() * 2) + (pic_title.length() * 2)
                + 4 + 8 + 4 + (pic_img_location.length() * 2) + 8 + utfextras); // Changed 8th Aug
    }
    protected String getPic() {
        return new String(pic.toCharArray());
    }
    protected Date getAdd_date() {
        return (add_date == null ? (Date) null : new Date(add_date.getTime()));
    }
    protected Date getLast_access() {
        return (last_access == null ? (Date) null : new Date(last_access.getTime()));
    }
    protected String getPic_category(){
        return new String(pic_category.toCharArray());
    }
    protected String getPic_title(){
        return new String(pic_title.toCharArray());
    }
    protected float getPic_price(){
        return pic_price;
    }
    protected long getPic_sale(){
        return pic_sale;
    }
    protected float getPic_discount(){
        return pic_discount;
    }
    protected String getPic_img_location(){
        return new String(pic_img_location.toCharArray());
    }
    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        
        return getPic().concat(filher).concat(String.valueOf(getAdd_date().getTime())).concat(filher)
               .concat(String.valueOf(getLast_access() == null ? "":getLast_access().getTime())).concat(filher).concat(getPic_category())
               .concat(filher).concat(getPic_title()).concat(filher).concat(String.valueOf(getPic_price()))
               .concat(filher).concat(String.valueOf(getPic_sale())).concat(filher)
               .concat(String.valueOf(getPic_sale())).concat(filher).concat(String.valueOf(getPic_discount()))
               .concat(filher).concat(getPic_img_location()).concat(filher).concat(String.valueOf(getRowid()));
    }

    protected void setAdd_date(Date p_Add_date) {
        rowlock.acquireUninterruptibly();
        add_date = p_Add_date == null ? (Date) null : new Date(p_Add_date.getTime());
        rowlock.release();
    }
    protected void setLast_access(Date p_Last_access) {
        rowlock.acquireUninterruptibly();       
        last_access = p_Last_access == null ? (Date)null : new Date(p_Last_access.getTime());
        rowlock.release();
    }
    protected void setPic_category(String p_Pic_category){
        if (p_Pic_category == null)
            p_Pic_category = new String("Miscellaneous") ;

        rowlock.acquireUninterruptibly();
        pic_category = new String(p_Pic_category.toCharArray());
        rowlock.release();
    }
    protected void setPic_title(String p_Pic_title){
        if (p_Pic_title == null)
            p_Pic_title = "No Name";
        rowlock.acquireUninterruptibly();
        pic_title = new String(p_Pic_title.toCharArray());
        rowlock.release();
    }
    protected void setPic_price(float p_Pic_price){
        rowlock.acquireUninterruptibly();
        pic_price = p_Pic_price;
        rowlock.release();
    }
    protected void setPic_sale(long p_Pic_sale){
        rowlock.acquireUninterruptibly();
        pic_sale = p_Pic_sale;
        rowlock.release();
    }
    protected void setPic_discount(float p_Pic_discount){
        rowlock.acquireUninterruptibly();
        pic_discount = p_Pic_discount;
        rowlock.release();
    }
    protected void setPic_img_location(String p_Pic_img_location){
        if (p_Pic_img_location == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Pic_img_location=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        if (new File(LiveDB.convertFileName(p_Pic_img_location)).exists() == false)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Picture Image does not exist:" + p_Pic_img_location));
            System.exit(1);
        }
        
        if(LiveDB.command_line.equals("FileTableDataFileMap"))
        {
            pic_img_location = LiveDB.convertFileName(p_Pic_img_location);
        }
        else
            pic_img_location = new String(p_Pic_img_location.toCharArray());
        
        rowlock.release();
    }

    protected void setRowid(long p_Rowid){
        rowlock.acquireUninterruptibly();
        // 8th September 2012
        // the previous if condition was
        // if (LiveDB.LiveSharedDB[pdn].length == 0 && p_Rowid != 0)
        // which throws a null pointer exception
        if (LiveDB.LiveSharedDB[pdn] == null && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        
        // 8th September 2012 Added this extra check
        if (LiveDB.LiveSharedDB[pdn] != null)
        {
            // 13th September 2012
            if (LiveDB.LiveSharedDB[pdn].length > 1)// Added on 25th June 2012
            {
                thisrow = LiveDB.LiveSharedDB[pdn].length -2;
                if (p_Rowid != ((PictureDatabase)LiveDB.LiveSharedDB[pdn][thisrow]).getRowid() + 1)
                {
                    LiveDB.ALRTLVL = 7;
                    System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
                    System.exit(1);
                }
            }
        }        
        
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
        int thisrow = 0;
        String file = null;

        // first find the rowid for this table pdn in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(pdn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("PictureDatabase misses entry in indxinstalledTables"));
            // perhaps the index is corrupted or has not been updated someplace so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == pdn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("PictureDatabase misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(pdn));

        thisrow = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno();
        file = ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][thisrow]).getFilename();
        file = LiveDB.serializedfilespath + file + ".ser";
        srf = new File (file);

        // outsize should be negative before every write
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(-1L);
        rowlock.acquireUninterruptibly();
        
        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);        
        rowlock.release();
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        rowlock.acquireUninterruptibly();
        * 
        */
        
        try{
              stream.writeUTF(pic);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(add_date.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field add_date"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{              
              stream.writeLong(last_access == null ? 0L:last_access.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field last_access"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(pic_category);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_category"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(pic_title);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_title"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeFloat(pic_price);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_price"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(pic_sale);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_sale"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeFloat(pic_discount);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_discount"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(pic_img_location);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_img_location"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }

        
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PictureDatabase serialization"));
            System.exit(1);
        }
        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(0L);


        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        * 
        */
        
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, ClassNotFoundException {
     // now read in all values here
        rowlock.acquireUninterruptibly();
        try{
              pic = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field pic"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic"));
              throw new IOException ();
        }

        try{
              add_date = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field add_date"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field add_date"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field add_date"));
              throw new IOException ();
        }

        try{
              last_access = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field last_access"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field last_access"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field last_access"));
              throw new IOException ();
        }

        try{
              pic_category = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_category"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_category"));
              throw new IOException ();
        }

        try{
              pic_title = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_title"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_title"));
              throw new IOException ();
        }

        try{
              pic_price = stream.readFloat();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_price"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_price"));
              throw new IOException ();
        }

        try{
              pic_sale = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_sale"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_sale"));
              throw new IOException ();
        }

        try{
              pic_discount = stream.readFloat();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_discount"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_discount"));
              throw new IOException ();
        }

        try{
              pic_img_location = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pic_img_location"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pic_img_location"));
              throw new IOException ();
        }

        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    // This commit has been provided to assist Dynamic additions to the PictureDatabase.
    // You may later have a stock of additional pictures to load.
    @Override
    public final boolean COMMIT(){

        boolean fnd = false;
        // we are committing the ChangeTracker table itself so mark those rows with change_flag 'F'
        for (int rw =  (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == pdn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                fnd = true;
                // insert an entry in ChangeTracker table for this table
                // Basically what this does is to backup this entirerow where you
                // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                thisrow = LiveDB.growMore(ct);
                LiveDB.LiveSharedDB[ct][thisrow] =
                new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length -1);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which may be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][pdn]).getFilename() + ".ser",
                LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR +
                ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][pdn]).getFilename() + ".ser" + ".res");
            if ( LiveDB.commit_database(pdn) == false )
                return false;
        }

        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(pdn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
