/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package MaintainData;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static MaintainData.LiveDB.ft;
import static MaintainData.LiveDB.ct;
// caution dont static import sc


/*
 * WIRING DIAGRAM :
 * @LOOKSY:ChangeTracker.java
 * @LOOKSY:PaymentType.java
 * @LOOKSY:PictureDatabase.java
 * @LOOKSY:CustomerRegister.java
 * @LOOKSY:Spy.dump()
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.CopyFile, LiveDB.commit_database(), LiveDB.LASTLOCKMETHOD
 * @LOOKSY:LiveDB.indxchangeTracker, LiveDB.CTSR, LiveDB.LiveSharedDB, LiveDB.ALRTLVL, LiveDB.qsort()
 *
 * Class Name : class ShoppingCart implements Externalizable, Tables, Cloneable
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been fully implemented here including ROLLBACK and ROLLFORWARD.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indx1shoppingCart
 * searchkey = AnyValidCID + AnyValidPictureCode + Sub_opt_id
 * searchkey returns Long rowid in the table
 * LiveDB.indx2shoppingCart  -- This second index is meant to identify unique CID's IN ShoppingCart table
 * searchkey = AnyValidCID   -- The second index is an ODDBALL index
 * searchkey returns Long rowid in the table
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists all the shopping cart activity. It is the status of billed field
 * which will tell the system if a purchase order needs to be raised. Also if invoicing is required. So
 * every row in this table is to be viewed as an Instance of window shopping by customers until billed = 1.
 * Even if there are further changes to this particular instance of shopping by the customer it will be
 * reflected. For example the rma field will contain an rma number (random generation) the moment customer
 * requests an RMA. Then the refund field also needs to contain a value. The constructor goes thru some
 * quick validations to ensure correct data.
 *
 * Technical References : Buy and Read Asterix the Legionary.
 * Table Dedicated to : Dizzy Gillespie (The Winter in Lisbon)
 *
 * Start Date: 18th October 2010
 * End Date: 15th May 2011
 * Last Modification Date :15th September 2011
 * Last Modifications : The old table schema was as under
 * private String cid;
 * private String pic;
 * private String ponumber;
 * private long opt_id;
 * private long sub_opt_id;
 * private long pmtid;
 * private long pic_qty;
 * private float total;
 * private long billed;
 * private Date shopdate;
 * private long ordernumber;
 * private float orderdiscount;
 * private long rma;
 * private float refund;
 * private long rowid;
 *
 * The new table schema is as under. The field long opt_id has been removed since it is superfluous.
 * Also the field long sub_opt_id has been changed to a String. This field will now contain user
 * choices of sub opts for each of the prod opts seperated by '*'. So it will be a string like
 * "*1*4*5*2*9*2*6*"
 *
 * private String cid;
 * private String pic;
 * private String ponumber;
 * private String sub_opt_id;
 * private long pmtid;
 * private long pic_qty;
 * private float total;
 * private long billed;
 * private Date shopdate;
 * private long ordernumber;
 * private float orderdiscount;
 * private long rma;
 * private float refund;
 * private long rowid;
 *
 * Due to change in schema the constructor also changes and the index LiveDB.indx1shoppingCart
 * also changes. The new Index is as follows
 * searchkey = AnyValidCID + AnyValidPictureCode + Sub_opt_id
 * searchkey returns Long rowid in the table
 *
 * So the same customer may place orders for the same picture with two different set of sub_opt
 * choices. So the index will return a different rowid for each of the sub_opt choices since the
 * underlying table will contain two rows for the same picture code and customer but the sub_opts
 * will be different.
 *
 * Also changes to clone () and SIZEOF ()
 *
 * Last Modification Date : 10th April 2012
 * Last Modifications : Added this extra check before pic_qty must be checked before issuing RMA
 *                      and refunds.
        if (pic_qty <= 0 || pic_qty > LiveDB.MAXMONEYBACK)
            rma = 0;
 *
 * Last Modification Date : 3rd June 2012
 * Last Modifications : Made changes to setter setRowid to validate parameter p_rowid for negative vals
 * before assigning to rowid
 * Also changed some hardcodings in initialization blocks scn = LiveDB.sc, itn = LiveDB.it
 *
 * Last Modification Date : 8th August 2012
 * Last Modifications : As Below
        // Every writeUTF writes two extra bytes of length information to file/stream
        // So we need to add these two extra bytes to the return value of SIZEOF
        int utfextras = 4 * 2;
 * 
 * 
 * Last Modification Date : 15th Augustus 2012
 * Last Modification : Through out this source file I have added this call to LiveDB.growMore(ct)
 * whereever I am growing the LiveDB.LiveSharedDB array with a call to new.
 * growMore is a popular enriched bomb grade fertilizer used in the Indian Agricultural sector
 * to spice up the fruit and it contains both Urea and Potash to make Matash. And when you cook
 * vegetable add Harissa (mild Afrikan spice from North Afrika) one ladel along with Bagar.
 * 
 * Last Modification Date : 8th September 2012
 * Last Modifications : Marked as 8th September 2012
 * 
 * Last Modification Date : 10th September 2012
 * Last Modification : declared new variables thisrow, thatrow and its usage
 * 
 * Last Modification Date : 11th September 2012
 * Last Modifications : removed rowlock.acquireUninterruptibly(); and rowlock.release()
 * statements from the clone method because each of the setters will do the above.
 * 
 * 
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

class ShoppingCart implements Externalizable, Tables, Cloneable {
    private String cid; // customer identification
    private String pic; // picture code
    private String ponumber;  /* pruchase order number */
    private String sub_opt_id; // Sub Option Id for example at website www.artleaves.net GOLD, RED, BLACK etc
    private long pmtid; // valid payment id from PaymentType.java
    private long pic_qty; // order quantity
    private float total; // the total bill
    private long billed; /* Billing Status 0 not yet clicked on check out
                         * 1 requested billing meaning clicked and/or sent purchase order
                         * 2 deleted/cancelled purchase of a single item
                         * 3 For RMA requested, 4 for Printed Invoice
                         * 5 For Raised Invoice to Customer
                         * 6 Customer Clicked Cancelled after submit order
                         * 7 Customer emptied basket
                         * 8 Invoice Cancellation from superuser
                         */
    private Date shopdate; // date of shopping
    private long ordernumber; /* first order 1 and so on ...... */
    private float orderdiscount; /* total discount */
    private long rma; // return material authorization number
    private float refund; // the refund amount after taking away re-stock fee as a percentage
    private long rowid; /* 14 FIELDS */

    private static final long serialVersionUID = 99799L;
    
    private transient int thisrow;
    private transient int thatrow;
    private transient long fieldchanges; // this database can support at best 64 fields for present
    private transient int scn;
    private transient int itn;
    private transient File srf;
    private transient ShoppingCart sc;
    private transient Semaphore rowlock;

    {
       thisrow = 0; 
       thatrow = 0;
       fieldchanges = 0L; // for Ultra Fast Commits later
       scn = LiveDB.sc;
       itn = LiveDB.it;
       srf = null;
       sc = null;
       rowlock = new Semaphore(1, true);
    }

    public ShoppingCart (){
        cid = " ";
        pic = " ";
        ponumber = " ";
        sub_opt_id = " ";
        pmtid = 0;
        pic_qty = 0;
        total = 0;
        billed = 0;
        shopdate = new Date();
        ordernumber = 0;
        orderdiscount = 0.0f;
        rma = 0;
        refund = 0.0f;
        rowid = 0;        
    }
    
    public ShoppingCart (String p_Cid, String p_Pic, String p_PO, String p_Sub_opt_id,
            long p_Pmtid, long p_Pic_qty, float p_Total, long p_Billed, Date p_Shopdate,
            long p_Ordernumber, float p_Orderdiscount, long p_Rma, long p_Rowid )
    {        
        setCid(p_Cid);
        setPic(p_Pic);
        setPonumber(p_PO);        
        setSub_opt_id(p_Sub_opt_id);
        setPmtid(p_Pmtid);
        setPic_qty(p_Pic_qty);
        rowlock.acquireUninterruptibly();
        rma = p_Rma; // return material authorization number
        rowlock.release();
        setTotal(p_Total);
        setBilled(p_Billed);
        setShopdate(p_Shopdate);
        setOrdernumber(p_Ordernumber);
        setOrderdiscount(p_Orderdiscount);
        setRowid(p_Rowid);
        
    }
    @Override
    protected Object clone(){
       
        try{
        sc = (ShoppingCart) super.clone();
        sc.setCid(sc.getCid());
        sc.setPic(sc.getPic());
        sc.setPonumber(sc.getPonumber());
        sc.setSub_opt_id(sc.getSub_opt_id());
        sc.setShopdate(sc.getShopdate());
        }catch (CloneNotSupportedException ex_NOKIA91){          
          throw new InternalError (ex_NOKIA91.toString());
        }
        
        return sc;
    }
    protected int SIZEOF(){
        // Every writeUTF writes two extra bytes of length information to file/stream
        // So we need to add these two extra bytes to the return value of SIZEOF
        int utfextras = 4 * 2;
        return ((cid.length() * 2) + (pic.length() * 2) + (ponumber.length() * 2)
                + (sub_opt_id.length() * 2) + (8 * 2) + 4 + (8 * 3) + 4 
                + 8 + 4 + 8 + utfextras ); // Changed 8th August 2012
    }
    protected String getCid(){
        return new String(cid.toCharArray()).trim();
    }
    protected String getPic(){
        return new String(pic.toCharArray()).trim();
    }
    protected String getPonumber(){
        return new String(ponumber.toCharArray()).trim();
    }

    protected String getSub_opt_id()
    {
        return new String(sub_opt_id.toCharArray()).trim();
    }

    protected long getPmtid()
    {
        return pmtid;
    }
    protected long getPic_qty()
    {
        return pic_qty;
    }

    protected float getTotal()
    {
        return total;
    }
    
    protected long getBilled(){
        return billed;
    }
    protected Date getShopdate(){
        return new Date(shopdate.getTime());
    }
    protected long getOrdernumber(){
        return ordernumber;
    }
    protected float getOrderdiscount(){
        return orderdiscount;
    }
    protected long getRma(){
        return rma;
    }
    protected Float getRefund(){
        return refund;
    }
    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        return getCid().concat(filher).concat(getPic()).concat(filher).concat(getPonumber()).concat(filher)
               .concat(filher).concat(getSub_opt_id())
               .concat(filher).concat(String.valueOf(getPmtid())).concat(filher)
               .concat(String.valueOf(getPic_qty())).concat(filher).concat(String.valueOf(getTotal()))
               .concat(filher).concat(String.valueOf(getBilled())).concat(filher)
               .concat(String.valueOf(getShopdate().getTime())).concat(filher).concat(String.valueOf(getOrdernumber()))
               .concat(filher).concat(String.valueOf(getOrderdiscount())).concat(filher)
               .concat(String.valueOf (getRma())).concat(filher).concat(String.valueOf(getRefund())).concat(filher)
               .concat(String.valueOf(getRowid()));
    }

    private void setCid(String p_Cid){
        rowlock.acquireUninterruptibly();
        if (p_Cid == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Cid=null"));
            System.exit(1);
        }
        if (LiveDB.indxcustomerRegister.get(p_Cid) == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Cid:" + String.valueOf(p_Cid)));
            System.exit(1);
        }
        // String is immutable but the reference is copied and the reference is a Call by Value(local)
        // which will cease to exist after method exits
        this.cid = new String(p_Cid.toCharArray());
//.concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Cid.length()));
        rowlock.release();
    }
    private void setPic(String p_Pic){
        if (p_Pic == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Pic=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        if (LiveDB.indxpictureDatabase.get(p_Pic) == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Pic:" + String.valueOf(p_Pic)));
            System.exit(1);
        }
        // String is immutable but the reference is copied and the reference is a Call by Value(local)
        // which will cease to exist after method exits
        this.pic = new String(p_Pic.toCharArray());
//.concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Pic.length()));
        rowlock.release();
    }
    void setPonumber(String p_Po){
        if (p_Po == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Po=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        this.ponumber = new String(p_Po.toCharArray());
//.concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Po.length()));
        rowlock.release();
    }

    protected void setSub_opt_id (String p_Sub_opt_id){
        if (p_Sub_opt_id == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Sub_opt_id=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        if (p_Sub_opt_id == null) p_Sub_opt_id = " " ;
        sub_opt_id = new String(p_Sub_opt_id.toCharArray());
//.concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Sub_opt_id.length()));
        rowlock.release();
    }
    protected void setPmtid(long p_Pmtid){
        rowlock.acquireUninterruptibly();
        this.pmtid = p_Pmtid;
        rowlock.release();
    }
    protected void setPic_qty(long p_Pic_qty){
        rowlock.acquireUninterruptibly();
        this.pic_qty = p_Pic_qty;
        if (pic_qty <= 0 || pic_qty > LiveDB.MAXMONEYBACK)
            rma = 0;
        rowlock.release();
    }
    protected void setTotal (float p_Total){
        rowlock.acquireUninterruptibly();
        this.total = p_Total;
        if (rma != 0)
           refund = total - (total * LiveDB.RESTOCKPERCENT/100);
        else refund = 0.0F;        
        rowlock.release();
    }
    protected void setBilled (long p_Billed){
        rowlock.acquireUninterruptibly(); 
        this.billed = p_Billed;
        rowlock.release();
    }
    protected void setShopdate (Date p_Shopdate){
        rowlock.acquireUninterruptibly();
        this.shopdate = new Date(p_Shopdate.getTime());
        rowlock.release();
    }
    protected void setOrdernumber (long p_Ordernumber){
        rowlock.acquireUninterruptibly();
        this.ordernumber = p_Ordernumber;
        rowlock.release();
    }
    protected void setOrderdiscount (float p_Orderdiscount){
        rowlock.acquireUninterruptibly(); 
        this.orderdiscount = p_Orderdiscount;
        rowlock.release();
    }

    void setRma (long p_RMA)
    {
        rowlock.acquireUninterruptibly();
        rma = p_RMA;
        rowlock.release();
    }
    
    protected void setRowid (long p_rowid){
        rowlock.acquireUninterruptibly();
        // 8th September 2012
        if (LiveDB.LiveSharedDB[scn] == null && p_rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_rowid)));
            System.exit(1);
        }
        // I am not checking if the rowid is in sequence etc because this table has some
        // ROLLBACK and ROLLFORWARD and it may create problems
        if (p_rowid >= 0)
            rowid = p_rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_rowid));
            System.exit(1);
        }
        rowlock.release();
    }
    // although each instance of this Class ShoppingCart represents a single row
    // we are committing the entire table here but before that a few checks and balances
    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;

        // we are committing the ShoppingCart so mark those rows with change_flag 'F'
        for (int rw = (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                   ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                   fnd = true;
                   // insert an entry in ChangeTracker table for this table
                   // Basically what this does is to backup this entirerow where you
                   // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                   thisrow = LiveDB.growMore(ct);
                   LiveDB.LiveSharedDB[ct][thisrow] =
                   new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                   LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length -1);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which MAY be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                    + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][scn]).getFilename() + ".ser",
                LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR
                + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][scn]).getFilename() + ".ser" + ".res");
            
            if ( LiveDB.commit_database(scn) == false )
                return false;
        }
        
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){

        if (LiveDB.LiveSharedDB[scn].length == 0)
            return false;
        return LiveDB.write_IndexFile(scn);
    }

    @Override
    public final boolean REFRESH(){
        
        if (RefreshDatabase.read_database(scn) == false)
           return false;
        return true;
    }
    protected void acquireRowlock(){
        rowlock.acquireUninterruptibly();
    }

    protected void releaseRowlock(){
        rowlock.release();
    }
    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){

        Long [] Upr = null; // updated rows to rollback

        boolean ok = false;


        if (lastsavepoint == false)
        { // rollback just the last record in table not this rowid !!!
          // there may be a disaster if you RollBack this row
           
           ok = RollBackSingleRow (LiveDB.LiveSharedDB[scn].length -1);
        }
        else
        {  // we have to RollBack until just after the last savepoint which is a fastcommit 'F'
           // so all rowids after the last savepoint

           for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= (int)LiveDB.CTSR; rw--)
           {
               if (LiveDB.pid == ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getPid())
               {
                   if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'R'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'J'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn )
                   {
                       Upr[i] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid();
                       i++;
                   }
                   else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn )
                       break;
               }
           }

           // so all rows affected in this table are in an array
           // so start rolling back each row starting with the highest row number
           LiveDB.qsort (Upr);

           for (int t = Upr.length - 1 ; t >= 0 ; t--)
               ok = RollBackSingleRow (t);
        } // else


        return ok;
    }

    private boolean RollBackSingleRow (int p_whichrow){

           Long rid = 0L;
           Object OldRow = null; // Old value of ShoppingCart Row prior to Update
           

           if (LiveDB.indxchangeTracker.get(Integer.toString(scn) + Integer.toString(p_whichrow))!= null)
              rid = LiveDB.indxchangeTracker.get(Integer.toString(scn) + Integer.toString(p_whichrow));

           // before nullyfying this row check in change tracker if change_flag = 'I'
           if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'I' )
           {
              // new ChangeTracker Entry here and then nullify
              /* lock up CustomerRegister table in fileTableDataFileMap */
              ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setTable_lock ("L");
              // insert an entry in ChangeTracker table for this table
              thisrow = LiveDB.growMore(ct);
              LiveDB.LiveSharedDB[ct][thisrow] =
              new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][scn],
              (long)LiveDB.LiveSharedDB[ct].length -1);
              LiveDB.LASTLOCKMETHOD = "RollBackSingleRow";
              /* process id that locked this table */
              ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setPid (LiveDB.pid);
              // insert an entry in ChangeTracker table for this table
              thisrow = LiveDB.growMore(ct);
              LiveDB.LiveSharedDB[ct][thisrow] =
              new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][scn],
              (long)LiveDB.LiveSharedDB[ct].length -1);

               OldRow = ((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).clone();

               // here we dont acquire the rowlock because we are nullyfying the row
               // we can release the lock after nullyfying row
               LiveDB.LiveSharedDB[scn][p_whichrow] = null;
               // since we are nullyfying the last row in the table insert an entry in ChangeTracker first
               // insert an entry in ChangeTracker table for this table
               thisrow = LiveDB.growMore(ct);
               LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (scn, p_whichrow, 'R', LiveDB.pid, new Date (),
               (ShoppingCart)OldRow, // save this row for a Rollforward
               (long)LiveDB.LiveSharedDB[ct].length -1);

               if (LiveDB.indx1shoppingCart.remove(((ShoppingCart)OldRow).getCid()
                  + ((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getPic()                  
                  + ((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getSub_opt_id()) == null)
               {
                   System.err.println(LiveDB.grassHopper("indx1shoppingCart could not remove"
                   + " for RollBack ShoppingCart Table Rowid:" + String.valueOf(p_whichrow)));
                   Spy.dump();
                   // dont exit not a super sensitive error let it remain in the index it wont harm
               }

               if (LiveDB.indx2shoppingCart.remove(((ShoppingCart)OldRow).getCid()) == null)
               {
                   System.err.println(LiveDB.grassHopper("indx2shoppingCart could not remove"
                   + " for RollBack ShoppingCart Table Rowid:" + String.valueOf(p_whichrow)));
                   Spy.dump();
               }
               
               if (LiveDB.indxchangeTracker.remove(Integer.toString(scn) + Integer.toString(p_whichrow)) == null)
                   System.err.println(LiveDB.grassHopper("indxchangeTracker remove fail "
                   + "for RollBack CustomerRegister Table Rowid:" + p_whichrow));
               // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
               // the previous state of the record that has been RolledBack let us use another flag
               // 'r' (lowercase) to mark this record in ChangeTracker because this original row
               // which was rolled back had ChangeFlag 'I' but it must be ignored by FastCommit.
               // Since this ChangeTracker row on which I am doing a setChange_flag does not
               // contain an EntireRow which is a backup of the previous state prior to RollBack
               // use lowercase flag 'r' instead
               ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
               // insert an entry in ChangeTracker table for this table
               thisrow = LiveDB.growMore(ct);
               LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
               (long)LiveDB.LiveSharedDB[ct].length -1);
               /* unlock CustomerRegister table in fileTableDataFileMap */
               ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setTable_lock ("U");
               // insert an entry in ChangeTracker table for this table
               thisrow = LiveDB.growMore(ct);
               LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ft][scn], (long)LiveDB.LiveSharedDB[ct].length -1);
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'U' )
           {
               // we have to get the last state for this row and restore that
               // so search ChangeTracker for the same tableno and rowid

               for (int rw = (int)LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
               {
                  if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid() == p_whichrow
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid() == rid )

                       OldRow = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getEntirerow();
                  // continue to spin in the loop until you find the last occurence of OldRow
                  // that condition in between has been commented out because if
                  // row re-numbering takes place that could create problems real big so for
                  // now comment that out
               }
               if (OldRow == null)
               {
                   LiveDB.ALRTLVL = 4;
                   System.err.println(LiveDB.grassHopper("could not find OldRow "
                   + "for RollBack ShoppingCart Table Rowid:" + p_whichrow));
                   System.err.println ("LINE " +  ": " + LiveDB.logerrorno + " "
                   + "Dump Files for All Relevant Tables on disk Please Check");
                   // do some dumps here
                   Spy.dump();
               }
               else
               {
                    /* lock up CustomerRegister table in fileTableDataFileMap */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setTable_lock ("L");
                    // insert an entry in ChangeTracker table for this table
                    thisrow = LiveDB.growMore(ct);
                    LiveDB.LiveSharedDB[ct][thisrow] =
                    new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][scn],
                    (long)LiveDB.LiveSharedDB[ct].length -1);
                    LiveDB.LASTLOCKMETHOD = "RollBackSingleRow";
                    /* process id that locked this table */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setPid (LiveDB.pid);
                    // insert an entry in ChangeTracker table for this table
                    thisrow = LiveDB.growMore(ct);
                    LiveDB.LiveSharedDB[ct][thisrow] =
                    new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][scn],
                    (long)LiveDB.LiveSharedDB[ct].length -1);

                    // CHECK DURING BETA TESTING
                    // insert an entry in ChangeTracker table for this table before the actual
                    // dml on the table so as to save the previous row state
                    thisrow = LiveDB.growMore(ct);
                    LiveDB.LiveSharedDB[ct][thisrow] =
                    new ChangeTracker (scn,p_whichrow, 'R', LiveDB.pid, new Date (),
                    (CustomerRegister)OldRow, // OldRow previous state prior to RollBack for Rollforward
                    (long)LiveDB.LiveSharedDB[ct].length -1);

                    ((CustomerRegister)LiveDB.LiveSharedDB[scn][p_whichrow]).acquireRowlock();
                    LiveDB.LiveSharedDB[scn][p_whichrow] = (ShoppingCart)OldRow;
                    ((CustomerRegister)LiveDB.LiveSharedDB[scn][p_whichrow]).releaseRowlock();

                    // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
                    // the previous state of the record that has been RolledBack let us use another flag
                    // 'r' (lowercase) to mark this record in ChangeTracker because this original row
                    // which was rolled back had ChangeFlag 'U' but it must be ignored by FastCommit.
                    // Since this ChangeTracker row on which I am doing a setChange_flag does not
                    // contain an EntireRow which is a backup of the previous state prior to RollBack
                    // use lowercase flag 'r' instead
                    ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
                    // insert an entry in ChangeTracker table for this table
                    thisrow = LiveDB.growMore(ct);
                    LiveDB.LiveSharedDB[ct][thisrow] =
                        new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
                        LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
                        (long)LiveDB.LiveSharedDB[ct].length -1);
                    
                    /* unlock CustomerRegister table in fileTableDataFileMap */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][scn])).setTable_lock ("U");
                    // insert an entry in ChangeTracker table for this table
                    thisrow = LiveDB.growMore(ct);
                    LiveDB.LiveSharedDB[ct][thisrow] =
                    new ChangeTracker (ft,scn, 'U', LiveDB.pid, new Date (),
                    LiveDB.LiveSharedDB[ft][scn], (long)LiveDB.LiveSharedDB[ct].length -1);

                    return true;
               }
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'D' )
           {
               ((CustomerRegister)LiveDB.LiveSharedDB[scn][p_whichrow]).acquireRowlock();
               // save the previous state of the row for a Rollforward
               thisrow = LiveDB.growMore(ct);
               LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (scn,p_whichrow, 'R', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[scn][p_whichrow], // previous state prior to RollBack for Rollforward
               (long)LiveDB.LiveSharedDB[ct].length -1);
               // undelete this row restore the entirerow
               LiveDB.LiveSharedDB[scn][p_whichrow] =
                     ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getEntirerow();
               ((CustomerRegister)LiveDB.LiveSharedDB[scn][p_whichrow]).releaseRowlock();
               // Update Indexes since a RollBack has undeleted the record

               LiveDB.indx1shoppingCart.put(((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getCid()
                            + ((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getPic()                  
                            + ((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getSub_opt_id(),
                            (long)((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getRowid());
                    /* Build a second index just to identify unique cid's in shopping cart duplicates will be overwritten */
                    // for example here you could have used a single dual purpose index but i am making it two for clarity
               LiveDB.indx2shoppingCart.put(((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getCid(),
                            (long)((ShoppingCart)LiveDB.LiveSharedDB[scn][p_whichrow]).getRowid());
               // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
               // the previous state of the record that has been RolledBack let us use another flag
               // 'r' (lowercase) to mark this record in ChangeTracker because this original row
               // which was rolled back had ChangeFlag 'D' but it must be ignored by FastCommit.
               // Since this ChangeTracker row on which I am doing a setChange_flag does not
               // contain an EntireRow which is a backup of the previous state prior to RollBack
               // use lowercase flag 'r' instead
               ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
               // insert an entry in ChangeTracker table for this table
               thisrow = LiveDB.growMore(ct);
               LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
               (long)LiveDB.LiveSharedDB[ct].length -1);
               return true;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'R' )
           {
               // if a RollBack has already been done dont repeat just return true;
               return true;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'F'
                   || ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'C')
           {
               // this has already been fastcommitted so you cant RollBack
               return false;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'J' )
           { // This if is un-necessary i may want to fill it up in next release
               
               // you have just RollForward and now you want a RollBack
               // Who do you think you are bloody bugar request turned down thumbs down
               // Dont do this to me a Jillion + 1 times
               // Even leading databases commerical dont support looneys like you
               // Is this some kind of Army Drill ? Paisa Mungta Hai
               return false;
               // actually it is possible Jillion + 1 times only but I have to find a smarter and faster
               // algorithm to program that darned ChangeTracker table since it will get extremely
               // large and unweildy after 6 months of heavy usage of this application
           }
           else
           {
               if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'K'
                  && ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'T'
                  && ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'r')
               {
                    // if the change_flag is none of the above ChangeTracker is corrupted
                    LiveDB.ALRTLVL = 7;
                    System.err.println(LiveDB.grassHopper("invalid change_flag "
                        + "for RollBack ShoppingCart Table Rowid:" + p_whichrow));
                    System.err.println ("LINE " +  ": " + LiveDB.logerrorno + " "
                        + "Dump Files for All Relevant Tables on disk Please Check");
                    // do some dumps here
                    Spy.dump();
               }
           }

           return false; // if none of the above change_flag return false
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        // find the previous ROLLBACK record and then call ROLLFORWARD with dates
        Date from = null;

        for (int rw = LiveDB.LiveSharedDB[ct].length -1; rw >= LiveDB.CTSR; rw--)
        {
             if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T')
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn )
               // so if there is a commit record then we cant Rollforward any RollBack
              // that happened on this table prior to this commit or a previous Rollforward
                              break;             

             if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn) // this table
             {
                from = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate();
                break;
             }
        }

        if (from ==  null)
            return true;

        return ROLLFORWARD (from, new Date());
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){

        if ((p_fromtime == null && p_upto_time != null) || (p_fromtime != null && p_upto_time == null ))
            return false;

        Long RB [][] = null; // RollBack record numbers in ChangeTracker table for table cr

        for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= LiveDB.CTSR; rw--)
        {
                if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T')
                    && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn )
                    // so if there is a commit record then we cant Rollforward any RollBack
                    // that happened on this table prior to this commit or a previous Rollforward
                              break;
                if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                    && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == scn) // this table
                {
                    if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_upto_time) <= 0
                        && p_upto_time != null) // these rows in ChangeTracker fall within date ranges for crn
                        && (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_fromtime) >= 0
                        && p_fromtime != null))
                    {
                         RB[i][0] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid();
                         RB[i][1] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid();
                         i++;
                    } // if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_upto_time) <= 0
                } // if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
        } // for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= 0; rw--)

        AuxilliaryFunction (RB);
        return true;
    }

    private void AuxilliaryFunction (Long p_RB[][]){
       // This parameter contains all the RollBack records after the last commit/ fast commit has taken
       // place. We will now attempt to Rollforward these RollBack records

       // march backwards thru this array take the oldest RollBack record first and then Rollforward
       // Index p_RB[b][0] will contain ctrowid
       // Index p_RB[b][1] will contain Rowid
       for (int b = p_RB.length -1; b >= 0; b--)
       {
            // apply this entire row now to the table
            thisrow = (int)(long)p_RB[b][1];
            thatrow = (int)(long)p_RB[b][0];
            LiveDB.LiveSharedDB[scn][thisrow] =
            ((ShoppingCart)((ChangeTracker)LiveDB.LiveSharedDB[ct][thatrow]).getEntirerow()).clone();
            // we also have to update the indexes once again over here
            // this Entirerow was placed readymade for a quick rollforward
            // This Entirerow indicates the previous state of the row prior to a RollBack
            // So now that we have restored it in a RollForward we have to get the Rowid
            // for EntireRow and check if the oldchangeflag of the Row was 'I' so when
            // RollBack was done it the index entry would have been removed so now we
            // have to add it again. If oldchangeflag was 'D' and it was Undeleted because
            // of a RollBack then the index entry would have been added during a RollBack
            // Now we need to delete that index entry once again because of a Rollforward.

            thisrow = LiveDB.growMore(ct);
            LiveDB.LiveSharedDB[ct][thisrow] =
               new ChangeTracker (scn,(int)(long)p_RB[b][1], 'J', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[scn][(int)(long)p_RB[b][1]], // this is the Rollforward record
               (long)LiveDB.LiveSharedDB[ct].length -1);
            // Update Indexes
            for (int row = LiveDB.LiveSharedDB[ct].length -1; row >= 0 ; row--)
            {
                 thisrow = (int)(long)p_RB[b][1];
                 if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getRowid()
                     == ((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getRowid())
                 {
                     if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'I'
                         || (LiveDB.ENABLEFASTCOMMIT
                             && ((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'F'
                             && ((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getOldchangeflag() == 'I'))
                         // this second condition is to deal with a quick intermittent bug that may appear
                         // even before we setChange_flag('r') in RollBack the FastCommit mechanism may
                         // have picked the row and fastcommitted it. Should not happen rare occurence
                         // anyways that record will be over-written by the next FastCommit
                     {
                            thisrow = (int)(long)p_RB[b][1];
                            LiveDB.indx1shoppingCart.put(((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getCid()
                            + ((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getPic()                            
                            + ((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getSub_opt_id(),
                            (long)((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getRowid());
                            /* Build a second index just to identify unique cid's in shopping cart duplicates will be overwritten */
                            // for example here you could have used a single dual purpose index but i am making it two for clarity
                            LiveDB.indx2shoppingCart.put(((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getCid(),
                            (long)((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getRowid());
                            break;
                     }
                     else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'D')
                     {
                         thisrow = (int)(long)p_RB[b][1];
                         // so do a remove from the index
                         if (LiveDB.indx1shoppingCart.remove(((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getCid()
                             + ((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getPic()                             
                             + ((ShoppingCart)LiveDB.LiveSharedDB[scn][thisrow]).getSub_opt_id()) == null)
                        {
                            System.err.println(LiveDB.grassHopper("indx1shoppingCart could not remove"
                            + " for RollBack ShoppingCart Table Rowid:" + String.valueOf((int)(long)p_RB[b][1])));
                            Spy.dump();
                            // dont exit not a super sensitive error let it remain in the index it wont harm
                        }

                        if (LiveDB.indx2shoppingCart.remove(((CustomerRegister)LiveDB.LiveSharedDB[scn][thisrow]).getCid()) == null)
                        {
                            System.err.println(LiveDB.grassHopper("indx2shoppingCart could not remove"
                            + " for RollBack ShoppingCart Table Rowid:" + String.valueOf((int)(long)p_RB[b][1])));
                            Spy.dump();
                        }
                        break;
                     }
                 }
            } // for (int row = LiveDB.LiveSharedDB[ct].length -1; row >= LiveDB.CTSR; row--)
       }
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
        int thisrow = 0;
        String file = null;

        // first find the rowid for this table scn = 3 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(scn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("ShoppingCart misses entry in indxinstalledTables"));
            // perhaps the index is corrupted or has not been updated someplace so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == scn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("ShoppingCart misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(scn));

        thisrow = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno();
        file = ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][thisrow]).getFilename();
        file = LiveDB.serializedfilespath + file + ".ser";
        
        srf = new File (file);

        // outsize should be negative before every write
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(-1L);
        rowlock.acquireUninterruptibly();
        
        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);        
        rowlock.release();
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        rowlock.acquireUninterruptibly();
        * 
        */
        
        try{
              stream.writeUTF(cid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field cid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(pic);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pic"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(ponumber);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field ponumber"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(sub_opt_id);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field sub_opt_id"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(pmtid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pmtid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(pic_qty);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pic_qty"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeFloat(total);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field total"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(billed);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field billed"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(shopdate.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field shopdate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(ordernumber);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field ordernumber"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeFloat(orderdiscount);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field orderdiscount"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(rma);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field rma"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeFloat(refund);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field refund"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field rowid"));
              throw new IOException ();
        }

        // this flush is valid if you are using UltraFast Commits uncomment all those flushes above
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ShoppingCart serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        rowlock.release();
        if (srf.exists())
           ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        else ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(0L);
        
        // The following code has been commented out to avoid a java.lang.OutOfMemoryError
        // Its not all that important to backup the InstalledTables row each and every time
        /*
        thisrow = LiveDB.growMore(ct);
        LiveDB.LiveSharedDB[ct][thisrow] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], thisrow);
        * 
        */
        
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, ClassNotFoundException {
        rowlock.acquireUninterruptibly();
        // now read in all values here
        try{
              cid = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field cid"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "ShoppingCart Field cid"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field cid"));
              throw new IOException ();
        }
        try{
              pic = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field pic"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "ShoppingCart Field pic"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pic"));
              throw new IOException ();
        }
        try{
              ponumber = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field ponumber"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "ShoppingCart Field ponumber"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field ponumber"));
              throw new IOException ();
        }

        try{
              sub_opt_id = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field sub_opt_id"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "ShoppingCart Field sub_opt_id"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
        }catch (IOException ex_NOKIA91) {
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field sub_opt_id"));
              throw new IOException ();
        }
        
        try{
              pmtid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field pmtid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pmtid"));
              throw new IOException ();
        }
        try{
              pic_qty = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field pic_qty"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field pic_qty"));
              throw new IOException ();
        }
        try{
              total = stream.readFloat();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field total"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field total"));
              throw new IOException ();
        }
        try{
              billed = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field billed"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field billed"));
              throw new IOException ();
        }
        try{
              shopdate = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field shopdate"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field shopdate"));
              throw new IOException ();
        }
        try{
              ordernumber = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field ordernumber"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field ordernumber"));
              throw new IOException ();
        }
        try{
              orderdiscount = stream.readFloat();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field orderdiscount"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field orderdiscount"));
              throw new IOException ();
        }
        try{
              rma = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field rma"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field rma"));
              throw new IOException ();
        }
        try{
              refund = stream.readFloat();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field refund"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field refund"));
              throw new IOException ();
        }
        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "ShoppingCart Field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ShoppingCart Field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }
}
