/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static DatabaseEngine.LiveDB.ft;
import static DatabaseEngine.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:ChangeTracker.java
 * @LOOKSY:Spy.dump()
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.CopyFile, LiveDB.commit_database(), LiveDB.LASTLOCKMETHOD
 * @LOOKSY:LiveDB.indxchangeTracker, LiveDB.CTSR, LiveDB.LiveSharedDB, LiveDB.ALRTLVL, LiveDB.qsort()
 *
 * Class Name : CustomerRegister implements Externalizable, Tables, Cloneable
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been fully implemented here including ROLLBACK and ROLLFORWARD.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxcustomerRegister
 * searchkey = AnyValidCID
 * searchkey returns Long rowid in the table
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists all the reistered customer details. The constructor does some
 * usual validations to ensure data compliance.
 *
 * Technical References : Buy and Read Asterix and Obelix all out at Sea.
 * Table Dedicated to : Louis ArmStrong (A potrait of Louis ArmStrong)
 *
 * Start Date: 18th October 2010
 * End Date: 15th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

class CustomerRegister implements Externalizable, Tables, Cloneable {
    private String cid;  /* Operating system login name followed by @nodename */
    private String custname;
    private Date custbirthdate;   /* we will email some freebies like */
    private String custbirthcity; /* we sing em a birthday song */
    private String custbirthstate;  /* we mail horrorscopes */
    private String custbirthnation;
    private char custsex ;
    private String custmaritalstatus; /* M - Married, S Single, D divorced,
                                       * W widow, V widower, L living in, E engaged
                                       */
    private Date custlogintime;   /* login time into the system */
    private Date custlogouttime;   /* log out time from system */
    private Date custregisterdate; /* customer registration date */
    private String custaddr1;
    private String custaddr2;
    private String custaddr3;
    private String custaddr4; /* The fourth address line since some addresses in India
                               * are extremely long with associated landmarks
                               * for courier delivery
                               */
    private String landmark; /* for Courier Delivery Boy for example "Near Ghanta Ghar Choron Wali Gali Ignore Postal Address" */
    private String custcity;
    private String custnation;
    private String custzip;
    private String custemail;
    private String custphone;
    private String custfax;
    private char custflag ; /* "O" Alive, "B" Bad, "C" Celebrity,
                              * "T" Top Gun, "S" Stayin Alive
                              * "D" Dead Customers, "F" Possible Felon
                              */
    private Date custlastdate;  /* Last Date you had with me */
    private long rowid;  /* 25 FIELDS */

    private transient long fieldchanges; // for Ultra Fast Commits
    private transient File srf;
    private transient int crn;
    private transient int itn = 14;
    private transient CustomerRegister cr;
    private transient Semaphore rowlock;

    {
       fieldchanges = 0L;
       srf = null;
       crn = LiveDB.cr;
       cr = null;
       rowlock = new Semaphore(1, true);
    }

    public CustomerRegister (String p_Cid, String p_Custname,
            Date p_Custbirthdate, String p_Custbirthcity, String p_Custbirthstate,
            String p_Custbirthnation, char p_Custsex, String p_Custmaritalstatus,
            Date p_Custlogintime, Date p_Custlogouttime,Date p_Custregisterdate, String p_Custaddr1, 
            String p_Custaddr2, String p_Custaddr3, String p_Custaddr4, String p_Landmark, String p_Custcity,
            String p_Custnation, String p_Custzip, String p_Custemail, String p_Custphone, String p_Custfax,
            char p_Custflag, Date p_Custlastdate, long p_Rowid){        
        
        setCid(p_Cid);
        setCustname(p_Custname);        
        setCustbirthdate(p_Custbirthdate);
        setCustbirthcity(p_Custbirthcity);        
        setCustbirthstate(p_Custbirthstate);        
        setCustbirthnation(p_Custbirthnation);        
        setCustsex(p_Custsex);
        setCustmaritalstatus(p_Custmaritalstatus);        
        setCustlogintime(p_Custlogintime);
        setCustlogouttime(p_Custlogouttime);
        setCustregisterdate(p_Custregisterdate);
        setCustaddr1(p_Custaddr1);        
        setCustaddr2(p_Custaddr2);        
        setCustaddr3(p_Custaddr3);        
        setCustaddr4(p_Custaddr4);        
        setLandmark(p_Landmark);        
        setCustcity(p_Custcity);        
        setCustnation(p_Custnation);        
        setCustzip(p_Custzip);        
        setCustemail(p_Custemail);        
        setCustphone(p_Custphone);        
        setCustfax(p_Custfax);        
        setCustflag(p_Custflag);
        setCustlastdate(p_Custlastdate);        
        setRowid(p_Rowid);        
        
    }
    
    // dont use final here at all because it may bypass virtual method invocation
    @Override
    protected Object clone(){
        rowlock.acquireUninterruptibly();
        try{
        cr = (CustomerRegister)super.clone();
        cr.setCid(cr.getCid()); // if u dont do this it will copy reference dodo
        cr.setCustname(cr.getCustname()); //so both original,clone are pointing same place
        cr.setCustbirthcity(cr.getCustbirthcity());
        cr.setCustbirthstate(cr.getCustbirthstate());
        cr.setCustbirthnation(cr.getCustbirthnation());
        cr.setCustmaritalstatus(cr.getCustmaritalstatus());
        cr.setCustaddr1(cr.getCustaddr1());
        cr.setCustaddr2(cr.getCustaddr2());
        cr.setCustaddr3(cr.getCustaddr3());
        cr.setCustaddr4(cr.getCustaddr4());
        cr.setLandmark(cr.getLandmark());
        cr.setCustcity(cr.getCustcity());
        cr.setCustnation(cr.getCustnation());
        cr.setCustzip(cr.getCustzip());
        cr.setCustemail(cr.getCustemail());
        cr.setCustphone(cr.getCustphone());
        cr.setCustfax(cr.getCustfax());
        cr.setCustlogintime(cr.getCustlogintime());
        cr.setCustlogouttime(cr.getCustlogouttime());
        cr.setCustregisterdate(cr.getCustregisterdate());
        cr.setCustbirthdate(cr.getCustbirthdate());
        cr.setCustlastdate(cr.getCustlastdate()); // do the same with dates also
        }catch (CloneNotSupportedException ex_NOKIA99){
            throw new InternalError (ex_NOKIA99.toString()); // catch outside in code wherever you clown
        }
        rowlock.release();
        return cr;
    }

    // dont use final here at all because it may bypass virtual method invocation
    // and report a wrong sizeof
    protected int SIZEOF(){

        // dont use a getter here because it will trim and return
        return ((cid.length() * 2) + (custname.length() * 2) + 8 + (custbirthcity.length() * 2)
        + (custbirthstate.length() * 2) + (custbirthnation.length() * 2) + 2
        + (custmaritalstatus.length() * 2) + 8 + 8 + 8 + (custaddr1.length() * 2)
        + (custaddr2.length() * 2) + (custaddr3.length() * 2) + (custaddr4.length() * 2)
        + (landmark.length() * 2) + (custcity.length() * 2)
        + (custnation.length() * 2) + (custzip.length() * 2) + (custemail.length() * 2)
        + (custphone.length() * 2) + (custfax.length() * 2) + 2 + 8 + 8 );
    }

    // dont use keyword final with getters and setter because the compiler may choose to optimize
    // and bypass virtual method invocation. This should not happen with valid instances but why
    // take chances if in your pgram code you have some redundant references
    protected String getCid(){
        return new String(cid.toCharArray()).trim();
    }
    protected String getCustname(){
        return new String(custname.toCharArray()).trim();
    }

    protected String getUsername(){
        return new String(getCustname().substring(1, getCustname().indexOf('>')).toCharArray()).trim();
    }

    // this gets just the name alone without username
    protected String getName (){
        return new String(getCustname().substring(getCustname().indexOf('>') + 1, getCustname().length()).toCharArray()).trim();
    }
    
    protected Date getCustbirthdate(){
        return new Date(custbirthdate.getTime());
    }
    protected String getCustbirthcity(){
        return new String(custbirthcity.toCharArray()).trim();
    }
    protected String getCustbirthstate(){
        return new String(custbirthstate.toCharArray()).trim();
    }
    protected String getCustbirthnation(){
        return new String(custbirthnation.toCharArray()).trim();
    }
    protected char getCustsex(){
        return custsex;
    }
    protected String getCustmaritalstatus(){
        return new String(custmaritalstatus.toCharArray()).trim();
    }
    protected Date getCustlogintime(){
        return new Date(custlogintime.getTime());
    }
    protected Date getCustlogouttime(){
        return new Date(custlogouttime.getTime());
    }
    protected Date getCustregisterdate(){
        return new Date(custregisterdate.getTime());
    }
    protected String getCustaddr1(){
        return new String(custaddr1.toCharArray()).trim();
    }
    protected String getCustaddr2(){
        return new String(custaddr2.toCharArray()).trim();
    }
    protected String getCustaddr3(){
        return new String(custaddr3.toCharArray()).trim();
    }
    protected String getCustaddr4(){
        return new String(custaddr4.toCharArray()).trim();
    }
    protected String getLandmark(){
        return new String(landmark.toCharArray()).trim();
    }
    protected String getCustcity(){
        return new String(custcity.toCharArray()).trim();
    }
    protected String getCustnation (){
        return new String(custnation.toCharArray()).trim();
    }
    protected String getCustzip (){
        return new String(custzip.toCharArray()).trim();
    }
    protected String getCustemail (){
        return new String(custemail.toCharArray()).trim();
    }
    protected String getCustphone (){
        return new String(custphone.toCharArray()).trim();
    }
    protected String getCustfax (){
        return new String(custfax.toCharArray()).trim();
    }
    protected char getCustflag (){
        return custflag;
    }
    protected Date getCustlastdate (){
        return new Date(custlastdate.getTime());
    }
    protected long getRowid(){
        return rowid;
    }

    // dont make this final because the compiler may attempt optimization and do a direct call
    // instead of virtual method invocation
    @Override
    public String toString(){

        String filher = "%";

        return getCid().concat(filher).concat(getCustname()).concat(filher)
               .concat(String.valueOf(getCustbirthdate().getTime())).concat(filher).concat(getCustbirthcity())
               .concat(filher).concat(getCustbirthstate()).concat(filher).concat(getCustbirthnation())
               .concat(filher).concat(String.valueOf(getCustsex())).concat(filher).concat(getCustmaritalstatus())
               .concat(filher).concat(String.valueOf(getCustlogintime().getTime())).concat(filher)
               .concat(String.valueOf(getCustlogouttime())).concat(filher).concat(String.valueOf(getCustregisterdate()))
               .concat(filher).concat(getCustaddr1()).concat(filher).concat(getCustaddr2()).concat(filher)
               .concat(getCustaddr3()).concat(filher).concat(getCustaddr4()).concat(filher).concat(getLandmark())
               .concat(filher).concat(getCustcity()).concat(filher).concat(getCustnation()).concat(filher)
               .concat(getCustzip()).concat(filher).concat(getCustemail()).concat(filher).concat(getCustphone())
               .concat(filher).concat(getCustfax()).concat(filher).concat(String.valueOf(getCustflag()))
               .concat(filher).concat(String.valueOf(getCustlastdate().getTime())).concat(filher)
               .concat(String.valueOf(getRowid()));
    }

    private void setCid(String p_Cid){

        if (p_Cid == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Cid=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.cid = new String(p_Cid.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Cid.length()));
        rowlock.release();
    }
    
    protected void setCustname(String p_Custname){

        if (p_Custname == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custname=null"));
            System.exit(1);
        }

        String usrname = "<" + LiveDB.USRNAME + ">";
        rowlock.acquireUninterruptibly();
        this.custname =  new String(usrname.toCharArray()).concat(new String(p_Custname.toCharArray())).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - (p_Custname.length() + usrname.length())));
        rowlock.release();
    }
    protected void setCustbirthdate(Date p_Custbirthdate){
        rowlock.acquireUninterruptibly();
        this.custbirthdate = new Date(p_Custbirthdate.getTime());
        rowlock.release();
    }
    protected void setCustbirthcity(String p_Custbirthcity){

        if (p_Custbirthcity == null)
            p_Custbirthcity = " " ;

        rowlock.acquireUninterruptibly();
        this.custbirthcity = new String(p_Custbirthcity.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custbirthcity.length()));
        rowlock.release();
    }
    protected void setCustbirthstate(String p_Custbirthstate){

        if (p_Custbirthstate == null)
            p_Custbirthstate = " ";

        rowlock.acquireUninterruptibly();
        this.custbirthstate = new String(p_Custbirthstate.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custbirthstate.length()));
        rowlock.release();
    }
    protected void setCustbirthnation(String p_Custbirthnation){

        if (p_Custbirthnation == null)
            p_Custbirthnation = " ";

        rowlock.acquireUninterruptibly();
        this.custbirthnation = new String(p_Custbirthnation.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custbirthnation.length()));
        rowlock.release();
    }
    protected void setCustsex(char p_Custsex){
        rowlock.acquireUninterruptibly();
        this.custsex = p_Custsex;
        rowlock.release();
     }
    protected void setCustmaritalstatus(String p_Custmaritalstatus){

        if (p_Custmaritalstatus == null)
            p_Custmaritalstatus = "S";

        rowlock.acquireUninterruptibly();
        this.custmaritalstatus = new String(p_Custmaritalstatus.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custmaritalstatus.length()));
        rowlock.release();
    }
    protected void setCustlogintime(Date p_Custlogintime){
        rowlock.acquireUninterruptibly();
        if (p_Custlogintime.after(getCustlogouttime()) && p_Custlogintime != null && getCustlogouttime() != null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Custlogintime after Custlogouttime"));
            System.exit(1);
        }
        if (p_Custlogintime == null)
            custlogintime = null;
        
        this.custlogintime = new Date(p_Custlogintime.getTime());
        rowlock.release();
    }
    protected void setCustlogouttime(Date p_Custlogouttime){
        rowlock.acquireUninterruptibly();
        if (p_Custlogouttime.before(getCustlogintime()) && p_Custlogouttime != null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Custlogouttime before Custlogintime"));
            System.exit(1);
        }
                
        if (p_Custlogouttime == null)
            custlogouttime = null;
        else
            custlogouttime = new Date(p_Custlogouttime.getTime());
        rowlock.release();
    }
    protected void setCustregisterdate(Date p_Custregisterdate){
        rowlock.acquireUninterruptibly();
        if (p_Custregisterdate == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Custregisterdate null"));
            System.exit(1);
        }
        if (p_Custregisterdate.before(getCustbirthdate()) && getCustbirthdate() != null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Custregisterdate before he was born"));
            System.exit(1);           
        }
        
        this.custregisterdate = new Date(p_Custregisterdate.getTime());
        rowlock.release();
    }
    protected void setCustaddr1(String p_Custaddr1){

        if (p_Custaddr1 == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custaddr1=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custaddr1 = new String(p_Custaddr1.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custaddr1.length()));
        rowlock.release();
    }
    protected void setCustaddr2(String p_Custaddr2){

        if (p_Custaddr2 == null)
            p_Custaddr2 = " ";

        rowlock.acquireUninterruptibly();
        this.custaddr2 = new String(p_Custaddr2.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custaddr2.length()));
        rowlock.release();
    }
    protected void setCustaddr3(String p_Custaddr3){

        if (p_Custaddr3 == null)
            p_Custaddr3 = " ";

        rowlock.acquireUninterruptibly();
        this.custaddr3 = new String(p_Custaddr3.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custaddr3.length()));
        rowlock.release();
    }
    protected void setCustaddr4(String p_Custaddr4){

        if (p_Custaddr4 == null)
            p_Custaddr4 = " ";
        
        rowlock.acquireUninterruptibly();
        this.custaddr4 = new String(p_Custaddr4.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custaddr4.length()));
        rowlock.release();
    }
    protected void setLandmark( String p_Landmark ){

        if (p_Landmark == null)
            p_Landmark = " ";

        rowlock.acquireUninterruptibly();
        this.landmark = new String(p_Landmark.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Landmark.length()));
        rowlock.release();
    }
    protected void setCustcity(String p_Custcity){

        if (p_Custcity == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custcity=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custcity = new String(p_Custcity.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custcity.length()));
        rowlock.release();
    }
    protected void setCustnation (String p_Custnation){

        if (p_Custnation == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custnation=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custnation = new String(p_Custnation.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custnation.length()));
        rowlock.release();
    }
    protected void setCustzip (String p_Custzip){

        if (p_Custzip == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custzip=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custzip = new String(p_Custzip.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custzip.length()));
        rowlock.release();
    }
    protected void setCustemail (String p_Custemail){

        if (p_Custemail == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custemail=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custemail = new String(p_Custemail.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custemail.length()));
        rowlock.release();
    }
    protected void setCustphone (String p_Custphone){

        if (p_Custphone == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Custphone=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();
        this.custphone = new String(p_Custphone.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custphone.length()));
        rowlock.release();
    }
    protected void setCustfax (String p_Custfax){

        if (p_Custfax == null)
            p_Custfax = " ";

        rowlock.acquireUninterruptibly();
        this.custfax = new String(p_Custfax.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Custfax.length()));
        rowlock.release();
    }
    protected void setCustflag (char p_Custflag){
        rowlock.acquireUninterruptibly();
        /* "O" Alive, "B" Bad, "C" Celebrity,
         * "T" Top Gun, "S" Stayin Alive
         * "D" Dead Customers
         */
        if (p_Custflag != 'O' && p_Custflag != 'B' && p_Custflag != 'C' && p_Custflag != 'T'
            && p_Custflag != 'S' && p_Custflag != 'D')
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Custflag:" + p_Custflag));
            System.exit(1);
        }
        
        this.custflag = p_Custflag ;
        rowlock.release();
    }
    protected void setCustlastdate (Date p_Custlastdate){
        rowlock.acquireUninterruptibly();
        if (p_Custlastdate.before(getCustregisterdate()))
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Custlastdate before Custregisterdate"));
            System.exit(1);
        }
        
        this.custlastdate = new Date(p_Custlastdate.getTime());
        rowlock.release();
    }
    protected void setRowid (long p_rowid){
        rowlock.acquireUninterruptibly();        
        if (LiveDB.LiveSharedDB[crn].length == 0 && p_rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Args p_Rowid " + String.valueOf(p_rowid)));
            System.exit(1);
        }     
        rowid = p_rowid;
        if (p_rowid >= 0)
            rowid = p_rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_rowid));
            System.exit(1);
        }
        rowlock.release();
    }

    // although each instance of this Class CustomerRegister represents a single row
    // we are committing the entire table here but before that a few checks and balances
    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;
 
        // we are committing the CustomerRegister so mark those rows with change_flag 'F'
        for (int rw = (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                fnd = true;
                // insert an entry in ChangeTracker table for this table
                // Basically what this does is to backup this entirerow where you
                // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which may be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            // So do a blind copy
            LiveDB.CopyFile(LiveDB.serializedfilespath + 
                    ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][crn]).getFilename() + ".ser",
                    LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR
                    + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][crn]).getFilename() + ".ser" + ".res");
            if ( LiveDB.commit_database(crn) == false )
                return false;
        }
        
        return true;
    }

    // final because you allow it to be bypassed by the virtual method invocation
    @Override
    public final boolean COMMITINDEXES(){
        
        if (LiveDB.LiveSharedDB[crn].length == 0)
            return false;
        return LiveDB.write_IndexFile(crn);
    }

    // final because you allow it to be bypassed by the virtual method invocation
    @Override
    public final boolean REFRESH(){
        
        if (RefreshDatabase.read_database(crn) == false)
           return false;
        return true;
    }

    // dont use keyword final here
    protected void acquireRowlock(){
        rowlock.acquireUninterruptibly();
    }

    // dont use keyword final here
    protected void releaseRowlock(){
        rowlock.release();
    }
    
    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        
        Long [] Upr = null; // updated rows to rollback
        
        boolean ok = false;

        
        if (lastsavepoint == false)  
        { // rollback just the last record in table not this rowid !!!
          // there may be a disaster if you RollBack this row
           
           ok = RollBackSingleRow (LiveDB.LiveSharedDB[crn].length -1);
        }
        else
        {  // we have to RollBack until just after the last savepoint which is a fastcommit 'F'
           // so all rowids after the last savepoint 

           
           for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= (int)LiveDB.CTSR; rw--)
           {
               if (LiveDB.pid == ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getPid())
               {
                   if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'R'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'J'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn )
                   {
                       Upr[i] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid();
                       i++;
                   }
                   else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn )
                       break;
               }
           }
           
           // so all rows affected in this table are in an array
           // so start rolling back each row starting with the highest row number
           LiveDB.qsort (Upr);

           for (int t = Upr.length - 1 ; t >= 0 ; t--)
               ok = RollBackSingleRow (t);           
        } // else

        
        return ok;
    }

    private boolean RollBackSingleRow (int p_whichrow){
           
           Long rid = 0L;
           Object OldRow = null; // Old value of CustomerRegister Row prior to Update/Insert/....
           

           if (LiveDB.indxchangeTracker.get(Integer.toString(crn) + Integer.toString(p_whichrow))!= null)
              rid = LiveDB.indxchangeTracker.get(Integer.toString(crn) + Integer.toString(p_whichrow));

           // before nullyfying this row check in ChangeTracker if change_flag = 'I'
           if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'I' )
           {
              // new ChangeTracker Entry here and then nullify
              /* lock up CustomerRegister table in fileTableDataFileMap */
              ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setTable_lock ("L");
              // insert an entry in ChangeTracker table for this table
              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
              new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][crn],
              (long)LiveDB.LiveSharedDB[ct].length);
              LiveDB.LASTLOCKMETHOD = "RollBackSingleRow";
              /* process id that locked this table */
              ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setPid (LiveDB.pid);
              // insert an entry in ChangeTracker table for this table
              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
              new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][crn],
              (long)LiveDB.LiveSharedDB[ct].length);

              OldRow = ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).clone();

               // here we dont acquire the rowlock because we are nullyfying the row
               // we cant release the lock after nullyfying row
               LiveDB.LiveSharedDB[crn][p_whichrow] = null;
               // since we are nullyfying the last row in the table insert an entry in ChangeTracker first
               // insert an entry in ChangeTracker table for this table
               LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (crn, p_whichrow, 'R', LiveDB.pid, new Date (),
               OldRow, // save this row for a Rollforward
               (long)LiveDB.LiveSharedDB[ct].length);

               if (LiveDB.indxcustomerRegister.remove(((CustomerRegister)OldRow).getCid()) == null)                   
                  System.err.println(LiveDB.grassHopper("indxcustomerRegister remove fail "
                  + "for RollBack CustomerRegister Table Rowid:" + p_whichrow));

               if (LiveDB.indxchangeTracker.remove(Integer.toString(crn) + Integer.toString(p_whichrow)) == null)
                   System.err.println(LiveDB.grassHopper("indxchangeTracker remove fail "
                   + "for RollBack CustomerRegister Table Rowid:" + String.valueOf(p_whichrow)));

               // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
               // the previous state of the record that has been RolledBack let us use another flag
               // 'r' (lowercase) to mark this record in ChangeTracker because this original row
               // which was rolled back had ChangeFlag 'I' but it must be ignored by FastCommit.
               // Since this ChangeTracker row on which I am doing a setChange_flag does not
               // contain an EntireRow which is a backup of the previous state prior to RollBack
               // use lowercase flag 'r' instead
               ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
               // insert an entry in ChangeTracker table for this table
               LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
               (long)LiveDB.LiveSharedDB[ct].length);

               /* unlock CustomerRegister table in fileTableDataFileMap */
               ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setTable_lock ("U");
               // insert an entry in ChangeTracker table for this table
               LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ft][crn], (long)LiveDB.LiveSharedDB[ct].length);
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'U' )
           {
               // we have to get the last state for this row and restore that
               // so search ChangeTracker for the same tableno and rowid

               for (int rw = (int)LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
               {
                  if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid() == p_whichrow
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid() == rid )

                       OldRow = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getEntirerow();
               }
               if (OldRow == null)
               {
                   LiveDB.ALRTLVL = 4;
                   System.err.println(LiveDB.grassHopper("could not find OldRow "
                   + "for RollBack CustomerRegister Table Rowid:" + String.valueOf(p_whichrow)));
                   System.err.println ("LINE " +  ": " + String.valueOf(LiveDB.logerrorno) + " "
                   + "Dump Files for All Relevant Tables on disk Please Check");
                   // do some dumps here
                   Spy.dump();
               }
               else
               {
                    /* lock up CustomerRegister table in fileTableDataFileMap */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setTable_lock ("L");
                    // insert an entry in ChangeTracker table for this table
                    LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                    new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][crn],
                    (long)LiveDB.LiveSharedDB[ct].length);
                    LiveDB.LASTLOCKMETHOD = "RollBackSingleRow";
                    /* process id that locked this table */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setPid (LiveDB.pid);
                    // insert an entry in ChangeTracker table for this table
                    LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                    new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (), LiveDB.LiveSharedDB[ft][crn],
                    (long)LiveDB.LiveSharedDB[ct].length);

                    // insert an entry in ChangeTracker table for this table before the actual
                    // dml on the table so as to save the previous row state
                    LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                    new ChangeTracker (crn,p_whichrow, 'R', LiveDB.pid, new Date (),
                    (CustomerRegister)OldRow, // OldRow previous state prior to RollBack for Rollforward
                    (long)LiveDB.LiveSharedDB[ct].length);

                    ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).acquireRowlock();
                    LiveDB.LiveSharedDB[crn][p_whichrow] = (CustomerRegister)OldRow;
                    ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).releaseRowlock();

                    // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
                    // the previous state of the record that has been RolledBack let us use another flag
                    // 'r' (lowercase) to mark this record in ChangeTracker because this original row
                    // which was rolled back had ChangeFlag 'U' but it must be ignored by FastCommit.
                    // Since this ChangeTracker row on which I am doing a setChange_flag does not
                    // contain an EntireRow which is a backup of the previous state prior to RollBack
                    // use lowercase flag 'r' instead
                    ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
                    // insert an entry in ChangeTracker table for this table
                    LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                    new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
                    LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
                    (long)LiveDB.LiveSharedDB[ct].length);
               
                    /* unlock CustomerRegister table in fileTableDataFileMap */
                    ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][crn])).setTable_lock ("U");
                    // insert an entry in ChangeTracker table for this table
                    LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                    new ChangeTracker (ft,crn, 'U', LiveDB.pid, new Date (),
                    LiveDB.LiveSharedDB[ft][crn], (long)LiveDB.LiveSharedDB[ct].length);
                    
                    return true;
               }
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'D' )
           {
               ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).acquireRowlock();
               // save the previous state of the row for a Rollforward
               LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (crn,p_whichrow, 'R', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[crn][p_whichrow], // previous state prior to RollBack for Rollforward
               (long)LiveDB.LiveSharedDB[ct].length);
               // undelete this row restore the entirerow
               LiveDB.LiveSharedDB[crn][p_whichrow] =
                     ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getEntirerow();
               ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).releaseRowlock();
               // Update Indexes since a RollBack has undeleted the record
               LiveDB.indxcustomerRegister.put(((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).getCid(),
                       ((CustomerRegister)LiveDB.LiveSharedDB[crn][p_whichrow]).getRowid());
               // Since any row in ChangeTracker with ChangeFlag 'R' must have a Entirerow which is
               // the previous state of the record that has been RolledBack let us use another flag
               // 'r' (lowercase) to mark this record in ChangeTracker because this original row
               // which was rolled back had ChangeFlag 'D' but it must be ignored by FastCommit.
               // Since this ChangeTracker row on which I am doing a setChange_flag does not
               // contain an EntireRow which is a backup of the previous state prior to RollBack
               // use lowercase flag 'r' instead
               ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).setChange_flag('r');
               // insert an entry in ChangeTracker table for this table
               LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (ct, rid, 'U', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[ct][(int)(long)rid], // backup this row
               (long)LiveDB.LiveSharedDB[ct].length);
               return true;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'R' )
           {
               // if a RollBack has already been done dont repeat just return true;
               return true;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'F'
                   || ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'C')
           {
               // this has already been fastcommitted so you cant RollBack
               return false;
           }
           else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() == 'J' )
           {
               // you have just RollForward and now you want a RollBack
               // Who do you think you are bloody bugar request turned down thumbs down
               // Dont do this to me a Jillion + 1 times
               // Even leading databases commerical dont support looneys like you
               // Is this some kind of Army Drill ? Paisa Mungta Hai
               return false;
               // actually it is possible Jillion + 1 times only but I have to find a smarter and faster
               // algorithm to program that darned ChangeTracker table since it will get extremely
               // large and unweildy after 6 months of heavy usage of this application 
           }
           else
           {
               if (((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'K'
                  && ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'T'
                  && ((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)rid]).getChange_flag() != 'r')
               {
                    // if the change_flag is none of the above ChangeTracker is corrupted
                    LiveDB.ALRTLVL = 7;
                    System.err.println(LiveDB.grassHopper("invalid change_flag "
                        + "for RollBack CustomerRegister Table Rowid:" + String.valueOf(p_whichrow)));
                    System.err.println ("LINE " +  ": " + String.valueOf(LiveDB.logerrorno) + " "
                        + "Dump Files for All Relevant Tables on disk Please Check");
                    // do some dumps here
                    Spy.dump();
               }
           }

           return false; // if none of the above change_flag return false
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    // A Rollforward is just doing a RollBack on the previous RollBack so it is equal to cancelling
    // the previous RollBack
    @Override
    public final boolean ROLLFORWARD (){
        // find the previous ROLLBACK record and then call ROLLFORWARD with dates
        Date from = null;
                
        for (int rw = LiveDB.LiveSharedDB[ct].length -1; rw >= LiveDB.CTSR; rw--)
        {
             if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T')
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn )
               // so if there is a commit record then we cant Rollforward any RollBack
              // that happened on this table prior to this commit or a previous Rollforward
                              break; 

             if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn) // this table
             {
                from = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate();
                break;
             }
        } 
        
        if (from ==  null)
            return true;

        return ROLLFORWARD (from, new Date()); 
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
         
        if ((p_fromtime == null && p_upto_time != null) || (p_fromtime != null && p_upto_time == null ))
            return false;
                
        Long RB [][] = null; // RollBack record numbers in ChangeTracker table for table cr
        
        for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= LiveDB.CTSR; rw--)
        {
                if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'F'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'C'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'J'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'K'
                    || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'T')
                    && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn )
                    // so if there is a commit record then we cant Rollforward any RollBack
                    // that happened on this table prior to this commit or a previous Rollforward
                              break;
                if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
                    && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == crn) // this table
                {
                    if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_upto_time) <= 0
                        && p_upto_time != null) // these rows in ChangeTracker fall within date ranges for crn
                        && (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_fromtime) >= 0
                        && p_fromtime != null))
                    {
                         RB[i][0] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getCtrowid();
                         RB[i][1] = ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid();
                         i++;
                    } // if ((((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().compareTo(p_upto_time) <= 0
                } // if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'R'
        } // for (int rw = LiveDB.LiveSharedDB[ct].length -1, i = 0; rw >= 0; rw--)

        AuxilliaryFunction (RB);
        return true;
    }

    private void AuxilliaryFunction (Long p_RB[][]){
       
       // This parameter contains all the RollBack records after the last commit/ fast commit has taken
       // place. We will now attempt to Rollforward these RollBack records

       // march backwards thru this array take the oldest RollBack record first and then Rollforward
       // Index p_RB[b][0] will contain ctrowid
       // Index p_RB[b][1] will contain Rowid
       for (int b = p_RB.length -1; b >= 0; b--)
       {
            // apply this entire row now to the table
            LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]] =
            ((CustomerRegister)((ChangeTracker)LiveDB.LiveSharedDB[ct][(int)(long)p_RB[b][0]]).getEntirerow()).clone();
            // we also have to update the indexes once again over here
            // this Entirerow was placed readymade during a RollBack for a quick rollforward
            // This Entirerow indicates the previous state of the row prior to a RollBack
            // So now that we have restored it in a RollForward we have to get the Rowid
            // for EntireRow and check if Changeflag of the Row was 'I' so when
            // RollBack was done the index entry would have been removed so now we
            // have to add it again. If Changeflag was 'D' and it was Undeleted because
            // of a RollBack then the index entry would have been added during a RollBack
            // Now we need to delete that index entry once again because of a Rollforward.
            // we will only check the Changeflag and not the oldchangeflag because both
            // RollBack and Rollforward are dealing only with rows that have not yet been
            // 'C' committed or 'F' Fast committed.
            
            LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
               new ChangeTracker (crn,(int)(long)p_RB[b][1], 'J', LiveDB.pid, new Date (),
               LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]], // this is the RolledForward restored record
               (long)LiveDB.LiveSharedDB[ct].length);

            // Update Indexes
            for (int row = LiveDB.LiveSharedDB[ct].length -1; row >= 0 ; row--)
            {
                 if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getRowid()
                     == ((CustomerRegister)LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]]).getRowid())
                 {
                     if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'I'
                         || (LiveDB.ENABLEFASTCOMMIT 
                             && ((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'F'
                             && ((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getOldchangeflag() == 'I'))
                         // this second condition is to deal with a quick intermittent bug that may appear
                         // even before we setChange_flag('r') in RollBack the FastCommit mechanism may
                         // have picked the row and fastcommitted it. Should not happen rare occurence
                         // anyways that record will be over-written by the next FastCommit
                     {
                        // so do a put into the index
                        LiveDB.indxcustomerRegister.put(((CustomerRegister)LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]]).getCid(),
                        ((CustomerRegister)LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]]).getRowid());
                        break;
                     }
                     else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][row]).getChange_flag() == 'D')
                     {
                         // so do a remove from the index
                         if (LiveDB.indxcustomerRegister.remove(((CustomerRegister)LiveDB.LiveSharedDB[crn][(int)(long)p_RB[b][1]]).getCid()) == null)
                             System.err.println(LiveDB.grassHopper("indxcustomerRegister remove fail "
                                   + "for RollBack CustomerRegister Table Rowid:" + (int)(long)p_RB[b][1]));
                         break;
                     }
                 }
            } // for (int row = LiveDB.LiveSharedDB[ct].length -1; row >= LiveDB.CTSR; row--)
       }      
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
 
        // first find the rowid for this table itn = 14 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(crn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("CustomerRegister misses entry in indxinstalledTables"));
            // perhaps the index is corrupted or has not been updated someplace so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == crn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("CustomerRegister misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(crn));

        srf = new File (((FileTableDataFileMap)LiveDB.LiveSharedDB[ft]
                [(int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno()]).getFilename());

        // outsize should be negative before every write
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);


        try{
              stream.writeUTF(cid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field cid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(custname);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custname"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{

              stream.writeLong(custbirthdate.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custbirthdate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custbirthcity);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custbirthcity"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custbirthstate);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custbirthstate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custbirthnation);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custbirthnation"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeChar(custsex);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custsex"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custmaritalstatus);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custmaritalstatus"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(custlogintime.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlogintime"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(custlogouttime.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlogouttime"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(custregisterdate.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custregisterdate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custaddr1);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr1"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custaddr2);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr2"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custaddr3);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr3"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custaddr4);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr4"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(landmark);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field landmark"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custcity);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custcity"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custnation);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custnation"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custzip);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custzip"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custemail);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custemail"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custphone);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custphone"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(custfax);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custfax"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeChar(custflag);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custflag"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(custlastdate.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlastdate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }


        // This last flush is valid after each row Uncomment the above flushes for UltraFastCommits
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during CustomerRegister serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
        rowlock.release();
    }
                                    // what method throws is not part of signature
    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {
        
        rowlock.acquireUninterruptibly();
        // now read in all values here
        try{
              cid = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field cid"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field cid"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field cid"));
              throw new IOException ();
        }

        try{
              custname = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custname"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custname"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custname"));
              throw new IOException ();
        }
        
        try{
              custbirthdate = new Date(stream.readLong());
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custbirthdate"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custbirthdate"));
              throw new IOException ();
        }

        try{
              custbirthcity = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "custbirthcity"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "custbirthcity"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "custbirthcity"));
              throw new IOException ();
        }

        try{
              custbirthstate = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "custbirthstate"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "custbirthstate"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "custbirthstate"));
              throw new IOException ();
        }

        try{
              custbirthnation = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "custbirthnation"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "custbirthnation"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "custbirthnation"));
              throw new IOException ();
        }
 
        try{
              custsex = stream.readChar();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custsex"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custsex"));
              throw new IOException ();
        }

        try{
              custmaritalstatus = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custmaritalstatus"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custmaritalstatus"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custmaritalstatus"));
              throw new IOException ();
        }

        try{
              custlogintime = new Date(stream.readLong());
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custlogintime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlogintime"));
              throw new IOException ();
        }

        try{
              custlogouttime = new Date(stream.readLong());
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custlogouttime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlogouttime"));
              throw new IOException ();
        }

        try{
              custregisterdate = new Date(stream.readLong());
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custregisterdate"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custregisterdate"));
              throw new IOException ();
        }
        
        try{
              custaddr1 = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custaddr1"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custaddr1"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr1"));
              throw new IOException ();
        }

        try{
              custaddr2 = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custaddr2"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custaddr2"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr2"));
              throw new IOException ();
        }

        try{
              custaddr3 = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custaddr3"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custaddr3"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr3"));
              throw new IOException ();
        }
 
        try{
              custaddr4 = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custaddr4"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custaddr4"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custaddr4"));
              throw new IOException ();
        }

        try{
              landmark = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "landmark"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "landmark"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "landmark"));
              throw new IOException ();
        }

        try{
              custcity = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custcity"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custcity"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custcity"));
              throw new IOException ();
        }

        try{
              custnation = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custnation"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custnation"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custnation"));
              throw new IOException ();
        }

        try{
              custzip = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custzip"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custzip"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91,"field custzip"));
              throw new IOException ();
        }

        try{
              custemail = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custemail"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custemail"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custemail"));
              throw new IOException ();
        }

        try{
              custphone = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custphone"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custphone"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custphone"));
              throw new IOException ();
        }

        try{
              custfax = stream.readUTF();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custfax"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field custfax"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custfax"));
              throw new IOException ();
        }

        try{
              custflag = stream.readChar();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custflag"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custflag"));
              throw new IOException ();
        }

        try{
              custlastdate = new Date(stream.readLong());
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field custlastdate"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field custlastdate"));
              throw new IOException ();
        }
        try{
              rowid = stream.readLong();
        }catch (EOFException ex_YESKIA92){
              System.err.println(LiveDB.grassHopper(ex_YESKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }
}
