/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.UTFDataFormatException;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.Externalizable;
import java.io.File;
import java.io.EOFException;
import java.io.IOException;
import java.util.concurrent.Semaphore;
import java.util.Date;
/* WATCH OUT:- USING STATIC IMPORTS */
import static DatabaseEngine.LiveDB.LOCKWAITPERIOD;
import static DatabaseEngine.LiveDB.LOCKWAITTIME;
import static DatabaseEngine.LiveDB.LASTLOCKMETHOD;
import static DatabaseEngine.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:LiveDB.tablenameidentifier, LiveDB.REALTABLES, 'InstalltionDirectory'/filetabledatafile.txt,
 *         LiveDB.indxinstalledTables, LiveDB.LOCKWAITPERIOD, LiveDB.LOCKWAITTIME
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:InstallDB.java method protected static boolean validateFileTableDataFileMap()
 * @LOOKSY:ArtLeavesInterfaces.java
 *
 * Class Name : FileTableDataFileMap implements Externalizable, Tables
 *
 * Methods : Look in Class definition
 * The Constructor method does a thorough validation before accepting data. If data is non-compliant it
 * does a System.exit(1) rather than a LiveDB.SmoothExitCondition.signalAll() because the object has
 * already been created the constructor is only filling in values into the instance.
 * There are no Setters for this table at all. There are two constructors for this table one is without
 * parameters the default which is meant to be used by fsck to help simulate a REFRESH.
 *
 * The Tables Interface has been faked here except for REFRESH, COMMIt, COMMITINDEXES because it is too
 * very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxfileTableDataFileMap
 * searchkey = Integer.toString(AnyLiveDB.REALTABLES);
 * searchkey returns Long rowid in the table
 *
 * Data File : YES Data present at Install time specified by filename filetabledatafile.txt in local directory
 * from which you installed the program with command line $InstallDB FileTableDataFileMap. Without dataentry
 * file for this table the application cannot install. It is mandatory to specify data for Critical table zero
 * in file filetabledatafile.txt at Install time. It looks for this file only. A readymade data entry file will
 * be provided to you which you need to customize specific to your installation. dont loose it take a physical
 * backup and keep it safe some place.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This is not a LiveDB.SerializeJustOnceTable(long p_tbl) it does see a frequent change
 * to the table_lock and pid fields but the underlying disk size of the table wont change because both fields
 * are constant in size although they are updated frequently. This is one of the three critical tables in the
 * database without which even an fsck will fail. This table will only contain entries for LiveDB.REALTABLES
 * zero thru seventeen. Tables Eighteen onwards are considered fake tables.
 *
 * Technical References : Buy and Read Asterix the Gladiator
 *
 * Start Date: 18th October 2010
 * End Date: 12th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class FileTableDataFileMap implements Externalizable, Tables, Cloneable {

    private long tableno ; /* store it here also for convenient programming */
    private String datafilename; // the comma seperated notepad data filename with path
    private String filename; /* the persistent storage file name for table without pathname */
    private String table_identifier_name; /* Table identifier name LiveDB.tablenameidentifier */
    private String classname; /* Class Name associated with this Table */
    private String table_lock; /* The Lock status for this table
                               * table_lock = "L" table locked by other process
                               * table_lock = "U" table unlocked.
                               * table_lock = "E" Emergency Declared
                               * table_lock = "POTA" Print Non-Bailable Arrest Warrants
                               * for terrorist customers and also that bugar on the console.
                               * Wait for next release no time now. There will be another
                               * thread called SystemCop which will print Arrest Warrants
                               * "You are Served"  "Volley Will Follow".
                               */
    private long pid ; /* internal process id that locked this table */
    private long rowid;  /* 8 FIELDS */

    private transient FileTableDataFileMap ft;
    private transient long fieldchanges;
    private transient File srf;
    private transient int ftn;
    private transient int itn;
    private transient Semaphore rowlock;

    {
       fieldchanges = 0L;
       srf = null;
       ft = null;
       ftn = 0;
       itn = 14;       
       rowlock = new Semaphore(1, true);
    }


    public FileTableDataFileMap (){
                
        tableno = 0;
        datafilename = "somejunk";
        filename = "somejunk";
        classname = "somejunk";
        table_identifier_name = "somejunk";
        table_lock = "U";
        pid = 1234;
        rowid = 0;
    }

    public FileTableDataFileMap (long p_Tableno, String p_DataFileName, String p_Filename,
        String p_Table_Identifier_Name, String p_Classname, String p_table_lock, long p_Rowid )
    {
        rowlock.acquireUninterruptibly();
        if (p_Tableno < 0 || p_Tableno > LiveDB.REALTABLES)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("FileTableDataFileMap constructor invalid tableno " + String.valueOf(p_Tableno)));
            System.exit(1);
        }
        tableno = p_Tableno;
        /* datafilename with path ... This has already been validated at Install time */
        if (p_DataFileName == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_DataFileName=null"));
            System.exit(1);
        }
        if ((p_DataFileName = LiveDB.convertFileName(p_DataFileName)) != null)
        {
            if (LiveDB.command_line.equals("FileTableDataFileMap"))
                datafilename = new String((p_DataFileName = LiveDB.convertFileName(p_DataFileName)).toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_DataFileName.length()));
            else datafilename = new String(p_DataFileName.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_DataFileName.length()));
        }
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_DataFileName in FileTableDataFileMap"));
            System.exit(1);
        }            
        rowlock.release();
        /* persistent storage filename with path */
        setFilename(p_Filename);
        setTable_identifier_name(p_Table_Identifier_Name);
        rowlock.acquireUninterruptibly();
        // this has already been validated from the contents of the text data entry file
        if (p_Classname ==null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid ClassName in FileTableDataFileMap"));
            System.exit(1);
        }

        classname = new String(p_Classname.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Classname.length()));
        rowlock.release();
        setTable_lock(p_table_lock);
        setPid(LiveDB.pid);
        rowlock.acquireUninterruptibly();
        if (LiveDB.LiveSharedDB[ftn].length == 0 && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid != ((InstalledTables)LiveDB.LiveSharedDB[ftn][LiveDB.LiveSharedDB[ftn].length -1]).getRowid() + 1)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        rowid = p_Rowid;
        rowlock.release();        
    }

    @Override
    protected Object clone(){
        rowlock.acquireUninterruptibly();
        try{
            ft = (FileTableDataFileMap)super.clone();
            ft.datafilename = new String(ft.datafilename.toCharArray());
            ft.classname = new String(ft.classname.toCharArray());
            ft.setFilename(ft.getFilename());
            ft.setTable_identifier_name(ft.getTable_identifier_name());
            ft.setTable_lock(ft.getTable_lock());
            // for entire row dont bother it will do a deep clone
        }catch (CloneNotSupportedException ex_NOKIA71){
            throw new InternalError (ex_NOKIA71.toString());
        }
        rowlock.release();
        return ft;
    }
    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables
        
        
        // first find the rowid for this table ftn = 0 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(ftn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("FileTableDataFileMap misses entry in indxinstalledTables"));
            // perhaps the index is corrupted so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == ftn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("FileTableDataFileMap misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(ftn));

        srf = new File (((FileTableDataFileMap)LiveDB.LiveSharedDB[ftn]
                [(int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno()]).getFilename());

        // outsize should be negative before every write
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
                
        
        try{
             stream.writeLong(tableno);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(datafilename);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field datafilename"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(filename);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field filename"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(table_identifier_name);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field table_identifier_name")); 
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(classname);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field classname")); 
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(table_lock);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field table_lock")); 
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(pid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pid")); 
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid")); 
              throw new IOException ();
        }

        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during FileTableDataFileMap serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
        rowlock.release();
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {
        rowlock.acquireUninterruptibly();
        try{
              tableno = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field tableno"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field tableno"));
              throw new IOException ();
        }

        try{
              datafilename = stream.readUTF();
        }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field datafilename"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field datafilename"));
              throw new IOException ();
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field datafilename"));
              throw new IOException ();
        }
        
        try{
              filename = stream.readUTF();
        }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field filename"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field filename"));
              throw new IOException ();
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field filename"));
              throw new IOException ();
        }

        try{
              table_identifier_name = stream.readUTF();
        }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field table_identifier_name"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field table_identifier_name"));
              throw new IOException ();
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field table_identifier_name"));
              throw new IOException ();
        }

        try{
              classname = stream.readUTF();
        }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field classname"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field classname"));
              throw new IOException ();
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field classname"));
              throw new IOException ();
        }

        try{
              table_lock = stream.readUTF();
        }catch (EOFException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field table_lock"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field table_lock"));
              throw new IOException ();
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field table_lock"));
              throw new IOException ();
        }

        try{
              pid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field pid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field pid"));
              throw new IOException ();
        }
        
        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }

    protected int SIZEOF(){
        return (8 + (datafilename.length() * 2) + (filename.length() * 2)
        + (table_identifier_name.length() * 2) + (classname.length() * 2)
        + (table_lock.length() * 2) + 8 + 8 );
    }   

    protected long getTableno(){
        return tableno;
    }
    protected String getDatafilename(){
        return new String(datafilename.toCharArray()).trim();
    }
    protected String getFilename(){
        return new String(filename.toCharArray()).trim();
    }
    protected String getTable_identifier_name (){
        return new String(table_identifier_name.toCharArray()).trim();
    }
    protected String getClassname(){
        return new String(classname.toCharArray()).trim();
    }
    protected String getTable_lock(){
        return new String(table_lock.toCharArray()).trim();
    }
    protected long getPid(){
        return pid;
    }
    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        return String.valueOf(getTableno()).concat(filher).concat(getDatafilename()).concat(filher)
               .concat(getFilename()).concat(filher).concat(getTable_identifier_name()).concat(filher)
               .concat(getClassname()).concat(filher).concat(getTable_lock()).concat(filher)
               .concat(String.valueOf(getPid())).concat(filher).concat(String.valueOf(getRowid()));
    }
    
    protected void setFilename(String p_Filename){

        rowlock.acquireUninterruptibly();
        if (p_Filename.indexOf(LiveDB.serializedfilespath) != -1
            || p_Filename.indexOf(LiveDB.FILESEPERATOR) != -1
            || Character.isDigit(p_Filename.toCharArray()[0])
            || checkFilename (p_Filename) == false )
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Filename:" + p_Filename));
            System.exit(1);
        }
        filename = new String(p_Filename.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Filename.length()));
        rowlock.release();
    }

    private boolean checkFilename(String p_Filename){

        if (p_Filename == null)
            return false;

        for (int i = 0; i < p_Filename.length(); i++)
        {
            if (Character.isDefined(p_Filename.toCharArray()[i])
                || Character.isHighSurrogate(p_Filename.toCharArray()[i])
                || Character.isLowSurrogate(p_Filename.toCharArray()[i])
                || Character.isISOControl(p_Filename.toCharArray()[i])
                || Character.isMirrored(p_Filename.toCharArray()[i])
                || Character.isUnicodeIdentifierPart(p_Filename.toCharArray()[i])
                || Character.isSpaceChar(p_Filename.toCharArray()[i]))
                return false;
        }        
        return true;
    }

    protected void setTable_identifier_name (String p_Table_identifier_name){
        boolean fnd = false;
        rowlock.acquireUninterruptibly();
        for (int i = 0; i < LiveDB.tablenameidentifier.length; i++)
        {
            if (LiveDB.tablenameidentifier[i][1].equals(p_Table_identifier_name))
            {
                fnd = true;
                break;
            }
        }        
        if (fnd == false)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid table identifier name:" + p_Table_identifier_name));
            System.exit(1);
        }        
        table_identifier_name = new String(p_Table_identifier_name.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Table_identifier_name.length()));
        rowlock.release();
    }
    
    protected void setTable_lock(String p_Table_lock){
       /* What if some other pid or some other thread of this pid has already locked table WAIT for lock to
        * release and then you grab the lock. Dont overdo this locking because if you forget to unlock "U"
        * the application will hang.
        */
        byte tick = 0;
        Date st = new Date ();

        if (p_Table_lock == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Table_lock=null"));
            System.exit(1);
        }

        rowlock.acquireUninterruptibly();

        
        while (getTable_lock().equals("E") && getPid() != LiveDB.pid)
        {
           if ((LOCKWAITTIME = (new Date ()).getTime() - st.getTime()) > LOCKWAITPERIOD)
           {
               tick += 1;
               
               if (tick < 3)
               {
                   LiveDB.ALRTLVL = 7;
                   System.err.println(LiveDB.grassHopper("WAITING FOR LOCK RELEASE FOR :" + String.valueOf(LOCKWAITTIME) + " "
                         + "LAST METHOD THAT LOCKED :" + LASTLOCKMETHOD));
               }
               else
               {
                   LiveDB.system_exit = 1; /* Do a smooth exit from the system no hurry */
                   LiveDB.SmoothExitCondition.signalAll();
               }
           }           
        }

        towaitforlock:
        while (getTable_lock().equals("L") && p_Table_lock.equals("L") && getPid() != LiveDB.pid)
            // you are requesting a lock but some other process has that lock already so wait
            continue towaitforlock;

        if (getTable_lock().equals("L") || getTable_lock().equals("E") || getTable_lock().equals("U"))
            table_lock = new String(p_Table_lock.toCharArray()) + new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Table_lock.length());
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Lock flag " + p_Table_lock));
            System.exit(1);
        }
        
        rowlock.release();
    }

    protected void setPid(long p_Pid){
        rowlock.acquireUninterruptibly();
        if (p_Pid != LiveDB.pid)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid Pid " + String.valueOf(p_Pid)));
            System.exit(1);
        }        
        this.pid = p_Pid;  // you dont want the peoples republic setting this so
        rowlock.release(); // package private
    }
    // although each instance of this Class InstalledTables represents a single row
    // we are committing the entire table here but before that a few checks and balances

    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;

        // we are committing the FileTableDataFileMap so mark those rows with change_flag 'C'
        for (int rw =  (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == ftn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                fnd = true;
                // insert an entry in ChangeTracker table for this table
                // Basically what this does is to backup this entirerow where you
                // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which may be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                   + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ftn][ftn]).getFilename() + ".ser",
                   LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR +
                   ((FileTableDataFileMap)LiveDB.LiveSharedDB[ftn][ftn]).getFilename() + ".ser" + ".res");
            if ( LiveDB.commit_database(ftn) == false )
                return false;
        }
        
        return true;
    }

    @Override
    public final boolean COMMITINDEXES(){
        
        if (LiveDB.LiveSharedDB[ftn].length == 0)
            return false;
        return LiveDB.write_IndexFile(ftn);
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(ftn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
