/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.Serializable;
import java.util.Date;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.concurrent.Semaphore;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:FileTableDataFileMap.java
 * @LOOKSY:ArtLeavesInterfaces.java
 *
 * Class Name : class LanguageSupport implements Serializable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH because it is too very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxlanguageSupport
 * searchkey = toString(AnyValid two or three letter language code );
 * searchkey returns Long rowid in the table
 *
 * LiveDB.indxLanguageIDs
 * searchkey = Diaplay Name of Language
 * searchkey returns ISO Language code
 *
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table contains all the languages supported by this Application. The constructor
 * does some usual validations and also plugs in a language name based on the two or three letter ISO language
 * code that you passed to the constructor.
 *
 * Technical References : Buy and Read Asterix at the Olympic Games.
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 * Last Modification Date : 6th October 2011
 * Last Modifications : LiveDB.indxlanguageSupport the index has been modified to fetch a rowid for
 * a given two or three letter language code.
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class LanguageSupport implements Serializable, Tables {
    private long lid; /* language id unique */
    private String language ; /* ISO Language code two letter */
    private String name; /* display name of the Language */
    private long rowid ; /* 4 FIELDS */

    private transient int lsn;
    private transient Semaphore rowlock;

    {
       rowlock = new Semaphore(1, true);
       lsn = 4;
    }

    // this is a default contructor without arguments needed for extend
    protected LanguageSupport ()
    {
        lid = rowid = 0L;
        language = name = null;
    }
    
    public LanguageSupport (long p_Lid, String p_Language, long p_Rowid) {
        rowlock.acquireUninterruptibly();
        lid = p_Lid;
        rowlock.release();
        setLanguage(p_Language);
        rowlock.acquireUninterruptibly();
        name = new Locale(language).getDisplayLanguage(new Locale(language));        
        if (LiveDB.LiveSharedDB[lsn].length == 0 && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);            
        }
        if (p_Rowid != ((LanguageSupport)LiveDB.LiveSharedDB[lsn][LiveDB.LiveSharedDB[lsn].length -1]).getRowid() + 1)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }
    protected long getLid(){
        return lid;
    }
    protected String getLanguage(){
        return new String(language.toCharArray());
    }
    protected String getName(){
        return new String(name.toCharArray());
    }
    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        return String.valueOf(getLid()).concat(filher).concat(getLanguage()).concat(filher).concat(getName())
               .concat(filher).concat(String.valueOf(getRowid()));
    }
    
    // carefull dont use this package private
    protected void setLanguage(String p_Language){
        boolean fnd = false;
        String iso3 = null;

        rowlock.acquireUninterruptibly();
        if (p_Language.length() > 2 || p_Language.length() < 2)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid language code:" + p_Language));
            System.exit(1);
        }

        if (p_Language.length() == 2)
        {
            for (int i = 0; i < Locale.getISOLanguages().length; i++)
            {
                if (p_Language.equals(Locale.getISOLanguages()[i]))
                {
                    fnd = true;
                    break;
                }
            }
            
            if (fnd == false)
            {
                LiveDB.ALRTLVL = 7;
                System.err.println(LiveDB.grassHopper("Invalid language code:" + p_Language));
                System.exit(1);
            }
        }

        if (p_Language.length() == 3)  // let this patch of code remain for later use
        {
            try{
            iso3 = LiveDB.DefaultLocale.getISO3Language();
            }catch (MissingResourceException ex_NOKIA91){}
            if (p_Language.equals(iso3) == false)
            {
                LiveDB.ALRTLVL = 3;
                System.err.println(LiveDB.grassHopper("Possible Invalid language code:" + p_Language));
            }
        }
 
        this.language = new String(p_Language.toCharArray());
        rowlock.release();
    }

    @Override
    public final boolean COMMIT(){
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(lsn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()

    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
