/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static DatabaseEngine.LiveDB.ft;
import static DatabaseEngine.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:LanguageSupport.java
 * @LOOKSY:PictureDatabase.java
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:LiveDB.indxinstalledTables
 *
 * Class Name : class LanguageTitles implements Externalizable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH because it is too very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxlanguageTitles
 * searchkey = langcode + englishlabel
 * searckey returns language_text in the table
 * 
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap. The data file
 * will contain the language transalations for each langcode and each englishlabel. A readymade data
 * entry file will be provided to you dont loose that take a physical backup and keep it safe some place.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table contains all the languages titles for all languages supported by this Application.
 * The constructor does some usual validations. Basically for each component of the GUI we need a language
 * transalation at run time as the user switches between languages. So this table provides a ready lookup.
 *
 * Technical References : Buy and Read Asterix in Britain.
 * Table Dedicated to : Jim Morrison (Doors)
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 * Last Modified Date : 12th September 2011
 * Last Modification : Total change in table schema. The old schema was
 * private long lid;
 * private long gui_comp_no;
 * private String pic;
 * private String gui_comp_type;
 * private String language_text;
 * private long rowid;
 *
 * The new schema is very simplified so for any english label used in the GUI you can get the
 * equivalent transalation for selected language with an index fetch.
 *   private String langcode; two letter ISO language code 
 *   private String englishlabel;  The label in English 
 *   private String language_text;  Language Text in Unicode 
 *   private long rowid;  4 FIELDS 
 *
 * Due to changes to table schema writeExternal and readExternal methods have also changed.
 * 
 * The indexes will also change. The two previous indexes were
 * Relevant Indexes : LiveDB.indx1languageTitles
 * searchkey = toString(AnyValidLid from LanguageSupport) + toString(AnyValidGUIComponentNumber)
 * searchkey returns Long rowid in the table
 * LiveDB.indx2languageTitles
 * searchkey = toString(AnyValidLid from LanguageSupport) + toString(AnyValidPictureCode from PictureDatabase)
 * searchkey returns Long rowid in the table 
 * 
 * Just one index now LiveDB.indxlanguageTitles
 * searchkey = langcode + englishlabel
 * searckey returns language_text in the table
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class LanguageTitles implements Externalizable, Tables {
    private String langcode; /* two letter ISO language code */
    private String englishlabel; /* The label in English */
    private String language_text; /* Language Text in Unicode */
    private long rowid; /* 4 FIELDS */

    private transient Semaphore rowlock;
    private transient File srf;
    private transient int itn;
    private transient int ltn;

    {
       rowlock = new Semaphore(1, true);
       srf = null;
       itn = 14;
       ltn = 5;
    }

    // this is a default constructor needed for an extend and call to super ()
    protected LanguageTitles ()
    {
        rowid = 0L;
        langcode = englishlabel = language_text = null;
    }
    
    public LanguageTitles ( String p_langcode, String p_englishlabel, String p_Language_Text,
                            long p_Rowid)
    {

        rowlock.acquireUninterruptibly();
        if (p_langcode == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_langcode=null"));
            System.exit(1);
        }
        langcode = new String (p_langcode.toCharArray());
        rowlock.release();
        setenglishLabel (p_englishlabel);
        setLanguage_text(p_Language_Text);

        rowlock.acquireUninterruptibly();
        if (LiveDB.LiveSharedDB[ltn].length == 0 && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid != ((LanguageTitles)LiveDB.LiveSharedDB[ltn][LiveDB.LiveSharedDB[ltn].length -1]).getRowid() + 1)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }

    protected int SIZEOF(){ // dont use a getter here ever
        return ((langcode.length() * 2) + (englishlabel.length() * 2) + (language_text.length() * 2) + 8);
    }

    protected String getlangCode ()
    {
        return new String(langcode.toCharArray());
    }

    protected String getenglishLabel (){
        return new String(englishlabel.toCharArray());
    }

    protected String getLanguage_text(){
        return new String(language_text.toCharArray());
    }

    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";

        return getlangCode().concat(filher).concat(getenglishLabel ()).concat(filher)
               .concat(getLanguage_text()).concat(filher).concat(String.valueOf(getRowid()));
    }

    protected void setenglishLabel (String p_label){
        if (p_label == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_label=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        englishlabel = new String(p_label.toCharArray());
        rowlock.release();
    }

    protected void setLanguage_text(String p_Language_text){
        if (p_Language_text == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Language_text=null"));
            System.exit(1);
        }
        // no blank padding needed since this is a Load Once Serialize Once table
        rowlock.acquireUninterruptibly();
        this.language_text = new String(p_Language_text.toCharArray());
        rowlock.release();
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables

        // first find the rowid for this table ltn = 5 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(ltn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("LanguageTitles misses entry in indxinstalledTables"));
            // perhaps the index is corrupted or has not been updated someplace so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == ltn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("LanguageTitles misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(ltn));

        srf = new File (((FileTableDataFileMap)LiveDB.LiveSharedDB[ft]
                [(int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno()]).getFilename());

        // outsize should be negative before every write
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);


        try{
              stream.writeUTF(langcode);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field langcode"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during LanguageTitles serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        
        try{
              stream.writeUTF(englishlabel);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field englishlabel"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during LanguageTitles serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        
        try{
              stream.writeUTF(language_text);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field language_text"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during LanguageTitles serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }

        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during LanguageTitles serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
        rowlock.release();
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {
        rowlock.acquireUninterruptibly();
        // now read in all values here


        try{
              langcode = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field langcode"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field langcode"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field langcode"));
              throw new IOException ();
        }
        
        try{
              englishlabel = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field englishlabel"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field englishlabel"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field englishlabel"));
              throw new IOException ();
        }

        try{
              language_text = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field language_text"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "field language_text"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field language_text"));
              throw new IOException ();
        }

        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }

    @Override
    public final boolean COMMIT(){
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(ltn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
