/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.Serializable;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static DatabaseEngine.LiveDB.pt;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:IntranetCompliantRadioTaggedCollar.java
 * @LOOKSY:RefreshDatabase.read_database()
 * @LOOKSY:ArtLeavesInterfaces.java
 *
 * Class Name : class PaymentType implements Serializable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH because it is too very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxpaymentType
 * searchkey = toString(AnyValidPaymentID)
 * searchkey returns Long rowid in the table
 *
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap. The data file
 * will contain the payment types for accepting payment. A readymade data entry file will be provided to you
 * dont loose that take a physical backup and keep safe.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists all different payment types which will be supported by Application.
 *
 * Technical References : Buy and Read Asterix and the Cauldron.
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class PaymentType implements Serializable, Tables {
    /* Each instance of this object stores one Row of Data */
    private long pmtid;  /* Payment ID */
    private String pmt_type; /* Description like Cheque, CC, WUMT, Cash, COD etc */
    private long rowid; /* 3 FIELDS */

    private transient int ptn ;
    private transient Semaphore rowlock;

    {
       rowlock = new Semaphore(1, true);
       ptn = pt;
    }

    protected PaymentType ()
    {
        pmtid = rowid = 0L;
        pmt_type = null;
    }
    
    public PaymentType (long p_Pmtid, String p_Pmt_Type, long p_Rowid ) {
        rowlock.acquireUninterruptibly();
        pmtid = p_Pmtid; /* Store one Data Row */
        rowlock.acquireUninterruptibly();
        setPmt_type(p_Pmt_Type);

        rowlock.acquireUninterruptibly();
        if (LiveDB.LiveSharedDB[ptn].length == 0 && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid != ((PaymentType)LiveDB.LiveSharedDB[ptn][LiveDB.LiveSharedDB[ptn].length -1]).getRowid() + 1)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }
    /* Getters and Setters provide a public interface to private data */
    /* They follow the rulebook of encapsulation */
    /* Dont provide Setters for all fields specially key fields !! */
    /* If you have to provide a setter for some sensitive data then
     * make that setter package private or protected
     */
    protected long getPmtid(){
        return pmtid;
    }
    protected String getPmt_type(){
        return new String(pmt_type.toCharArray());
    }
    protected long getRowid(){
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        return String.valueOf(getPmtid()).concat(filher).concat(getPmt_type()).concat(filher)
               .concat(String.valueOf(getRowid()));
    }
    
    protected void setPmt_type(String p_Pmt_type){
        if (p_Pmt_type == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Pmt_type=null"));
            System.exit(1);
        }            
        rowlock.acquireUninterruptibly();
        pmt_type = new String(p_Pmt_type.toCharArray());
        rowlock.release();
    }

    @Override
    public final boolean COMMIT(){
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(ptn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
