/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;


import java.io.ObjectInput;
import java.io.Externalizable;
import java.io.ObjectOutput;
import java.io.File;
import java.io.UTFDataFormatException;
import java.io.EOFException;
import java.io.IOException;
import java.util.concurrent.Semaphore;
import java.io.Serializable;
import java.util.Date;
import static DatabaseEngine.LiveDB.ft;
import static DatabaseEngine.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:WriteTables.java  --- Take a Quick Look
 * @LOOKSY:InstalledTables.java  --- Take a Quick Look
 * @LOOKSY:FileTableDataFileMap.java  --- Take a Quick Look
 * @LOOKSY:ArtLeavesInterfaces.java  --- Take a Quick Look
 * @LOOKSY:RefreshDatabase.read_database()  --- Take a Quick Look
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.CopyFile, LiveDB.commit_database(), LiveDB.write_IndexFile()
 * @LOOKSY:LiveDB.LiveSharedDB, LiveDB.ALRTLVL, LiveDB.Action, LiveDB.maxtables, LiveDB.IPCSTrigger,
 * @LOOKSY:LiveDB.TrainsInWaiting, LiveDB.IPCSCondition, LiveDB.IPCSLOCK
 *
 * Class Name : PlatformClearance implements Externalizable, Cloneable, Tables
 *              Contains Instance of class TrainRunningInformation implements Serializable, Cloneable
 *
 * Methods : Look in Class definition
 * This constructor method is very complicated and so i will take a few lines to explain what it does.
 * TERMINALA replaces rowlock. It is just another semaphore. I find it more sensible to use these for
 * rowlocks. A couple of setters follow. Now whenever there is an Action that involves a write to the
 * database table. How do you know that the other bugar is not writing also ? So we create a .syc file
 * on disk using a LiveDB.IPCSTrigger(Object number) which is the table number in this release. It can
 * also be a sort number and index number. But we are not using it thataways yet. So the constructor
 * needs to delete the old .syc file first before creating a new one.  That trigger will check for the 
 * existence of a file on disk for the same ObjectNumber and create it if not found. That trigger takes
 * care of placing the thread on hold to keep waiting for the IPCSCondition and also does a SignalAll 
 * automatically. You wont have to bother about it in your code at all. In short every process that 
 * writes must create a File name synchronizedfilespath + "ipcssync_objid".syc and delete that file when
 * it has finished writing. CONFUSING ?? Dont read this at all. I have also allowed my imagination to
 * run wet and wild on those advanced type enums dont let that happen or you will never get to bed with
 * your spouse.
 *
 * The Tables Interface has been faked except COMMIT, COMMITINDEXES, REFRESH because I dont want you doing
 * naughty things with this table.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indx1platformClearance
 * searchkey = Integer.toString(((PlatformClearance)getEntirerow()).getRequestno()) + Long.toString(((PlatformClearance)getEntirerow()).getSid())
 * searchkey returns Long rowid which is the rowid for this table.
 *
 * LiveDB.indxplatformClearance
 * searchkey = AnyValidDate (getRequesttime)
 * searchkey returns Long rowid which is the rowid for this table
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : The Queue LiveDB.TrainsInWaiting is based on this table. This queue along with this
 * table acts as a synchronization mechanism for all LiveDB.Action posted by this instance of running app
 * and also for various tables. The Queue (although an arraydeque synonymous with a queue) accepts requests
 * on a FIFO (First In First Out) basis. And when the request is completed it is also de-queued from
 * LiveDB.TrainsInWaiting. So every DML that happens on this table is indirectly thru TrainsInWaiting.
 * So any thread has to offer a request for the Queue TrainsInWaiting to accept. Now the understanding is
 * that for any database related activity your thread must read only after a write has happened by some
 * other thread if a request has been placed in the queue. Whatever be the type of request you must wait
 * your turn patiently. This is as far as the queue TrainsInWaiting is concerned. Now the underlying table
 * PlatformClearance will contain a historical record of all database activity which I am purging frequently
 * in Spy.java method public boolean LostSouls (). You dont have to worry about synchronization with other
 * processes on the system since the queue contains a reference to this table always and this table gets
 * refreshed constantly and also the constructor method creates a file on disk to avoid collisions.
 * This TrainsInWaiting is specific to your process (JVM instance of this application) and I am not merging
 * the different queues anywhere periodically. So there is a tiny bug here which I will sort out just when
 * I have obtained my turn to proceed with my request number by using some indexes. Hope it works !!!
 *
 * Technical References : Buy and Read Asterix the Soothsayer.
 * Table Dedicated To: Joe Zawinul and Wayne Shorter
 *
 * Start Date: 18th October 2010
 * End Date: 16th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

final class PlatformClearance implements Externalizable, Cloneable, Tables {
        private int requestno;  /* random generation */
        private long sid ;  /* station id or process id */
        private int request_type; /* valid request types enum Action */
        private Date request_time;
        private TrainRunningInformation trainid;
        private long rowid; /* 6 FIELDS */

        private transient long fieldchanges;
        private transient TrainRunningInformation tri; // this is just for manipulating and returning trainid
        private transient int pcn ;
        private transient File srf;
        private transient int itn;        
        private transient PlatformClearance pf;        
        private transient Semaphore TERMINALA;  // for international flights

        // initialization block
        {
           tri = null;
           fieldchanges = 0L;
           pcn = LiveDB.pc;
           srf = null;
           itn = LiveDB.it;
           pf = null;
           TERMINALA = new Semaphore (1,true);           
        }
        
        public PlatformClearance ( int p_Requestno, long p_Sid, LiveDB.Action p_Request_type, 
                                   Date p_Request_time, TrainRunningInformation p_Trainid, long p_Rowid)
        {            
            setRequestno(p_Requestno);
            TERMINALA.acquireUninterruptibly();
            sid = p_Sid;
            TERMINALA.release();
            setRequest_type(p_Request_type);
            setRequest_time(p_Request_time);
            setTrainid(p_Trainid);

            TERMINALA.acquireUninterruptibly();
            if (LiveDB.LiveSharedDB[pcn].length == 0 && p_Rowid != 0)
            {
                LiveDB.ALRTLVL = 7;
                System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
                System.exit(1);
            }
            if (p_Rowid >= 0)
                rowid = p_Rowid;
            else
            {
                LiveDB.ALRTLVL = 7;
                System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
                System.exit(1);
            }
            
            if (this.request_type == LiveDB.Action.WRITEOBJECT.ordinal() || this.request_type == LiveDB.Action.FASTCOMMIT.ordinal() ||
                this.request_type == LiveDB.Action.WRITELINE.ordinal())
            {
                // delete the previous ipcsfile created if certain conditions are met
                if ((((TrainRunningInformation)
                   ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getTrainid()).getTrain() == this.trainid.getTrain()
                   && ( ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.FASTCOMMIT.ordinal() || ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.WRITEOBJECT.ordinal() || ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.WRITELINE.ordinal() )))
                {
                    /* if the previous write was on the same tableno or index number then this process created that ipcsfile */
                    /* so go into a loop until accidenttime or arrivaltime is null. After that delete the file */
                    tohover:
                    while ((((TrainRunningInformation)
                        ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getTrainid()).getArrivaltime()
                         == null && ((TrainRunningInformation)
                        ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getTrainid()).getAccident() == null))
                                          continue tohover;
                    LiveDB.IPCSTrigger(-1); // -1 or less than zero value will delete old ipcsfile
                }
                else if ((((TrainRunningInformation)
                   ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getTrainid()).getTrain() != this.trainid.getTrain()
                   && ( ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.FASTCOMMIT.ordinal() || ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.WRITEOBJECT.ordinal() || ((PlatformClearance)(LiveDB.LiveSharedDB[pcn][LiveDB.LiveSharedDB[pcn].length -2])).getRequest_type()
                   == LiveDB.Action.WRITELINE.ordinal() )))
                   LiveDB.IPCSTrigger(-1); // -1 or less than zero value will delete old ipcsfile
              
                LiveDB.IPCSTrigger(this.trainid.getTrain()); // now the trigger will create a new ipcsfile
            }
            TERMINALA.release();
        }

        @Override
        protected Object clone () {
            TERMINALA.acquireUninterruptibly();
            try{
              pf = (PlatformClearance)super.clone();
              pf.setRequest_time(pf.getRequest_time());
              pf.trainid = (TrainRunningInformation)trainid.clone();
            }catch (CloneNotSupportedException ex_NOKIA46){
                // this is not possible unless a programming screwup so throw unchecked error
                throw new InternalError (ex_NOKIA46.toString());
            }
            TERMINALA.release();
            return pf;
        }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {

        TERMINALA.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables

        // first find the rowid for this table pcn = 17 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(pcn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("PlatformClearance misses entry for self in indxinstalledTables"));
            // perhaps the index is corrupted so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == pcn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("PlatformClearance misses entry for self in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(pcn));

        srf = new File (((FileTableDataFileMap)LiveDB.LiveSharedDB[ft]
                [(int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno()]).getFilename());

        // outsize should be negative before every write
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeInt(requestno);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91,"writeExternal Class PlatformClearance Field requestno"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(sid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91,"writeExternal Class PlatformClearance Field sid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeInt(request_type);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class PlatformClearance Field request_type"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getRequest_time().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class PlatformClearance Field request_time"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              trainid.writeExternal(stream);              
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class PlatformClearance Field trainid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class PlatformClearance Field rowid"));
              throw new IOException ();
        }

        
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during PlatformClearance serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
        TERMINALA.release();
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {
   
        TERMINALA.acquireUninterruptibly();

        try{
              requestno = stream.readInt();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field requestno"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field requestno"));
              throw new IOException ();
        }

        try{
              sid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field sid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field sid"));
              throw new IOException ();
        }

        try{
              request_type = stream.readInt();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field request_type"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field request_type"));
              throw new IOException ();
        }

        try{
              request_time = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field request_time"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field request_time"));
              throw new IOException ();
        }

        try{
            trainid.readExternal(stream);
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field trainid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field trainid"));
              throw new IOException ();
        }

        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal Class PlatformClearance Field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal Class PlatformClearance Field rowid"));
              throw new IOException ();
        }
        TERMINALA.release();
   }

   protected int SIZEOF(){
            return (4 + 8 + 4 + 8 + this.trainid.SIZEOF() + 8 );
   }

   protected int getRequestno (){
        return this.requestno;
   }
   protected long getSid(){
        return this.sid;
   }
   protected int getRequest_type(){
        return this.request_type;
   }
   protected Date getRequest_time() {
        return new Date(request_time.getTime());
   }
   protected TrainRunningInformation getTrainid(){
        tri = (TrainRunningInformation) this.trainid.clone();
        tri.trackname = tri.trackname.trim();
        tri.line = tri.line.trim();
        return tri;
   }
   protected long getRowid(){
       return this.rowid;
   }

   protected void setRequestno(int p_Request){
        TERMINALA.acquireUninterruptibly();
        this.requestno = p_Request;
        TERMINALA.release();
   }
   
   protected void setRequest_type(LiveDB.Action p_Type){
            TERMINALA.acquireUninterruptibly();
            this.request_type = p_Type.ordinal();
            if (this.request_type == LiveDB.Action.WRITEOBJECT.ordinal() 
                || this.request_type == LiveDB.Action.FASTCOMMIT.ordinal()
                || this.request_type == LiveDB.Action.WRITELINE.ordinal() )
                LiveDB.IPCSTrigger(this.trainid.getTrain());
            TERMINALA.release();
   }

   protected void setRequest_time (Date p_Time){
        TERMINALA.acquireUninterruptibly();
        this.request_time = new Date(p_Time.getTime());
        TERMINALA.release();
   }
   protected void setTrainid (TrainRunningInformation p_Trainid){
        TERMINALA.acquireUninterruptibly();
        this.trainid = (TrainRunningInformation)p_Trainid.clone();
        TERMINALA.release();
   }
   
   @Override
   public String toString(){
       String filher = "%";
       return String.valueOf(getRequestno()).concat(filher).concat(String.valueOf(getSid())).concat(filher)
              .concat(String.valueOf(getRequest_type())).concat(filher).concat(String.valueOf(getRequest_time().getTime()))
              .concat(filher).concat(getTrainid().toString()).concat(filher).concat(String.valueOf(getRowid()));
   }

   protected void setRowid (long p_row){
        TERMINALA.acquireUninterruptibly();
        this.rowid = p_row;
        TERMINALA.release();
   }

    // although each instance of this Class PlatformClearance represents a single row
    // we are committing the entire table here but before that a few checks and balances

    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;
 
        // we are committing the PlatformClearance table so mark those rows with change_flag 'F'
        for (int rw =  (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == pcn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                   ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                   fnd = true;
                   // insert an entry in ChangeTracker table for this table
                   // Basically what this does is to backup this entirerow where you
                   // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                   LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                   new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                   LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which MAY be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                    + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][pcn]).getFilename() + ".ser",
                   LiveDB.serializedfilespath + "backup" + LiveDB.FILESEPERATOR +
                   ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][pcn]).getFilename() + ".ser" + ".res");
            if ( LiveDB.commit_database(pcn) == false )
                return false;
        }
        
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){

        if (LiveDB.LiveSharedDB[pcn].length == 0)
            return false;
        return LiveDB.write_IndexFile(pcn);
    }

    @Override
    public final boolean REFRESH(){
        int pfc = 17;
        if (RefreshDatabase.read_database(pfc) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}

/*
 * WIRING DIAGRAM :
 * @LOOKSY:class PlatformClearance
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.ALRTLVL, LiveDB.Action, LiveDB.maxtables,
 *         LiveDB.TrainsInWaiting
 *
 * Class Name : TrainRunningInformation implements Serializable, Cloneable
 *
 * Methods : Look in Class definition
 * This constructor method is very complicated and so i will take a few lines to explain what it does.
 * Every time you construct a PlatformClearance object you also need to construct a TrainRunningInformation
 * object and pass it to that constructor. This instance of TrainRunningInformation contains a departure
 * time but accident time and arrival time will be null and running time will be zero. These are the times
 * for the various application actions listed at LiveDB.Action pertaining to a certain table number which
 * may be a real table or a fake table from 18 thru 50 like LiveDB.RunAwayBrideReport(). When LiveDB.Action
 * is completed the other stats like accident time (if error) or arrival time is filled in and cloned and
 * the PlatformClearance table is made to point to this clone.
 *
 * The Tables Interface is not implemented here
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxplatformClearance
 * searchkey = Integer.toString(((PlatformClearance)getEntirerow()).getRequestno()) + Long.toString(((PlatformClearance)getEntirerow()).getSid())
 * searchkey returns Long rowid which is the rowid for this table.
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : This is a contained class. It contains some usefull stats from a reporting angle
 * and you could also use this for ThunderBall tuning and posting speed limits for various threads. I am
 * not doing it as of now due to lack of time but will do so in next release. Remember that original shloka
 * aaj karanthaa kul kar. kul karanthaa parson. parson karantha narson from the city of hyderabad otherwise
 * the code will be filled with bugs (they taught you to count only 99 right ??).
 *
 * Technical References : Buy and Read Asterix the Soothsayer.
 * Table Dedicated To : Jaco Pastorius
 *
 * Start Date: 18th October 2010
 * End Date: 17th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

class TrainRunningInformation implements Externalizable, Cloneable {
        private int train; /* Either a tableno or index no or sorttable no */
        private long track; /* thread id Thread.currentThread.getId() */
        private int itrack; /* internal track or internal thread id as in ThreadController.DashBoardGPS */
        String trackname ; /* thread name */
        String line ; /* Which line South, North , North Western, East basically a threadgroup that created this thread Threadgroup.getName()*/
        private Date departuretime; /* departure time request execution start time */
        private Date arrivaltime; /* arrival time request completion time */
        private Date accident; /* accident time if request derails due to Yank or some other reason */
        private long runningtime; /* time taken to complete request */

        private transient TrainRunningInformation tri;
        private transient long fieldchanges;
        private transient int pcn;
        private transient int metadatasize;
        private transient Semaphore TERMINALB;

        {
            tri = null;
            fieldchanges = 0L;
            pcn = 17;
            metadatasize = 0;
            TERMINALB = new Semaphore(1,true);
        }
        public TrainRunningInformation ( int p_Train, long p_Track, String p_Trackname, String p_Line, Date p_Departuretime, Date p_Arrivaltime,
                Date p_Accident, long p_Runningtime ){
            train = p_Train;
            track = p_Track;
            itrack = ThreadController.getcurrenthread();
            setTrackname(p_Trackname);
            setLine(p_Line);
            setDeparturetime(p_Departuretime);
            setArrivaltime(p_Arrivaltime);
            setAccident(p_Accident);
            setRunningtime(p_Runningtime);
            
        }
        // Now do you understand why I have used primitive types inside classes
        // otherwise you will have to serialize any object and then deserialze
        // it into a new reference variable to create a copy of an object in Java
        // otherwise create a new instance and then copy yourself one attribute at
        // a time. Remember some objects such as Wrappers are immutable and so it
        // doesent make sense to clone and so they are not cloneable either.
        // but you can do your own copy constructor if the class is final
        // otherwise slicing can happen
        @Override
        protected Object clone () {
            try{
               tri = (TrainRunningInformation) super.clone ();
              /* Now do those funny things Both original and clone are
               * referencing the Same String objects trackname and line.
               * Therefore take the original and explicit copy and reassign
               * to clone. Do the same with any Object references other than
               * String.
               */
              // Thanks to a lesson from Bill Venneer              
               tri.setTrackname(tri.getTrackname());
               tri.setLine(tri.getLine());
               tri.setDeparturetime(tri.getDeparturetime());
               tri.setArrivaltime(tri.getArrivaltime());
               tri.setAccident(tri.getAccident());
            }catch (CloneNotSupportedException ex_NOKIA47){
                // this is not possible so throw unchecked error
                throw new InternalError (ex_NOKIA47.toString());
            }
            return tri;
        }

   @Override
   public final void writeExternal (ObjectOutput stream) throws IOException {

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeInt(train);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field train"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(track);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field track"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeInt(itrack);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field itrack"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(trackname);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field trackname"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeUTF(line);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field line"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getDeparturetime().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field departuretime"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getArrivaltime().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field arrivaltime"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(getAccident().getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field accident"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during TrainRunningInformation serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        try{
              stream.writeLong(runningtime);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class TrainRunningInformation Field runningtime"));
              throw new IOException ();
        }
    }


    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {

        // dont acquire rowlock here dummy
        
        try{
              train = stream.readInt();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field train"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field train"));
              throw new IOException ();
        }

        try{
              track = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field track"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field track"));
              throw new IOException ();
        }

        try{
              itrack = stream.readInt();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field itrack"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field itrack"));
              throw new IOException ();
        }

        try{
              trackname = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field trackname"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "readExternal TrainRunningInformation Field trackname"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field trackname"));
              throw new IOException ();
        }

        try{
              line = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field line"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "readExternal TrainRunningInformation Field line"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field line"));
              throw new IOException ();
        }

        try{
              departuretime = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field departuretime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field departuretime"));
              throw new IOException ();
        }
        
        try{
              arrivaltime = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field arrivaltime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field arrivaltime"));
              throw new IOException ();
        }

        try{
              accident = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field accident"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field accident"));
              throw new IOException ();
        }

        try{
              runningtime = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "readExternal TrainRunningInformation Field runningtime"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "readExternal TrainRunningInformation Field runningtime"));
              throw new IOException ();
        }
   }

   protected int SIZEOF(){
       return (4 + 8 + 4 + (getTrackname().length() * 2) + (getLine().length() * 2) + 8 + 8 + 8 + 8 );
   }
   
   protected int getTrain(){
       return train;
   }

   protected long getTrack(){
       return track;
   }

   protected int getITrack(){
       return itrack;
   }

   protected String getTrackname(){
       return new String(trackname.toCharArray()).trim();
   }

   protected String getLine(){
       return new String(line.toCharArray()).trim();
   }

   protected Date getDeparturetime(){
       return new Date(departuretime.getTime());
   }

   protected Date getArrivaltime(){
       return new Date(arrivaltime.getTime());
   }

   protected Date getAccident(){
       return new Date(accident.getTime());
   }

   protected long getRunningtime(){
       return runningtime;
   }
   
   @Override
   public String toString(){
       String filher = "#";
       return String.valueOf(getTrain()).concat(filher).concat(String.valueOf(getTrack())).concat(filher)
              .concat(String.valueOf(getITrack())).concat(filher).concat(getTrackname()).concat(filher)
              .concat(getLine()).concat(filher).concat(String.valueOf(getDeparturetime().getTime())).concat(filher)
              .concat(String.valueOf(getArrivaltime().getTime())).concat(filher).concat(String.valueOf(getAccident().getTime()))
              .concat(filher).concat(String.valueOf(getRunningtime()));
   }

   protected void setTrackname (String p_Trackname){
        if (p_Trackname == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Trackname=null"));
            System.exit(1);
        }
       TERMINALB.acquireUninterruptibly();
       trackname = new String(p_Trackname.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Trackname.length()));
       TERMINALB.release();
   }

   protected void setLine (String p_Line){
       TERMINALB.acquireUninterruptibly();
       this.line = new String(p_Line.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Line.length()));
       TERMINALB.release();
   }

   protected void setDeparturetime(Date p_Departuretime){
       TERMINALB.acquireUninterruptibly();
       if ((p_Departuretime != null && p_Departuretime.after(getArrivaltime()) && getArrivaltime() != null))
       {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Departure post Arrival! Departure time " + getDeparturetime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
       }
       if (p_Departuretime != null && p_Departuretime.after(getAccident()) && getAccident() != null)
       {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Departure post Accident! Departure time " + getDeparturetime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
       }
       this.departuretime = new Date(p_Departuretime.getTime());
       TERMINALB.release();
   }
   protected void setArrivaltime(Date p_Arrivaltime){
        TERMINALB.acquireUninterruptibly();
        if (p_Arrivaltime.before(getAccident()) && getAccident() != null && p_Arrivaltime != null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Arrival before Accident! Arrival time " + getArrivaltime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
        }

        if (p_Arrivaltime.before(getDeparturetime()) && getDeparturetime() != null && p_Arrivaltime != null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Arrival before Departure! Arrival time " + getArrivaltime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
        }
        
        this.arrivaltime = new Date(p_Arrivaltime.getTime());
        TERMINALB.release();
   }
   protected void setAccident(Date p_Accident){
        TERMINALB.acquireUninterruptibly();
        if (p_Accident != null && p_Accident.before(getDeparturetime()))
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("p_Accident before Departure " + getDeparturetime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
        }

        if (p_Accident != null && p_Accident.after(getArrivaltime()))
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("p_Accident after Arrival " + getArrivaltime().toString()));
            LiveDB.system_exit = 1;
            LiveDB.SmoothExitCondition.signalAll();
        }
        this.accident = new Date(p_Accident.getTime());
        TERMINALB.release();
   }
   protected void setRunningtime(long p_Runningtime){
        TERMINALB.acquireUninterruptibly();
        this.runningtime = p_Runningtime;
        TERMINALB.release();
    }
}
