/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.EOFException;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.UTFDataFormatException;
import java.util.Comparator;
import java.util.Date;
import java.util.concurrent.Semaphore;
import static DatabaseEngine.LiveDB.ft;
import static DatabaseEngine.LiveDB.ct;

/*
 * WIRING DIAGRAM :
 * @LOOKSY:Spy.java method public boolean LostSouls ()
 * @LOOKSY:SortandIndex.java
 * @LOOKSY:ChangeTracker.java
 * @LOOKSY:ArtLeavesInterfaces.java
 * @LOOKSY:RefreshDatabase.read_database()
 * @LOOKSY:LiveDB.indxinstalledTables, LiveDB.CopyFile, LiveDB.commit_database(), LiveDB.write_IndexFile()
 * @LOOKSY:LiveDB.CTSR, LiveDB.LiveSharedDB, LiveDB.ALRTLVL
 *
 * Class Name : ReportsTable implements Externalizable, Comparable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here other than COMMIT, COMMITINDEXES, REFRESH.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxreportGeneration  -- for a report id ascertain most recent report generation date
 * searchkey = AnyValidReportID(Integer)
 * searchkey returns Date which represents the most recent report generation date for this report id
 * LiveDB.indxreportAllocation
 * searchkey = AnyValidReportID(Integer)
 * searchkey returns the same number as a confirmation that this report id has been allocated to your process
 * LiveDB.indxreportRecentRow  -- for a given report generation date ascertain the report id in the table
 * searchkey = AnyValidDate(LiveDB.indxreportGeneration.get())
 * searchkey returns Long rowid for a the most recent report matching that date.
 * LiveDB.indxreportsTable -- find a certain document number
 * searchkey = Integer.toString(getDocumentno());
 * searchkey returns Long rowid in the table
 *
 * Relevant Sorts: LiveDB.srt1reportstable, LiveDB.srt2reportstable, LiveDB.srt3reportstable
 * srt1reportstable and srt3reportstable are not being utilized as of now but you might want to use those.
 * srt1reportstable is sorted on id,srt2reportstable is sorted on report date,srt3reportstable is sorted on status
 * Look at all Comparable and Comparator interfaces Implementation below for details.
 * srt2reportstable will sort ascending on report date using comparator DateComp below. It will be a ascending
 * sort since we are just returning the return value of compareTo
 *
 * Data File : None
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table is critical from the ReportWriter perspective and also for the running
 * thread SortandIndex. This table will contain all reports allocated to all processes historically which
 * is also being purged periodically with a cut off date in thread Spy.java method LostSouls(). If you
 * look at SortandIndex.java method sort_DataFile () top half you will get some clarity on how reports
 * specific to your process are being allocated and plugged into this table.
 *
 * Technical References : Buy and Read Asterix and the Great Crossing.
 * Table Dedicated to : Thelonius Monk (Monk in Italy)
 *
 * Start Date: 18th October 2010
 * End Date: 15th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

final class ReportsTable implements Externalizable, Comparable, Tables{
      private long documentno; /* random number generation */
      private long id; /* report id */
      private String doctype; /* type purchase order 'P' or other MIS reports */
      private String pathname; /* path and file name of document on disk */
      private long size; /* Intial size of report can change ... PER Report size and not file size */
      private Date reportdate ; /* Date of report generation */
      private char status ; /* report status 'C' completed 'I' Incomplete 'P' printing 'T' Taken/alloted
                             * 'A' Available for random allocation to any process id
                             */
      private long byteswritten; /* report bytes written so we can continue from last byte if Incomplete */
      private long pid; /* process id to which this report has been allocated randomly */
      private long rowid ;  /* 10 FIELDS */

      private transient long fieldchanges; // for later Ultra Fast Commits
      private transient int rtn;
      private transient int itn;
      private transient File srf;
      private transient ReportsTable rt;
      private transient Semaphore rowlock;

      {
         fieldchanges = 0L; // This database cannot support more than 64 fields per table for the present
         rtn = LiveDB.rt;
         itn = LiveDB.it;
         srf = null;
         rt = null;
         rowlock = new Semaphore(1, true);
      }

      public ReportsTable (long p_Documentno, long p_Id, String p_Doctype, String p_Pathname,
              long p_Size, Date p_Reportdate, char p_Status, long p_Byteswritten, long p_Pid,
              long p_Rowid) {
          rowlock.acquireUninterruptibly();
          documentno = p_Documentno;
          rowlock.release();
          rowlock.acquireUninterruptibly();
          id = p_Id;
          rowlock.release();
          setDoctype(p_Doctype); // validate later after checking code
          setPathname(p_Pathname);
          setSize(p_Size);
          setReportdate(p_Reportdate);
          setStatus(p_Status); // validate later after checking code
          setByteswritten(p_Byteswritten);
          setPid(p_Pid);
          setRowid(p_Rowid);
      }

      @Override
      protected Object clone(){
          rowlock.acquireUninterruptibly();
          try{
              rt = (ReportsTable) super.clone();
              rt.setDoctype(rt.getDoctype());
              rt.setPathname(rt.getPathname());
              rt.setReportdate(rt.getReportdate());
          }catch (CloneNotSupportedException ex_NOKIA44){
              throw new InternalError (ex_NOKIA44.toString());
          }
          rowlock.release();
          return rt;
      }
      
      protected int SIZEOF(){
        return (8 + 8 + (doctype.length() * 2) + (pathname.length() * 2) + 8 + 8 + 2 + 8 + 8 + 8);
      }
      
      protected long getDocumentno(){
          return documentno;
      }
      protected long getId(){
          return id;
      }
      protected String getDoctype(){
          return new String(doctype.toCharArray()).trim();
      }
      protected String getPathname(){
          return new String(pathname.toCharArray()).trim();
      }
      protected long getSize(){
          return size;
      }
      protected Date getReportdate(){
          return new Date(reportdate.getTime());
      }
      protected char getStatus(){
          return status;
      }
      protected long getByteswritten (){
          return byteswritten;
      }
      protected long getPid(){
          return pid;
      }
      protected long getRowid(){
          return rowid;
      }

      @Override
      public String toString(){
          String filher = "%";
          
          return String.valueOf(getDocumentno()).concat(filher).concat(String.valueOf(getId())).concat(filher)
                 .concat(getDoctype()).concat(filher).concat(getPathname()).concat(filher).concat(String.valueOf(getSize()))
                 .concat(filher).concat(String.valueOf(getReportdate().getTime())).concat(filher)
                 .concat(String.valueOf(getStatus())).concat(filher).concat(String.valueOf(getByteswritten()))
                 .concat(filher).concat(String.valueOf(getPid())).concat(filher).concat(String.valueOf(getRowid()));
      }

      protected void setDoctype (String p_Doctype){
          if (p_Doctype == null)
          {
              LiveDB.ALRTLVL = 7;
              System.err.println(LiveDB.grassHopper("Invalid p_Doctype=null"));
              System.exit(1);
          }
          rowlock.acquireUninterruptibly();          
          this.doctype = new String(p_Doctype.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Doctype.length()));
          rowlock.release();
      }
      protected void setPathname (String p_Pathname){
          if (p_Pathname == null)
          {
              LiveDB.ALRTLVL = 7;
              System.err.println(LiveDB.grassHopper("Invalid p_Pathname=null"));
              System.exit(1);
          }
          rowlock.acquireUninterruptibly();
          if(LiveDB.command_line.equals("FileTableDataFileMap") == false)
             pathname = new String(p_Pathname.toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Pathname.length()));
          else
          {
              if ((p_Pathname = LiveDB.convertFileName(p_Pathname)) != null)
                  pathname = new String((p_Pathname = LiveDB.convertFileName(p_Pathname)).toCharArray()).concat(new String(LiveDB.BlankPads.toCharArray(),0,LiveDB.MAXSTRINGFLDLEN - p_Pathname.length()));
          }
          rowlock.release();
      }
      protected void setSize ( long p_Size ){
          rowlock.acquireUninterruptibly();
          this.size = p_Size;
          rowlock.release();
      }
      protected void setReportdate (Date p_Reportdate){
          rowlock.acquireUninterruptibly();
          this.reportdate = new Date(p_Reportdate.getTime()); // carefull with this setting package private
          rowlock.release();
      }
      protected void setStatus (char p_Status){
          rowlock.acquireUninterruptibly(); 
          this.status = p_Status; // carefull with this setting package private
          rowlock.release();
      }
      protected void setByteswritten (long p_Byteswritten){
          rowlock.acquireUninterruptibly();
          this.byteswritten = p_Byteswritten; // carefull with this setting package private
          rowlock.release();
      }
      protected void setPid (long p_Pid){
          rowlock.acquireUninterruptibly();
          this.pid = p_Pid;  // carefull with this setting package private
          rowlock.release();
      }

      protected void setRowid (long p_rowid){
        rowlock.acquireUninterruptibly();
        if (LiveDB.LiveSharedDB[rtn].length == 0 && p_rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_rowid)));
            System.exit(1);
        }

        if (p_rowid >= 0)
            rowid = p_rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_rowid));
            System.exit(1);
        }
        rowlock.release();
     }

     /* TreeSet will by default look for natural order
      * but since this is available to help with sort
      * it will use this comparable method
      */
    @Override
    public int compareTo(Object p_O){
          if (id == ((ReportsTable)p_O).id)
              return 0;
          else if (id < ((ReportsTable)p_O).id)
              return -1;
          else
              return 1;
    }

    // although each instance of this Class ReportsTable represents a single row
    // we are committing the entire table here but before that a few checks and balances
    @Override
    public final boolean COMMIT(){
        
        boolean fnd = false;

        // we are committing the ReportsTable so mark those rows with change_flag 'F'
        for (int rw =  (int) LiveDB.CTSR; rw < LiveDB.LiveSharedDB[ct].length; rw++)
        {
            if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno() == rtn
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K'
                && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T')
            {
                   ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('C');
                   fnd = true;
                   // insert an entry in ChangeTracker table for this table
                   // Basically what this does is to backup this entirerow where you
                   // setChange_flag('C') otherwise database reconstruct of ChangeTracker will fail.
                   LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                   new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                   LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
            }
        }

        if (fnd)
        {
            // for the sake of dbfRepair we take a restore file which MAY be a BAD copy
            // since we are not checking for the presence of a .syc file on disk. If we
            // look for a .syc file on disk then this commit will take very very long.
            LiveDB.CopyFile(LiveDB.serializedfilespath 
                    + ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][rtn]).getFilename() + ".ser",
                LiveDB.serializedfilespath  + "backup" + LiveDB.FILESEPERATOR +
                ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][rtn]).getFilename() + ".ser" + ".res");
            
            if ( LiveDB.commit_database(rtn) == false )
                return false;
        }
        
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        if (LiveDB.LiveSharedDB[rtn].length == 0)
            return false;
        return LiveDB.write_IndexFile(rtn);
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(rtn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }

    @Override
    public final void writeExternal (ObjectOutput stream) throws IOException {
        rowlock.acquireUninterruptibly();
        int itr = -1; // Installed table row in InstalledTables

        // first find the rowid for this table ctn = 14 in InstalledTables
        if (LiveDB.indxinstalledTables.get(Integer.toString(rtn)) == null)
        {
            LiveDB.ALRTLVL = 4;
            System.err.println(LiveDB.grassHopper("ReportsTable misses entry in indxinstalledTables"));
            // perhaps the index is corrupted or has not been updated someplace so check the physical table
            for (int i = 0; i < LiveDB.LiveSharedDB[itn].length; i++)
            {
                if (((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getTableno() == rtn)
                   itr = (int)((InstalledTables)LiveDB.LiveSharedDB[itn][i]).getRowid();
            }

            if (itr == -1 || itr > LiveDB.maxtables)
            {
                LiveDB.ALRTLVL = 2;
                System.err.println(LiveDB.grassHopper("ReportsTable misses entry in indxinstalledTables"));
                System.exit(1);
            }
        }
        else
            itr = (int)(long)LiveDB.indxinstalledTables.get(Integer.toString(rtn));

        srf = new File (((FileTableDataFileMap)LiveDB.LiveSharedDB[ft]
                [(int)((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).getTableno()]).getFilename());

        // outsize should be negative before every write
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);

        try{
              stream.writeLong(documentno);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class ReportsTable Field documentno"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(id);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "writeExternal Class ReportsTable Field id"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(doctype);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field doctype"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeUTF(pathname);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field pathname"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(size);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field size"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(reportdate.getTime());
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field reportdate"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeChar(status);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field status"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(byteswritten);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "ReportsTable Field byteswritten"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(pid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field pid"));
              throw new IOException ();
        }

        /*
        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }*/

        // outsize should be negative before every write
        // uncomment this line if you are writing code for Ultra Fast commits
        // ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length() * -1);

        try{
              stream.writeLong(rowid);
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field rowid"));
              throw new IOException ();
        }


        try{
            stream.flush();
        }catch (IOException ex_NOKIA73){
            System.err.println(LiveDB.grassHopper(ex_NOKIA73, "flush failed during ReportsTable serialization"));
            System.exit(1);
        }

        // setOutsize to positive value after every successfull write
        // so if there is a power outage in between we will know that there has been a serialization
        // failure and the serialized file is corrupted and take necessary steps to restpre the
        // serialized file from CrashRecovery or ChangeTracker or if both CrashRecovery and ChangeTracker
        // are corrupted then restore from last autobackup
        ((InstalledTables)LiveDB.LiveSharedDB[itn][itr]).setOutsize(srf.length());
        LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
        new ChangeTracker (itn,itr, 'U', LiveDB.pid, new Date (),
        LiveDB.LiveSharedDB[itn][itr], (long)LiveDB.LiveSharedDB[ct].length);
        rowlock.release();
    }

    @Override
    public final void readExternal (ObjectInput stream) throws IOException, EOFException {
        rowlock.acquireUninterruptibly();
        // now read in all values here
        try{
              documentno = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field documentno"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field documentno"));
              throw new IOException ();
        }
        try{
              id = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field id"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field id"));
              throw new IOException ();
        }
        try{
              doctype = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field doctype"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "Field doctype"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field doctype"));
              throw new IOException ();
        }
        try{
              pathname = stream.readUTF();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field pathname"));
              throw new EOFException ();
        }catch (UTFDataFormatException ex_NOKIA93){
              System.err.println(LiveDB.grassHopper(ex_NOKIA93, "Field pathname"));
              throw new IOException (); // fake it to deque it from LiveDB.TrainsInWaiting
              // LiveDB.system_exit = 1;
              // LiveDB.SmoothExitCondition.signalAll();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field pathname"));
              throw new IOException ();
        }
        try{
              size = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field size"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field size"));
              throw new IOException ();
        }
        try{
              reportdate = new Date(stream.readLong());
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field reportdate"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field reportdate"));
              throw new IOException ();
        }
        try{
              status = stream.readChar();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field status"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field status"));
              throw new IOException ();
        }
        try{
              byteswritten = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field byteswritten"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field byteswritten"));
              throw new IOException ();
        }
        try{
              pid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field pid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field pid"));
              throw new IOException ();
        }
        
        try{
              rowid = stream.readLong();
        }catch (EOFException ex_NOKIA92){
              System.err.println(LiveDB.grassHopper(ex_NOKIA92, "Field rowid"));
              throw new EOFException ();
        }catch (IOException ex_NOKIA91){
              System.err.println(LiveDB.grassHopper(ex_NOKIA91, "Field rowid"));
              throw new IOException ();
        }
        rowlock.release();
    }
}

/* Comparator interfaces for comparing reportdate and status on ReportsTable */
// TreeSet will look for these to sort if defined.
class DateComp implements Comparator {
    @Override
    public int compare ( Object p_O1, Object p_O2 ) {
        return ((((ReportsTable)p_O1).getReportdate()).compareTo(((ReportsTable)(p_O2)).getReportdate()));
    }
}

class StatusComp implements Comparator {
    @Override
    public int compare (Object p_O1, Object p_O2){
        if (((ReportsTable)p_O1).getStatus() < ((ReportsTable)p_O2).getStatus())
                return -1;
        else if (((ReportsTable)p_O1).getStatus() == ((ReportsTable)p_O2).getStatus())
                return 0;
        else
            return 1;
    }
}
