/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.Serializable;
import java.util.Date;
import java.util.concurrent.Semaphore;

/*
 * WIRING DIAGRAM :
 * @LOOKSY: SearchDatabase.java
 * @LOOKSY:ArtLeavesInterfaces.java
 *
 * Class Name : public class SearchDatabase implements Serializable, Tables
 *
 * Methods : Look in Class definition
 *
 * The Tables Interface has been faked here except for REFRESH because it is too very dangerous to provide it.
 *
 * Data Attributes : Look in Class definition for a complete list also there is substantial documentation
 * next to every field definition to assist understanding.
 *
 * Relevant Indexes : LiveDB.indxsearchdatabase
 * searchkey = any keyword
 * returns String picture code.
 *
 * Data File : YES the name of the dataentry file will be contained in FileTableDataFileMap. The data file
 * will contain the different search keywords for each picture code. A readymade data entry file will be
 * provided to you dont loose that take a physical backup and keep safe.
 *
 * @Since Version 1.0
 *
 * Running Commentary : This table lists the different search keywords for each picture code.
 *
 * Technical References : Buy and Read Asterix and the Cauldron.
 * Table Dedicated to : Al Jarreau (TENDERNESS)
 *
 * Start Date: 18th October 2010
 * End Date: 13th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

public class SearchDatabase implements Serializable, Tables {
    private String keyword ; /* Keyword for search
                              * Some Keywords will be added
                              */
    private String pic ; /* Picture Code */
    private long rowid; /* 3 FIELDS */

    private transient int sdn;
    private transient Semaphore rowlock;

    {
       rowlock = new Semaphore(1, true);
       sdn = LiveDB.sd;
    }

    protected SearchDatabase (){
         keyword = pic = null;
         rowid = 0L;
    }

    public SearchDatabase (String p_Keyword, String p_Pic, long p_Rowid) {
        
        setKeyword(p_Keyword);
        rowlock.acquireUninterruptibly();
        if (p_Pic == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Pic=null"));
            System.exit(1);
        }
        pic = new String(p_Pic.toCharArray());
        if (LiveDB.LiveSharedDB[sdn].length == 0 && p_Rowid != 0)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid != ((SearchDatabase)LiveDB.LiveSharedDB[sdn][LiveDB.LiveSharedDB[sdn].length -1]).getRowid() + 1)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid rowid:" + String.valueOf(p_Rowid)));
            System.exit(1);
        }
        if (p_Rowid >= 0)
            rowid = p_Rowid;
        else
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Negative Rowid:" + p_Rowid));
            System.exit(1);
        }
        rowlock.release();
    }

    /* Getters and Setters provide a public interface to private data */
    /* They follow the rulebook of encapsulation */
    /* Dont provide Setters for all fields specially key fields !! */

    protected String getKeyword (){
        return new String(keyword.toCharArray());
    }
    protected String getPic(){
        return new String(pic.toCharArray());
    }
    protected long getRowid() {
        return rowid;
    }

    @Override
    public String toString(){
        String filher = "%";
        return getKeyword().concat(filher).concat(getPic()).concat(filher).concat(String.valueOf(getRowid()));
    }
    
    protected void setKeyword (String p_Keyword){
        if (p_Keyword == null)
        {
            LiveDB.ALRTLVL = 7;
            System.err.println(LiveDB.grassHopper("Invalid p_Keyword=null"));
            System.exit(1);
        }
        rowlock.acquireUninterruptibly();
        for (int i = 0; i < p_Keyword.length(); i++)
        {
            if (Character.isDefined(p_Keyword.toCharArray()[i]) == false
                || Character.isDigit(p_Keyword.toCharArray()[i])
                || Character.isMirrored(p_Keyword.toCharArray()[i]))
            {
                LiveDB.ALRTLVL = 15;
                System.err.println(LiveDB.grassHopper("Invalid keyword " + p_Keyword));
                System.exit(1);
            }
        }
        // String is immutable but the reference is copied and the reference is a Call by Value(local)
        // which will cease to exist after method exits
        this.keyword = new String(p_Keyword.toCharArray());
        rowlock.release();
    }
    @Override
    public final boolean COMMIT(){
        return true;
    }

    @Override
    public final boolean COMMITINDEXES (){
        return true;
    }

    @Override
    public final boolean REFRESH(){
        if (RefreshDatabase.read_database(sdn) == false)
           return false;
        return true;
    }

    // this is a programming interface if you wish to change the logic of the Thin Client
    // but it will be slow
    // CAUTION : THIS DOES NOT ROLLBACK THE ENTIRE TRANSACTION
    // IT WILL ROLBACK THIS TABLE
    // to ROLLBACK transaction use LiveDB.TRANSACTIONROLLBACK()
    @Override
    public final boolean ROLLBACK (boolean lastsavepoint){
        // not for this table it is very dangerous
        return false;
    }

    @Override
    public final boolean ROLLBACK(){
        return ROLLBACK(false);
    }

    @Override
    public final boolean ROLLFORWARD (){
        return ROLLFORWARD (null, null);
    }

    @Override
    public final boolean ROLLFORWARD (Date p_fromtime, Date p_upto_time){
       // this cannot be provided for this table too very dangerous
        return false;
    }
}
