/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package DatabaseEngine;

import java.io.File;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.Date;
import java.util.Random;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import javax.swing.JOptionPane;

/*
 * WIRING DIAGRAM :
 * THREAD NAME : InstallDB.threadwritetables
 * CLASS NAME : WriteTables
 * @LOOKSY : ChangeTracker.java
 * @LOOKSY : FileTableDataFileMap.java
 * @LOOKSY : CustomerRegister.java
 * @LOOKSY : ShoppingCart.java
 * @LOOKSY : CrashRecovery.java
 * @LOOKSY : InstalledTables.java
 * @LOOKSY : ReportsTable.java
 * @LOOKSY : PlatformClearance.java
 * @LOOKSY:ThreadController.TakeOffTimes[], ThreadController.LandingTimes[]
 *         Dont read the entire ThreadController.java you will go mad.
 * @LOOKSY: LiveDB.LogFileSabotageCheck(), LiveDB.ENABLEFASTCOMMIT, LiveDB.system_exit,
 *          LiveDB.SmoothExitCondition, LiveDB.TrainsInWaiting, LiveDB.SLEEPTIME
 * @LOOKSY:WriteTables.java run () method
 *
 * Method Name : public void run ()
 * Auxilliary Methods : private boolean write_changetracker (int p_uptorow)
 *                      private String get_pf (int p_tbl)
 *
 * Parameters : None
 *
 * @Since Version : 1.0
 *
 * Running Commentary : This is the fastcommit thread. Actually this is more of a Intermediate commit than
 * fastcommit. Ultra Fast Commits will come up in the next release where only the field changes will be
 * committed to disk. After this thread is started it keep running as a demon until LiveDB.system_exit == 0.
 * It will fast commit only those records appearing in ChangeTracker after you logged in as a user specific
 * to your LiveDB.pid provided getChange_flag() == 'U' || 'I' || 'D'. If change flag is any of these it
 * calculates the SIZEOF() for each table record and skips those many bytes in the underlying serialized
 * file and then writes the changed record spot on. But before the write it needs to queue the request
 * in LiveDB.TrainsInWaiting. Go thru the source code for the PlatformClearance.java table it does some
 * important synchronization using an .syc file. There is also a special full stop entry to chop this long
 * snake called ChangeTracker. check out the source below.
 * 
 * Technical Reference : None
 *
 * Start Date: 18th October 2010
 * End Date: 11th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

class WriteTables extends LiveDB implements Runnable {

    @Override
    public void run () {
        boolean ok = false;
        int tb = 2;
        int requestno = (new Random ()).nextInt();
        long rid = 0;
        int skb = 0; // skip bytes in File
        String sf = null;
        File psf = null; // persistent storage file
        FileOutputStream fos = null;        
        FileChannel ch = null;
        FileLock fl = null;        
        ObjectOutputStream oos = null;        
        Lock WriteObjectLock = new ReentrantLock ();
        TrainRunningInformation myrunninginfo = null;
        PlatformClearance mytrain = null;

        if (new File(LiveDB.serializedfilespath + "daemon").canWrite())
        {
            System.err.println(LiveDB.grassHopper("Perms Tampered File:" + LiveDB.serializedfilespath
                                                                                           + "daemon"));
        }
        else
        {
            if (new File(LiveDB.serializedfilespath + "daemon").setWritable(true, false) == false)
            {
                System.err.println(LiveDB.grassHopper("Cant setWritable File:"
                    + LiveDB.serializedfilespath + "daemon"));
                LiveDB.system_exit = 1;
                LiveDB.SmoothExitCondition.signalAll();
            }
        }

        try
        {
            new File (LiveDB.serializedfilespath + "daemon" + LiveDB.FILESEPERATOR + "threadwritetables"
                  + LiveDB.pid + LiveDB.NODENAME + ".daemon" ).createNewFile();
        }catch (IOException ex_YESKIA41){
            System.err.println(LiveDB.grassHopper(ex_YESKIA41, "Cant create DEMONIC File:" +
               LiveDB.serializedfilespath + "daemon" + LiveDB.FILESEPERATOR + "threadwritetables"
               + LiveDB.pid + LiveDB.NODENAME + ".daemon"));
        } // to hell with this error it wont happen continue processing

        do{
             LiveDB.LogFileSabotageCheck ();
             if (LiveDB.ENABLEFASTCOMMIT == true)
             {
             try{
                ThreadController.TakeOffTimes[5] = new Date ();

                // FASTCOMMIT ONLY TABLES, 0,1,2,3,6,13,14,15,17 THAT WILL SEE CHANGES
                // THE OTHER TABLES WILL BE LOADED AT INSTALL TIME AND THEN THEY ARE SERIALIZED AND THE
                // APPLICATION WILL AUTO EXIT AND YOU WILL HAVE TO RESTART THE APPLICATION FOR USE BY
                // NORMAL USERS. AFTER THIS THOSE TABLES REMAIN UN-TOUCHED AND SERIALIZED ON DISK
                // I AM BLANK PADDING ALL STRING DATA FIELDS BECAUSE OF THIS FASTCOMMIT OTHERWISE
                // A STRING FIELD MAY CONTAIN VARIABLE SIZE AND THEN I WOULD HAVE TO USE RANDOM ACCESS FILE
                // TO READ A BYTE BUFFER SEEK TO THE RECORD AND WRITE AND THEN RE-WRITE THE READ BUFFER
                // WHICH WILL BE A SLOW COMMIT INSTEAD OF A FAST COMMIT MECHANISM
                
                for (int rw = 0; rw < LiveDB.LiveSharedDB[ct].length; rw++)
                {
                    if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().after(LiveDB.today) == true
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'I'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'D'
                       && ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getPid() == LiveDB.pid )
                    {
                        sf = get_pf (tb = (int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getTableno());
                        psf = new File(sf);
                        try{
                              fos = new FileOutputStream(psf, true); // filechannel will possition at EOF
                           }catch (FileNotFoundException ex_NOKIA55){
                              write_changetracker (-1);
                              LiveDB.ALRTLVL = 2;
                              System.err.println(LiveDB.grassHopper(ex_NOKIA55, "Intruder wiped out underlying file  "
                              + Integer.toString(tb) + " " + sf));
                              // smooth exit
                              LiveDB.system_exit = 1;
                              LiveDB.SmoothExitCondition.signalAll();
                           }
                        try{
                              oos = new ObjectOutputStream (fos);
                            }catch (IOException ex_NOKIA65){
                              write_changetracker (-2);
                              System.err.println(LiveDB.grassHopper(ex_NOKIA65, "WriteTables while opening ObjectOutputStream  "
                              + Integer.toString(tb) + " " + sf));
                              // smooth exit
                              LiveDB.system_exit = 1;
                              LiveDB.SmoothExitCondition.signalAll();
                            }

                        // even if this flag is D write the row because if you dont then row will have to
                        // renumbered which will take a long time.
                        if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                            || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'D' )
                        {
                             for (int tbr = 0; tbr < ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid();
                                      tbr++)
                             {
                                 switch (tb){
                                     case ft:
                                         skb += ((FileTableDataFileMap)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case ct:
                                         skb += ((ChangeTracker)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case id:
                                         skb += ((InstallerDues)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case cr:
                                         skb += ((CustomerRegister)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case cp:
                                         skb += ((CustomerPayments)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case us:
                                         skb += ((UserSubmits)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case crr:
                                         skb += ((CrashRecovery)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                     case it:
                                         skb += ((InstalledTables)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case rt:
                                         skb += ((ReportsTable)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     case pc:
                                         skb += ((PlatformClearance)LiveDB.LiveSharedDB[tb][tbr]).SIZEOF();
                                         break;
                                     default:
                                        LiveDB.ALRTLVL = 4;
                                        System.err.println(LiveDB.grassHopper("non-Update Table has Flag in ChangeTracker table File:" + sf));
                                        Spy.dump();
			                JOptionPane.showMessageDialog( appframe,"IMPORTANT INFORMATION YOUR LOG FILE NAME FOR THIS SESSION IS : " + LiveDB.ErrorLogFile + "\n");
                                        System.exit(1);
                                        break;
                                 } // switch (tb)
                             } // for (int tbr = 0; tbr <= ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid();
                         } // if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U')

                         WriteObjectLock.lock();
                         IPCSCondition[ThreadController.getcurrenthread()] = WriteObjectLock.newCondition();
                         myrunninginfo = new TrainRunningInformation (tb,
                         Thread.currentThread().getId(), Thread.currentThread().getName(),
                         Thread.currentThread().getThreadGroup().getName(), null, null, null, 0L);
                         LiveDB.LiveSharedDB[pc][LiveDB.LiveSharedDB[pc].length] =
                         new PlatformClearance ( requestno, LiveDB.pid, LiveDB.Action.FASTCOMMIT,
                         new Date (), myrunninginfo, LiveDB.LiveSharedDB[pc].length);
                         mytrain = (PlatformClearance) LiveDB.LiveSharedDB[pc][LiveDB.LiveSharedDB[pc].length -1];
                         ch = fos.getChannel();
                         try{
                                  if (LiveDB.QueueTrainsInWaiting(mytrain, ch, fl, requestno) == false)
                                  {
                                      System.err.println(LiveDB.grassHopper("Error QueueTrainsInWaiting table :" +
                                      mytrain.getTrainid().getTrain()));
                                      write_changetracker (-3);
                                      LiveDB.system_exit = 1;
                                      LiveDB.SmoothExitCondition.signalAll();
                                  }
                                  /* since this is just an advisory lock on some operating systems
                                   * set the lock_flag to "L" in fileTableDataFileMap
                                   */
                                  ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setTable_lock("L");
                                  // insert an entry in ChangeTracker table for this table
                                  LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                  new ChangeTracker ((long)ft,tb, 'U',LiveDB.pid, new Date (),
                                  LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                  LiveDB.LASTLOCKMETHOD = "WriteTables";

                                  ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setPid (LiveDB.pid);
                                  // insert an entry in ChangeTracker table for this table
                                  LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                  new ChangeTracker ((long)ft,tb, 'U', LiveDB.pid, new Date (),
                                  LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                                                  
                                  if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'I')
                                     // do we have to flush to get correct ch.size() check during testing
                                     ch = ch.position(ch.size() + 1L); // some other pid/thread may have grown the file
                                  else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                                          || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'D' )
                                     ch = ch.position(skb + 1L);
                                                                                      
                                         switch (tb){
                                           case ft:
                                              ((FileTableDataFileMap)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case ct:
                                              ((ChangeTracker)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              // Basically what this does is to backup this entirerow where you
                                              // setChange_flag('F') otherwise database reconstruct of ChangeTracker will fail.
                                              // But in this case of ChangeTracker table this huge snake called ChangeTracker
                                              // has to full stop somewhere before it devours your hard disk at 512MB per year.
                                              // so instead of a call to constructor with update flag 'U'

                                              /* LiveDB.LiveSharedDB[ctn][LiveDB.LiveSharedDB[ctn].length] =
                                                 new ChangeTracker ((long)ctn,rw, 'U', LiveDB.pid, new Date (),
                                                 LiveDB.LiveSharedDB[ctn][rw],(long)LiveDB.LiveSharedDB[ctn].length); */

                                              // we will call the constructor with update flag 'F' so it will backup
                                              // the change that you just made and also chop the tail of this snake.
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'F',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case id:
                                              ((InstallerDues)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case cr:
                                              ((CustomerRegister)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case cp:
                                              ((CustomerPayments)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case us:
                                              ((UserSubmits)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case crr:
                                              ((CrashRecovery)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case it:
                                              // It should not have insert records at this FastCommit Stage
                                              if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'U'
                                                  || ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'D')
                                                  ((InstalledTables)LiveDB.LiveSharedDB[tb]
                                                  [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              else if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() == 'I')
                                              {
                                                  LiveDB.ALRTLVL = 4;
                                                  System.err.println(LiveDB.grassHopper("non-Insert Table has Flag in ChangeTracker table File:" + sf));
                                                  Spy.dump();
			                          JOptionPane.showMessageDialog(appframe, "IMPORTANT INFORMATION YOUR LOG FILE NAME FOR THIS SESSION IS : " + LiveDB.ErrorLogFile + "\n");
                                                  System.exit(1);
                                              }
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U',LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case rt:
                                              ((ReportsTable)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           case pc:
                                              ((PlatformClearance)LiveDB.LiveSharedDB[tb]
                                              [(int)((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getRowid()]).
                                                      writeExternal(oos);                                         
                                              ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
                                              // insert an entry in ChangeTracker table for this table
                                              LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                              new ChangeTracker ((long)ct,rw, 'U', LiveDB.pid, new Date (),
                                              LiveDB.LiveSharedDB[ct][rw],(long)LiveDB.LiveSharedDB[ct].length);
                                              break;
                                           default:                                              
                                              if (tb != ft && tb != ct && tb != cr && tb != id && tb != cp && tb != pc && tb != crr && tb != it && tb != rt )
                                              {
                                                  LiveDB.ALRTLVL = 7;
                                                  System.err.println(LiveDB.grassHopper("Invalid Table No:" + String.valueOf(tb)));
                                                  write_changetracker (-4);
                                                  LiveDB.system_exit = 1;
                                                  LiveDB.SmoothExitCondition.signalAll();
                                              }
                                              break;
                                         } // switch
                                     
                                     try{
                                           fl.release();
                                     }catch (ClosedChannelException ex_NOKIA52){
                                        System.err.println(LiveDB.grassHopper(ex_NOKIA52, "Error fl.release()"));
                                     }

                                     ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setTable_lock("U");
                                     // insert an entry in ChangeTracker table for this table
                                     LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                     new ChangeTracker ((long)ft,tb, 'U', LiveDB.pid, new Date (),
                                     LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                     ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setPid(LiveDB.pid);
                                     // insert an entry in ChangeTracker table for this table
                                     LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                     new ChangeTracker ((long)ft,tb, 'U', LiveDB.pid, new Date (),
                                     LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                     oos.close(); // CLOSE

                                     if (LiveDB.DeQueueTrainsInWaitingArrivals() == false)
                                     {
                                         LiveDB.ALRTLVL = 7; // programming screwup most probably
                                         System.err.println(LiveDB.grassHopper("Error DeQueueTrainsInWaitingArrivals()"));
                                     }
                            }catch (IOException ex_NOKIA43){
                                if (LiveDB.DeQueueTrainsInWaiting() == false)
                                {
                                   LiveDB.ALRTLVL = 7; // programming screwup most probably
                                   System.err.println(LiveDB.grassHopper("Error DeQueueTrainsInWaiting()"));
                                }
                                ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setTable_lock("U");
                                // insert an entry in ChangeTracker table for this table
                                LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                new ChangeTracker ((long)ft,tb, 'U', LiveDB.pid, new Date (),
                                LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                ((FileTableDataFileMap)(LiveDB.LiveSharedDB[ft][tb])).setPid(LiveDB.pid);
                                // insert an entry in ChangeTracker table for this table
                                LiveDB.LiveSharedDB[ct][LiveDB.LiveSharedDB[ct].length] =
                                new ChangeTracker ((long)ft,tb, 'U', LiveDB.pid, new Date (),
                                LiveDB.LiveSharedDB[ft][tb],(long)LiveDB.LiveSharedDB[ct].length);
                                System.err.println(LiveDB.grassHopper(ex_NOKIA43));
			        JOptionPane.showMessageDialog(appframe, "Fatal Error Check Log File Name " + LiveDB.ErrorLogFile + "\n");
			        JOptionPane.showMessageDialog(appframe, "OBJECT SERIALIZATION ERROR" + sf + "\n");
                                write_changetracker (rw);
                                LiveDB.system_exit = 1;
                                LiveDB.SmoothExitCondition.signalAll();
                          }finally{WriteObjectLock.unlock();}
                   } // if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getDmldate().after(LiveDB.today)
                   skb = 0;
                } // for (int rw = 0; rw < LiveDB.LiveSharedDB[ct].length; rw++)

            // write out the ChangeTracker entry because of that
            // full stop entry in case 1: where the update flag was faked as 'F'
            if (LiveDB.commit_database(ct) == false)
               System.err.println(LiveDB.grassHopper("ChangeTracker slow commit failed."));

            ThreadController.LandingTimes [5] = new Date ();
            // we really wont need this sleep call except in a windows environment where processes of
            // lower priority will get runnable state otherwise they will starve to death.
            Thread.sleep(LiveDB.SLEEPTIME); // so reduce this sleep time to a ridiculously low amt
                                            // since it is in a loop
            }catch (InterruptedException ex_YESKIA68) {
                   if (write_changetracker (-5) == false)
                   {
                        LiveDB.system_exit = 1;
                        LiveDB.SmoothExitCondition.signalAll();
                   }
            }
         } // if LiveDB.ENABLEFASTCOMMIT == true
         else
         {
            ThreadController.TakeOffTimes[5] = new Date ();
            ThreadController.LandingTimes [5] = new Date ();
         }            
            
        } while (LiveDB.FASTCOMMITEXIT == false);

        if (new File (LiveDB.serializedfilespath + "daemon" + LiveDB.FILESEPERATOR + "threadsysexit"
                  + LiveDB.pid + LiveDB.NODENAME + ".daemon" ).delete() == false)
        {
            System.err.println(LiveDB.grassHopper("Cant delete DEMONIC File:" +
               LiveDB.serializedfilespath + "daemon" + LiveDB.FILESEPERATOR + "threadsysexit"
               + LiveDB.pid + LiveDB.NODENAME + ".daemon"));
        }

        if (new File (LiveDB.serializedfilespath + "daemon").setReadOnly() == false)
        {
            LiveDB.ALRTLVL = 3;
            System.err.println(LiveDB.grassHopper("Cant setReadonly File:"
                    + LiveDB.serializedfilespath + "daemon"));
        }


        ThreadController.LandingTimes [5] = new Date ();
        // DONE BUT IF YOU WANT A TRULY ULTRA-FAST COMMIT PAISA MUNGTAH HAI USD $100
        // SPECIAL RELEASE FOR YOU WITHOUT SOURCE CODE !!! PAISA BECAUSE YOU CREEPS WILL DO A PHOKATCOPY
    }


    private String get_pf (int p_tbl){        
        return ((FileTableDataFileMap)LiveDB.LiveSharedDB[ft][p_tbl]).getFilename() + ".ser";
    }

/*
 * WIRING DIAGRAM :
 * @LOOKSY:ChangeTracker.java
 * @LOOKSY:LiveDB.commit_database()
 * @LOOKSY:WriteTables.java run () method
 *
 * Method Name : private boolean write_changetracker (int p_uptorow)
 * Parameters : int p_uptorow (pass a signed value also)
 * Returns : boolean true if LiveDB.commit_database(ct) is successfull
 *           boolean false if LiveDB.commit_database(ct) fails
 *
 * @Since Version : 1.0
 *
 * Running Commentary : This can do a commit of ChangeTracker table without updating those flags
 * There are certain exceptions thrown when you dont want to update those flags but you still want to
 * write the ChangeTracker table. For example if that write external throws a IOException you do want
 * to update those flag and also serialize the entire ChangeTracker table to disk using slow commit !!
 * This ChangeTracker table is a real pain in the ..... but it is needed for any kind of disaster
 * recovery. So dont loose it at any cost. Even a corupted rotten one might save the day somewhat.
 * 
 * Start Date: 18th October 2010
 * End Date: 11th May 2011
 *
 * AUTHOR : Padmanabhan Sai Krishna
 * Email : pskris@sify.com
 * Phone(s) : +91-80-41689002
 *            +91-80-23624473
 * Mobile : +91-9900516842
 * Residential Address for Darshan : C-1678, Sahakara Nagar, Bangalore - 92.
 */

    private boolean write_changetracker (int p_uptorow){  
        // pass this as signed int to prevent setChange_flag('F')
        
        // now update the ChangeTracker table for entries to ChangeTracker table itself
        for (int rw = 0; rw < p_uptorow ; rw++)
        {
              // FLAG K IS DATABASE ROLLBACK, FLAG T IS DATABASE RECONSTRUCT
              if (((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'F' &&
                  ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'C' &&
                  ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'K' &&
                  ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).getChange_flag() != 'T' )

                  ((ChangeTracker)LiveDB.LiveSharedDB[ct][rw]).setChange_flag('F');
        }
        // now commit ChangeTracker table itself which will always have changes to it
        return LiveDB.commit_database(ct);
 
    } // private void write_changetracker ()
}
